/*******************************************************************************
*
*                   Copyright 2003,MARVELL SEMICONDUCTOR ISRAEL, LTD.
* THIS CODE CONTAINS CONFIDENTIAL INFORMATION OF MARVELL.
* NO RIGHTS ARE GRANTED HEREIN UNDER ANY PATENT, MASK WORK RIGHT OR COPYRIGHT
* OF MARVELL OR ANY THIRD PARTY. MARVELL RESERVES THE RIGHT AT ITS SOLE
* DISCRETION TO REQUEST THAT THIS CODE BE IMMEDIATELY RETURNED TO MARVELL.
* THIS CODE IS PROVIDED "AS IS". MARVELL MAKES NO WARRANTIES, EXPRESSED,
* IMPLIED OR OTHERWISE, REGARDING ITS ACCURACY, COMPLETENESS OR PERFORMANCE.
*
* MARVELL COMPRISES MARVELL TECHNOLOGY GROUP LTD. (MTGL) AND ITS SUBSIDIARIES,
* MARVELL INTERNATIONAL LTD. (MIL), MARVELL TECHNOLOGY, INC. (MTI), MARVELL
* SEMICONDUCTOR, INC. (MSI), MARVELL ASIA PTE LTD. (MAPL), MARVELL JAPAN K.K.
* (MJKK), MARVELL SEMICONDUCTOR ISRAEL. (MSIL),  MARVELL TAIWAN, LTD. AND
* SYSKONNECT GMBH.
*
********************************************************************************
* main.c - C File for implementation of BIOS Intermediate Application Layer.
*
* DESCRIPTION:
*       None.
*
* DEPENDENCIES:
*   mvOs.h
*   mvSata.h.   
*   mvStorageDev.h
*   mvRegs.h
*
* FILE REVISION NUMBER:
*       $Revision: 1.7 $
*******************************************************************************/

#include "mvOs.h"
#include "mvBiosIal.h"
#include "mvSata.h"
#include "mvStorageDev.h"
#include "mvIALCommonUtils.h"

/* Defines */

/* CHS geometry definition */
#define HEADS   255
#define SECTORS 63
#define CYLINDER 1022
#define USE_REGISTER_ACCESS_PROTECTED_MODE

/* Global variables */
/* Global descriptors table */
unsigned int GDT[] = {  0xf, 0, 0, 0, 
    0xffff, 0, 0x9200, 0x8f};

/* mvSata Adapter location on PCI bus */
unsigned long mvSataBarOffset;
unsigned char mvSataResult;
unsigned char mvSataDeviceNum;
unsigned char mvSataBusNum;
unsigned short mvSataDeviceId;
unsigned char mvSataRevisionId;

/* Local stack */
#define stackSize 0x600
unsigned short stack[stackSize/2] = {0,};
unsigned short oldStack, oldStackSegment, oldDataSegment;

unsigned short savedInt13CodeSegment, savedInt13InstructionPointer;
unsigned int   alreadySavedInt13Vector = 0;
unsigned int   alreadyInitializedChannelsFlag = 0;
unsigned int   alreadyHookedInt13 = 0;
unsigned short bcvEsDataSegment;
unsigned short bcvDiPointer;
/* drivesNumbers is for checking if the adapter owns the drive with the int13 */
unsigned char drivesNumbers[MV_SATA_CHANNELS_NUM] = 
{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

/*
 * sataChannelNumner is used as a SATA channel in case a match is found using
 * drivesNumers
 */
MV_U8 sataChannelNumber[MV_SATA_CHANNELS_NUM] =
{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
/*
 * portMultiplierPort is used a device port number on the port multiplier in
 * case a match is found using drivesNumbers.
 */
MV_U8 portMultiplierPort[MV_SATA_CHANNELS_NUM] =
{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

unsigned short pnpCheckSegment, pnpCheckOffset, pnpCheckResult;
MV_U8          requestedAllignment = 2; /* Default buffer allignment is 2 bytes */
/* Data structure describing mvSata adapter and channels */
HW_ADAPTER_DESCRIPTION sataAdapter ={0,};
MV_BOOLEAN     useIoBar = MV_FALSE;

/* Local functions */
static MV_BOOLEAN findAndInitChannels(void);
MV_BOOLEAN StartChannel( PHW_ADAPTER_DESCRIPTION HwDeviceExtension,
                         MV_U8 channelIndex, MV_BOOLEAN PMEnabled, MV_U8 PMPort);
static MV_U16 readSystemClockTimer(void);
static MV_BOOLEAN
StartPM(PHW_ADAPTER_DESCRIPTION HwDeviceExtension, MV_U8 channelIndex,
        MV_BOOLEAN *driveList);
static void outCharacter(unsigned char character);
static unsigned char    getCurrentInstalledDrives(void);
static unsigned int getCurrentInt13CodeSegment(void);
static unsigned int getCurrentInt13InstructionPointer(void);
static void setCurrentInstalledDrives(unsigned char drive);
static void setCurrentInt13InstructionPointer(unsigned short value);
static void setCurrentInt13CodeSegment(unsigned short value);
static unsigned short   getStackSegment(void);
static unsigned short   getCodeSegment(void);
static void int13EntryPoint (void);
static int hookChannel(MV_U8 channel);
static unsigned short mvSataReadConfigWord(unsigned short offset);
static void oldInt13VectorCode(void);
static void initializeDriver(void);
static void activateA20Gate(void);
static void Ata2HostString(MV_U16 *source, MV_U16 *target, MV_U16 wordsCount);
static unsigned long readRegisterProtectedMode (register unsigned long offset);
static void writeRegisterProtectedMode (unsigned long offset, unsigned long value);
static void writeRegisterIoBar (unsigned short base_, unsigned long offset_,
                                unsigned long value_);
static unsigned long readRegisterIoBar (unsigned short base_,
                                        unsigned long offset_);


/* BCV Entry points */
void bcvEntryPoint_ch0(void);
void bcvEntryPoint_ch1(void);
void bcvEntryPoint_ch2(void);
void bcvEntryPoint_ch3(void);
void bcvEntryPoint_ch4(void);
void bcvEntryPoint_ch5(void);
void bcvEntryPoint_ch6(void);
void bcvEntryPoint_ch7(void);



typedef union
{
    unsigned short val;
    struct
    {
        unsigned char  lval;
        unsigned char  hval;
    } parts;
} reg16;

#define MV_EDMA_REQUEST_COMMANDS_NUM            11

typedef struct mvDmaRequestQueueEntry
{
    /* Fields set by CORE driver */
    volatile MV_U32       prdLowAddr;
    volatile MV_U32       prdHighAddr;
    volatile MV_U16       controlFlags;
    volatile MV_U16       command[MV_EDMA_REQUEST_COMMANDS_NUM];

} MV_DMA_REQUEST_QUEUE_ENTRY;


/*******************************************************************************
* main - Option ROM entry point
*
*
* DESCRIPTION:
*   Option ROM initializatoin function.
*   BIOS far calls this function for initializtion.
*
* INPUT:
*   ES:DI - PNP header (only if PNP BIOS)
*
* RETURN:
*   AX = 0x130 (indication IPL capable device)
*
* COMMENTS:
*   This function must be first function in main.c - it's compiled address
*   must be 0x30a (using Open Watcom compiler).
*
*******************************************************************************/
void main(void)
{
    _asm {
        mov cs:mvSataBusNum, ah
        mov cs:mvSataDeviceNum, al
        pop di
        pop si
        pop dx
        pop cx
        pop bx
        pop bp


        pusha
        push    fs
        push    gs
        push    es
        push    ds

        push    cs
        pop     ds
        mov     oldStack, sp
        mov     oldStackSegment,ss

        ; Change the stack segment and pointer
        push    cs
        pop     ss
        lea     sp, stack
        add     sp, stackSize - 0x10

        mov     pnpCheckSegment, es
        mov     pnpCheckOffset, di
        mov     pnpCheckResult, 0

        call initializeDriver

        cmp pnpCheckResult, 1  ; Check ifthis is a PNP BIOS
        je initOut          ;If so, BCV entries will initialize the adapter


        mov al, 0
        call hookChannel

        mov al, 1
        call hookChannel

        mov al, 2
        call hookChannel

        mov al, 3
        call hookChannel

        mov al, 4
        call hookChannel

        mov al, 5
        call hookChannel

        mov al, 6
        call hookChannel

        mov al, 7
        call hookChannel

        initOut :
        mov sp, oldStack
        mov ss, oldStackSegment
        pop ds
        pop es
        pop gs
        pop fs
        popa
        mov ax, 0x0130 ; Indicates that IPL Device supports int 13 found
        retf

    }
}

/*******************************************************************************
* bcvEntryPoint_chX - 8 Functions for BCV entry point
*
*
* DESCRIPTION:
*   8 functions needed for BCV entry point (PNP BIOS). Each function for a 
*   specific SATA channel.
*
* INPUT:
*   ES:DI - PNP header (only if PNP BIOS)
*
* RETURN:
*   AX = 0x130 (indication IPL capable device)
*
*******************************************************************************/
#define bcvEntryPointGenerate(x) void bcvEntryPoint_ch##x## (void)          \
{                                                                           \
_asm {      push    ds                              }                       \
_asm {      push    cs                              }                       \
_asm {      pop     ds                              }                       \
_asm {      mov     oldStackSegment, ss             }                       \
_asm {      push    cs                              }                       \
_asm {      pop     ss                              }                       \
_asm {      mov     oldStack, sp                    }                       \
_asm {      lea     sp, stack                       }                       \
_asm {      add     sp, stackSize-0x10              }                       \
_asm {      and     sp, 0xfff0                      }                       \
_asm {      pusha                                   }                       \
_asm {      push    fs                              }                       \
_asm {      push    gs                              }                       \
_asm {      push    es                              }                       \
_asm {      mov     bcvEsDataSegment, es            }                       \
_asm {      mov     bcvDiPointer, di                }                       \
    hookChannel(##x##);                                                     \
_asm {      pop     es                              }                       \
_asm {      pop     gs                              }                       \
_asm {      pop     fs                              }                       \
_asm {      popa                                    }                       \
_asm {      mov sp, oldStack                        }                       \
_asm {      mov ss, oldStackSegment                 }                       \
_asm {      pop ds                                  }                       \
_asm {      mov ax, 0x0130                          }                       \
_asm {      retf                                    }                       \
}

bcvEntryPointGenerate (0);
bcvEntryPointGenerate(1);
bcvEntryPointGenerate(2);
bcvEntryPointGenerate(3);
bcvEntryPointGenerate(4);
bcvEntryPointGenerate(5);
bcvEntryPointGenerate(6);
bcvEntryPointGenerate(7);



static MV_BOOLEAN
StartPM(PHW_ADAPTER_DESCRIPTION HwDeviceExtension,
        MV_U8 channelIndex, MV_BOOLEAN *driveList)
{
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_U8               PMPort;
    MV_SATA_PM_DEVICE_INFO PMInfo;
    MV_U8               temp;

    for(temp = 0 ; temp < MV_SATA_PM_MAX_PORTS ; temp ++)
    {
        driveList[temp] = MV_FALSE;
    }

    if(mvGetPMDeviceInfo(pSataAdapter, channelIndex, &PMInfo) == MV_FALSE)
    {
        MV_ERROR("MV IAL[%d %d]: Failed to get PortMultiplier Info\n",
                 pSataAdapter->adapterId, channelIndex);
        return MV_FALSE;
    }
    for(PMPort = 0; PMPort < PMInfo.numberOfPorts; PMPort++)
    {
        MV_U32  SStatus;
        /* 
         * Skip PMPort #0 - this should be already enabled due to legacy mode
         * transition of PM
         */

        if(PMPort > 0)
        {
            if(mvPMDevEnableStaggeredSpinUp(pSataAdapter, channelIndex, PMPort) ==
               MV_FALSE)
            {
                MV_ERROR("MV IAL[%d %d %d]: EnableStaggeredSpinUp Failed\n",
                         pSataAdapter->adapterId, channelIndex, PMPort);
                if(mvStorageDevATASoftResetDevice(pSataAdapter, channelIndex,
                                                  MV_SATA_PM_CONTROL_PORT, NULL) == MV_FALSE)
                {
                    MV_ERROR("MV IAL [%d %d]: failed to Soft Reset PM control port\n"
                             , pSataAdapter->adapterId, channelIndex);
                }
                continue;
            }
        }
        if(mvPMDevReadReg(pSataAdapter, channelIndex, PMPort,
                          MV_SATA_PSCR_SSTATUS_REG_NUM, &SStatus, NULL) ==
           MV_FALSE)
        {
            MV_ERROR("MV IAL[%d %d %d]: mvPMDevReadReg Failed\n",
                     pSataAdapter->adapterId, channelIndex, PMPort);
            if(mvStorageDevATASoftResetDevice(pSataAdapter, channelIndex,
                                              MV_SATA_PM_CONTROL_PORT, NULL) == MV_FALSE)
            {
                MV_ERROR("MV IAL [%d %d]: failed to Soft Reset PM control port\n"
                         , pSataAdapter->adapterId, channelIndex);
            }
            continue;
        }
        MV_PRINTD ("MV IAL[%d %d %x]: S-Status: 0x%x\n", pSataAdapter->adapterId,
                   channelIndex, PMPort, SStatus);
        if((SStatus & 0xf) != 3)
        {
            MV_PRINTD("MV IAL[%d %d %d]: No device found\n",
                      pSataAdapter->adapterId, channelIndex, PMPort);
            continue;
        }
        /* Disk is connected to PMPort device port on port multiplier */
        driveList[PMPort] = MV_TRUE;
    }
    return MV_TRUE;
}


/*******************************************************************************
* waitForKeystroke - Wait for a key pressed by user
*
*
* DESCRIPTION:
*   Utilizes BIOS 8 interrupt 16h, this function waits for a key pressed
*   by the user
*
* INPUT:
*   None
*
* RETURN:
*   None
*
*******************************************************************************/

void waitForKeystroke (void)
{
    _asm
    {
        pusha
        mov ah, 0x0
        waitForKeystrok_:
        int 0x16
        cmp al, 0
        je  waitForKeystrok_
        popa
        nop
        nop

    }
}

/*******************************************************************************
* readSegmentOffset8bit - Read 8bit from pointer indicated by segment:offset
*
*
* DESCRIPTION:
*   Read 16bit from pointer indicated by segment:offset
*
* INPUT:
*   segment_    -   Segment of requested data
*   offset_     -   Offset within the segment of the request data
*
* RETURN:
*   8bit of data that is requested
*
*******************************************************************************/
unsigned char readSegmentOffset8bit (unsigned short segment_, unsigned short offset_)
{
    unsigned char value = 0;
    _asm
    {
        pusha
        push fs
        mov di, segment_
        mov fs,di
        mov di, offset_
        mov cl, fs:[di]
        mov value, cl
        pop fs
        popa

    }
    return value;
}

/*******************************************************************************
* writeSegmentOffset8bit - Writes 8bit to a 8bit memory indicated by segment:offset
*
*
* DESCRIPTION:
*   Read 16bit from pointer indicated by segment:offset
*
* INPUT:
*   segment_    -   Segment of the memory to write to
*   offset_     -   Offset within the segment of the request data
*   value_      -   Value of the data to be written
*
* RETURN:
*   None
*
*******************************************************************************/
void writeSegmentOffset8bit (unsigned short segment_, unsigned short offset_, unsigned char value_)
{
    _asm
    {
        pusha
        push fs
        mov di, segment_
        mov fs,di
        mov di, offset_
        mov cl, value_
        mov fs:[di], cl
        pop fs
        popa

    }
    return;
}

/*******************************************************************************
* readSegmentOffset16bit - Read 16bit from pointer indicated by segment:offset
*
*
* DESCRIPTION:
*   Read 16bit from pointer indicated by segment:offset
*
* INPUT:
*   segment_    -   Segment of requested data
*   offset_     -   Offset within the segment of the request data
*
* RETURN:
*   16bit of data that is requested
*
*******************************************************************************/
unsigned short readSegmentOffset16bit (unsigned short segment_, unsigned short offset_)
{
    unsigned short value = 0;
    _asm
    {
        pusha
        push fs
        mov di, segment_
        mov fs,di
        mov di, offset_
        mov cx, fs:[di]
        mov value, cx
        pop fs
        popa

    }
    return value;
}

/*******************************************************************************
* writeSegmentOffset16bit - Writes 16bit to a 16bit memory indicated by segment:offset
*
*
* DESCRIPTION:
*   Read 16bit from pointer indicated by segment:offset
*
* INPUT:
*   segment_    -   Segment of the memory to write to
*   offset_     -   Offset within the segment of the request data
*   value_      -   Value of the data to be written
*
* RETURN:
*   None
*
*******************************************************************************/
void writeSegmentOffset16bit (unsigned short segment_, unsigned short offset_, unsigned short value_)
{
    _asm
    {
        pusha
        push fs
        mov di, segment_
        mov fs,di
        mov di, offset_
        mov cx, value_
        mov fs:[di], cx
        pop fs
        popa

    }
    return;
}

/*******************************************************************************
* memcpySegmentOffset32bitForward - Memory copy within segment
*
*
* DESCRIPTION:
*   Copies memory from source to destination within same segment, each read
*   write of 32 bit. The copy process is done from the end of the buffer
*   to the start, due to overlapping issues.
*
* INPUT:
*   segment_    -   Segment of requested data
*   offset_     -   Offset within the segment of the request data
*
* RETURN:
*   16bit of data that is requested
*
*******************************************************************************/
void memcpySegmentOffset32bitForward (unsigned short segment_,
                                      unsigned short destinationOffset_,
                                      unsigned short sourceOffset_,
                                      unsigned short byteCount_)
{
    _asm
    {
        pusha
        push fs
        push ecx
        mov di, segment_
        mov fs,di
        mov di, destinationOffset_
        mov si, sourceOffset_
        mov ax, byteCount_
        add di, ax
        add si, ax

        copyLoop:
        sub di, 4
        sub si, 4
        mov ecx, fs:[si]
        mov fs:[di], ecx
        sub ax, 4
        jnz copyLoop
        pop ecx
        pop fs
        popa

    }
}


#define WRITE_ATA_COMMAND_REG(addr, data, reg, isLast)              \
do{                                                                 \
    *(addr) = MV_CPU_TO_LE16((((MV_U8)(data)) & 0xff) | ((reg) << 8) | (isLast));   \
} while(0)

/* Calculate the base address of the registers for a SATA channel */
MV_U32 edmaRegOffst[MV_SATA_CHANNELS_NUM] = {0x22000, 0x24000, 0x26000, 0x28000,
    0x32000, 0x34000, 0x36000, 0x38000};

#define getEdmaRegOffset(x) edmaRegOffst[(x)]

MV_U32 generate32Addr (MV_U16 segment, MV_U16 offset)
{
    MV_U32 addr = 0;
    addr = (((MV_U32)segment) << 4) & 0x000ffff0;
    addr = addr + (MV_U32) offset;
    return addr;
}

MV_U8 requestIO (MV_U8 channelIndex, MV_U8 PMPort,
                 MV_U32 lowLBAAddress, MV_U16 highLBAAddress, MV_U16 sectorCount,
                 MV_UDMA_TYPE readWrite, MV_U16 bufferSegment, MV_U16 bufferOffset)
{
    volatile MV_U16  *pCommand;
    MV_SATA_EDMA_PRD_ENTRY  *prdTable;
    MV_DMA_REQUEST_QUEUE_ENTRY *pReqEntry;
    MV_U32 edmaRegsOffset;
    MV_U32 reqQEdmaAddr;
    MV_U16 stackSegment = 0;
    MV_U16 savedDataSegment = 0;
    MV_U16 timerBase, timerChanges;
    MV_U8 ATAStatus;
    MV_BOOLEAN isEXT;
    MV_U32 edmaErrorCause;
    MV_U8   bufferAllignment;
    MV_U8   savedBufferAllignment[8]; /* Used for 8byte buffer allignment */
    MV_U8   temp;
    MV_U32  bufferIOAddr = generate32Addr(bufferSegment, bufferOffset);

    /*
     * The following allocation is for PRD table and an EDMA request 
     * queue entry. The request queue entry is also used for the
     * response queue completion entry.
     * The PRD table and the request queue must be 32 byte alligned,
     * so we need to allocate 32*3 for the allignment.
     */
    unsigned char buffer[32*3];
    edmaRegsOffset = getEdmaRegOffset(channelIndex);
    /* Allign PRD table to 32 bytes */
    prdTable = (MV_SATA_EDMA_PRD_ENTRY *)(((MV_U16)buffer & 0xfff0) + 0x10);
    /* Request queue entry is 32bytes off the PRD table */
    pReqEntry = (MV_DMA_REQUEST_QUEUE_ENTRY *)(((MV_U16)buffer & 0xfff0) + 0x30);
    bufferAllignment = (MV_U8) bufferIOAddr & (requestedAllignment - 1);
    /* If not 32 byte alligned then add another 16byte */
    _asm {
        mov stackSegment, ss
        mov savedDataSegment, ds
        mov ds, stackSegment
    }
    if((generate32Addr (stackSegment, (MV_U16)prdTable) & MV_BIT4))
    {
        prdTable = (MV_SATA_EDMA_PRD_ENTRY *)((MV_U16) prdTable + 0x10);
        pReqEntry = (MV_DMA_REQUEST_QUEUE_ENTRY *)((MV_U16) pReqEntry + 0x10);
    }

    /* Populate the PRD table */
    prdTable[0].highBaseAddr = 0;
    prdTable[0].lowBaseAddr = bufferIOAddr - (MV_U32)bufferAllignment;
    prdTable[0].byteCount = (MV_U16) (sectorCount * 512);
    prdTable[0].flags = MV_EDMA_PRD_EOT_FLAG;
    prdTable[0].reserved = 0;

    /* Save the data for allignment */

    for(temp = 0 ; temp < bufferAllignment ; temp ++)
    {
        savedBufferAllignment[temp] =
        readSegmentOffset8bit(bufferSegment,
                              (bufferOffset - bufferAllignment) + temp);
    }

    /* Populate the request queue entry */
    pReqEntry->prdHighAddr = 0;
    pReqEntry->prdLowAddr = generate32Addr (stackSegment, (MV_U16)prdTable);
    reqQEdmaAddr = generate32Addr (stackSegment, (MV_U16)pReqEntry);
    pCommand = &pReqEntry->command[0];

    if((lowLBAAddress + (MV_U32)sectorCount) & 0xf0000000)
    { /* Use LBA48 */
        isEXT = MV_TRUE;
    } else
    {
        isEXT = MV_FALSE;
    }

    /* Set the direction of the transaction (read/write) */
    pReqEntry->controlFlags = (readWrite == MV_UDMA_TYPE_READ) ? 0x1 : 0x0;
    pReqEntry->controlFlags |= ((((MV_U16)PMPort) & 0x000f) << 12);

    if(isEXT == MV_TRUE)
    {   /* READ/WRITE DMA EXT */
        WRITE_ATA_COMMAND_REG(pCommand++,
                              (sectorCount & 0xFF00) >> 8,
                              MV_EDMA_ATA_SECTOR_COUNT_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (sectorCount) & 0xFF,
                              MV_EDMA_ATA_SECTOR_COUNT_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (lowLBAAddress & 0xFF000000) 
                              >> 24,
                              MV_EDMA_ATA_LBA_LOW_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++, 
                              (lowLBAAddress) &0xFF,
                              MV_EDMA_ATA_LBA_LOW_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (highLBAAddress & 0xFF),
                              MV_EDMA_ATA_LBA_MID_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (lowLBAAddress & 0xFF00) >> 8,
                              MV_EDMA_ATA_LBA_MID_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (highLBAAddress & 0xFF00) >> 8,
                              MV_EDMA_ATA_LBA_HIGH_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (lowLBAAddress & 0xFF0000)
                              >> 16,
                              MV_EDMA_ATA_LBA_HIGH_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++, MV_BIT6, MV_EDMA_ATA_DEVICE_ADDR,
                              0);

        if(readWrite == MV_UDMA_TYPE_READ)
        {
            WRITE_ATA_COMMAND_REG(pCommand++, MV_ATA_COMMAND_READ_DMA_EXT,
                                  MV_EDMA_ATA_COMMAND_ADDR, MV_BIT15);
        } else
        {
            WRITE_ATA_COMMAND_REG(pCommand++, MV_ATA_COMMAND_WRITE_DMA_EXT,
                                  MV_EDMA_ATA_COMMAND_ADDR, MV_BIT15);
        }
    } else
    { /* READ/WRITE DMA */
        WRITE_ATA_COMMAND_REG(pCommand++,
                              (sectorCount) & 0xFF,
                              MV_EDMA_ATA_SECTOR_COUNT_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (lowLBAAddress) & 0xFF,
                              MV_EDMA_ATA_LBA_LOW_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (lowLBAAddress & 0xFF00) >> 8,
                              MV_EDMA_ATA_LBA_MID_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              (lowLBAAddress & 0xFF0000)
                              >> 16,
                              MV_EDMA_ATA_LBA_HIGH_ADDR, 0);

        WRITE_ATA_COMMAND_REG(pCommand++,
                              MV_BIT6 | (MV_U8)((lowLBAAddress & 
                                                 0xF000000) >> 24),
                              MV_EDMA_ATA_DEVICE_ADDR, 0);

        if(readWrite == MV_UDMA_TYPE_READ)
        {
            WRITE_ATA_COMMAND_REG(pCommand++, MV_ATA_COMMAND_READ_DMA,
                                  MV_EDMA_ATA_COMMAND_ADDR, MV_BIT15);
        } else
        {
            WRITE_ATA_COMMAND_REG(pCommand++, MV_ATA_COMMAND_WRITE_DMA,
                                  MV_EDMA_ATA_COMMAND_ADDR, MV_BIT15);
        }
    }
    _asm
    {
        mov ds, savedDataSegment
    }
    /* Zero out the registers of EDMA */
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_REQUEST_Q_BAH_REG_OFFSET,
                       0);
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_RESPONSE_Q_BAH_REG_OFFSET,
                       0);
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_INTERRUPT_ERROR_CAUSE_REG_OFFSET,
                       0);

    /* Set request / response queue to point to the single EDMA entry */
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_REQUEST_Q_INP_REG_OFFSET,
                       reqQEdmaAddr);
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_REQUEST_Q_OUTP_REG_OFFSET,
                       reqQEdmaAddr & MV_EDMA_REQUEST_Q_OUTP_MASK);
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_RESPONSE_Q_OUTP_REG_OFFSET,
                       reqQEdmaAddr);
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_RESPONSE_Q_INP_REG_OFFSET,
                       reqQEdmaAddr & MV_EDMA_RESPONSE_Q_INP_MASK);

    /* This will clear any pending interrupts on drive */
    ATAStatus = MV_REG_READ_BYTE(mvSataBarOffset,
                                 edmaRegsOffset + MV_ATA_DEVICE_STATUS_REG_OFFSET);

    /* Enable EDMA */
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_COMMAND_REG_OFFSET,
                       MV_EDMA_COMMAND_ENABLE_MASK);
    MV_REG_READ_DWORD(mvSataBarOffset,
                      edmaRegsOffset + MV_EDMA_COMMAND_REG_OFFSET);

    /*
     * Advance request queue in pointer to next entry, thus triggering EDMA
     * to start the IO.
     * Beware that when advancing to wrap aroud the IN pointer by EDMA 
     * request queue size.
     */

    /* Need a wrap around ? */
    if((reqQEdmaAddr & 0x000003ffUL) == 0x3e0UL)
    {
        MV_REG_WRITE_DWORD(mvSataBarOffset,
                           edmaRegsOffset + MV_EDMA_REQUEST_Q_INP_REG_OFFSET,
                           reqQEdmaAddr & 0xfffffc00);
    } else
    {
        MV_REG_WRITE_DWORD(mvSataBarOffset,
                           edmaRegsOffset + MV_EDMA_REQUEST_Q_INP_REG_OFFSET,
                           reqQEdmaAddr + 0x20UL);

    }
    /* Poll on response queue in pointer for completion of command */
    timerChanges = 0;
    timerBase = readSystemClockTimer();
    while(1)
    {
        MV_U16 timerCurrent;
        if((MV_REG_READ_DWORD(mvSataBarOffset, 
                              edmaRegsOffset +
                              MV_EDMA_RESPONSE_Q_INP_REG_OFFSET)
            & MV_EDMA_RESPONSE_Q_INP_MASK) !=
           (reqQEdmaAddr & MV_EDMA_RESPONSE_Q_INP_MASK))
        {
            break;
        }

        timerCurrent = readSystemClockTimer();
        /* 
         * For measuring the time, the driver is waiting for changes on the 
         * system timer and count these changes.
         */
        if(timerCurrent != timerBase)
        {
            timerChanges ++;
            timerBase = timerCurrent;
        }

        /* 
         * Check if reached 10 seconds timeout. The 18*10 is because the system
         * timer is ~18Hz, and the required timeout is 10seconds
         */

        if(timerChanges == (18*10))
        {
#ifdef DEBUG_IAL
            printHexSingle ("Timeout !!!",0);
#endif
            break;
        }
    }
    /* Disable EDMA */
    MV_REG_WRITE_DWORD(mvSataBarOffset,
                       edmaRegsOffset + MV_EDMA_COMMAND_REG_OFFSET,
                       MV_EDMA_COMMAND_DISABLE_MASK);

    MV_REG_READ_DWORD(mvSataBarOffset,
                      edmaRegsOffset + MV_EDMA_COMMAND_REG_OFFSET);

    /* Do allignment if necessary */
    if(bufferAllignment)
    {
        memcpySegmentOffset32bitForward (bufferSegment,
                                         bufferOffset,
                                         bufferOffset - bufferAllignment,
                                         (MV_U16) (sectorCount * 512));
        for(temp = 0 ; temp < bufferAllignment ; temp ++)
        {
            writeSegmentOffset8bit(bufferSegment,
                                   (bufferOffset - bufferAllignment) + temp,
                                   savedBufferAllignment[temp]);
        }


    }
    edmaErrorCause = MV_REG_READ_DWORD (mvSataBarOffset, edmaRegsOffset +
                                        MV_EDMA_INTERRUPT_ERROR_CAUSE_REG_OFFSET);
    ATAStatus = MV_REG_READ_BYTE(mvSataBarOffset,
                                 edmaRegsOffset + MV_ATA_DEVICE_STATUS_REG_OFFSET);

    /* Clear EDMA error cause register */
    MV_REG_WRITE_DWORD (mvSataBarOffset, edmaRegsOffset +
                        MV_EDMA_INTERRUPT_ERROR_CAUSE_REG_OFFSET, 0);
    /* Clear unit cause register */
    if(channelIndex >= 4)
    {
        MV_REG_WRITE_DWORD (mvSataBarOffset,
                            MV_SATAHC_1_REGS_BASE_OFFSET +
                            MV_SATAHC_INTERRUPT_CAUSE_REG_OFFSET, 0);
    } else
    {
        MV_REG_WRITE_DWORD (mvSataBarOffset,
                            MV_SATAHC_0_REGS_BASE_OFFSET +
                            MV_SATAHC_INTERRUPT_CAUSE_REG_OFFSET, 0);
    }
    /* Check if disk is busy */
    if(ATAStatus & MV_BIT7)
    {
#ifdef DEBUG_IAL
        printHexSingle ("Disk Busy",0);
#endif

        /* Return timeout ; disk busy */
        return 0x80;

    }
    /* Check if ATA error occured */
    if(ATAStatus & MV_BIT0)
    {
        MV_U8 ATAError;
        ATAError = MV_REG_READ_BYTE(mvSataBarOffset,
                                    edmaRegsOffset + MV_ATA_DEVICE_ERROR_REG_OFFSET);
#ifdef DEBUG_IAL
        printHexSingle ("Error reg",ATAError);
#endif
        if(ATAError & MV_BIT6)
        {
            if(readWrite == MV_UDMA_TYPE_READ)
            {
                /* Return bad sector when UDMA read command (Medium error) */
                return 0x2;
            } else
            {
                /* Return write fault when UDMA write command (Write Protect) */
                return 0xcc;
            }
        }
        if(ATAError & MV_BIT7)
        {
            /* Return CRC error */
            return 0x10;
        }
        if(ATAError & MV_BIT4)
        {
            /* Return Sector not found */
            return 0x4;
        }
        if(ATAError & MV_BIT2)
        {
            /*
             * This should be returned as command aborted, but no such error
             * in BIOS codes, so returning Undefined error.
             */
            return 0xbb;
        }
        /* Undefined error */
        return 0xbb;
    }
    /* Check if EDMA error occured */
    if(edmaErrorCause & 0x1fff)
    {
#ifdef DEBUG_IAL
        printHexSingle ("EDMA Err",edmaErrorCause);
#endif

        /* Error in the IO */
        return 0x20;
    }

    return 0; /* All sectors transfered */

}

/*******************************************************************************
* int13ExecuteRequest - Main interrupt 13h handler
*
*
* DESCRIPTION:
*   Main interrupt 13h handler ; gets requests through AH and executes and
*   updates results accordingly
*
* INPUT:
*   All CPU registers through stack
*
* RETURN:
*   None
*
*******************************************************************************/
void interrupt
int13ExecuteRequest(volatile unsigned short gs,
                    volatile unsigned short fs,
                    volatile unsigned short es,
                    volatile unsigned short ds,
                    volatile reg16 di,
                    volatile reg16 si,
                    volatile reg16 bp,
                    volatile reg16 sp,
                    volatile reg16 bx,
                    volatile reg16 dx,
                    volatile reg16 cx,
                    volatile reg16 ax,
                    volatile unsigned short ip,
                    volatile unsigned short cs,
                    volatile unsigned short flags, 
                    ...)
{
    MV_U8 channelIndex = 0, PMPort = 0;
    int counter;
    MV_BOOLEAN incrementSP = MV_FALSE;
#ifdef DEBUG_IAL
    MV_BOOLEAN printAll = MV_FALSE;
    /* Comment and uncomment the following for full debug messages or debug messages on error only */
#if defined (MV_LOG_DEBUG)
    printAll = MV_TRUE; 
#endif
#endif

    /* Set data segment to point to code segment (the driver's data segment) */
    _asm
    {
        nop
        nop
        nop
        nop
        push cs
        pop ds
    }
#ifdef DEBUG_IAL
    initializeSerial ();
    if(printAll == MV_TRUE)
    {
        printHexSingle ("++++",0);
        printHexSingle ("+ AX",ax);
        printHexSingle ("+ BX",bx);
        printHexSingle ("+ CX",cx);
        printHexSingle ("+ DX",dx);
        printHexSingle ("+ ES",es);
        printHexSingle ("+ DI",di);
        printHexSingle ("+ DS",ds);
        printHexSingle ("++++",0);
    }
#endif
    /* If adapter's offset bit 20 is '1', then A20 line should be enabled */
    if((mvSataBarOffset & 0x00100000) && (useIoBar == MV_FALSE))
    {
        unsigned long requestQueueHiBase = 0;
        /* Check if it is already enabled ? */
        if(!(mvSataReadConfigWord(0x44) & 0x1000))
        {
            activateA20Gate();
            mvSataWriteConfigWord(0x44, mvSataReadConfigWord(0x44)
                                  + (1<<9));
        }
        requestQueueHiBase = mvSataReadReg(&sataAdapter.mvSataAdapter, 0x22010);
        if((requestQueueHiBase & 0xfffffffe0) != 
           (mvSataBarOffset & 0xffffffe0))
        {
#ifdef DEBUG_IAL
            printHexSingle ("Enabling A20",0);
#endif
            activateA20Gate();
        }
        requestQueueHiBase = mvSataReadReg(&sataAdapter.mvSataAdapter, 0x22010);
        if((requestQueueHiBase & 0xfffffffe0) != 
           (mvSataBarOffset & 0xffffffe0))
        {
#ifdef DEBUG_IAL
            printHexSingle ("Signature err",0);
#endif
            flags |= 0x1;
            ax.parts.hval = 0x20;
            return;
        }

    }

    flags &= ~0x01;     /* preset the carry flag to indicate no errors */

    for(counter = 0 ; counter < MV_SATA_CHANNELS_NUM ; counter ++)
    {
        if(drivesNumbers[counter] == dx.parts.lval)
        {
            channelIndex = sataChannelNumber[counter];
            PMPort = portMultiplierPort[counter];
            /* Check if PMPort is specified for the channel */
            if(PMPort == 0xff)
            {
                /*
                 * If not specified then set it for PMPort 0. This is necessary
                 * in order to set the PMPort in EDMA request queue entry 
                 * to 0
                 */
                PMPort = 0;
            }
            break;
        }
    }
    if(counter == MV_SATA_CHANNELS_NUM)
    {
        flags |= 0x1;
        ax.parts.hval = 0xbb;
#ifdef DEBUG_IAL
        printHexSingle ("ERR-invld CH",counter);
#endif
        return ;
    }
    if(channelIndex >= sataAdapter.mvSataAdapter.numberOfChannels)
    {
        flags |= 0x1;
        ax.parts.hval = 0xbb;
#ifdef DEBUG_IAL
        printHexSingle ("ERR-invld CH",channelIndex);
#endif
        return ;
    }

    switch(ax.parts.hval)
    {
    case 0x01:          /* Read disk Status */
        {
            ax.parts.hval = 0x1;
            ax.parts.lval = (MV_U8)(readSegmentOffset16bit (0x0040, 0x0041));
            break;
        }
    case 0x02:          /* Read Sectors */
    case 0x03:          /* Write Sectors */
        {
            MV_U32 lowLBAAddress;
            unsigned long track, head;
            unsigned char sector;
            sector = (unsigned short) (cx.parts.lval & 0x3f);
            track = (((unsigned short)(cx.parts.lval & 0xc0) << 2) | cx.parts.hval);
            head = (unsigned short) dx.parts.hval;
            lowLBAAddress = 
            (unsigned long) track * HEADS * SECTORS +
            (unsigned long) head * SECTORS +
            (unsigned long) (sector - 1) ;

            /* Check if sector count is more than 128 or equals 0 */
            if((ax.parts.lval > 128) || (ax.parts.lval == 0))
            {
#ifdef DEBUG_IAL
                printHexSingle ("Sect Cnt Err",readSegmentOffset16bit(ds,si.val+2));
#endif
                flags |= 0x1;
                ax.parts.lval = 0xa;
                break;
            }
            if((lowLBAAddress + ax.parts.lval) > 
               sataAdapter.channelsInfo[channelIndex].diskSize[PMPort])
            {
#ifdef DEBUG_IAL
                printHexSingle ("LBA OOR Lo",(lowLBAAddress) & 0xffff);
                printHexSingle ("LBA OOR Hi",((lowLBAAddress)>>16) & 0xffff);
#endif
                flags |= 0x1;
                ax.parts.lval = 0x4;
                break;
            }


#ifdef DEBUG_IAL
            if(printAll == MV_TRUE)
            {
                printHexSingle ("LBA Low",lowLBAAddress);
                printHexSingle ("LBA Hi",lowLBAAddress >> 16);
                printHexSingle ("Sec Count", ax.parts.lval);
                printHexSingle ("Buff Seg",es);
                printHexSingle ("Buff off",bx.val);
            }
#endif
            /* 
             * If SP < 0x200, then probably when calling requestIO it's going
             * to overflow. Add 0x200 to stack pointer and bp and substitute
             * 0x20 from the stack segment. This is the same from the int13
             * handler point of view, but when going back with iret the 
             * stack pointer, bp and stack segment must be returned to it's
             * previous value.
             */

            if(sp.val < 0x200)
            {
                incrementSP = MV_TRUE;
                _asm {
                    push si
                    add sp, 0x200
                    add bp, 0x200
                    mov si, ss
                    sub si, 0x20
                    mov ss, si
                    pop si

                }
#ifdef DEBUG_IAL
                printHexSingle ("Stack WO",sp.val);
#endif
            }
            ax.parts.hval = requestIO (channelIndex, PMPort,
                                       lowLBAAddress, 0, 
                                       ax.parts.lval,
                                       (ax.parts.hval == 0x2) ? MV_UDMA_TYPE_READ : MV_UDMA_TYPE_WRITE,
                                       es, bx.val);
            /* Revert sp, bp and ss to it's previous value */
            if(incrementSP == MV_TRUE)
            {
                _asm {
                    push si
                    sub sp, 0x200
                    sub bp, 0x200
                    mov si, ss
                    add si, 0x20
                    mov ss, si
                    pop si

                }
#ifdef DEBUG_IAL
                printHexSingle ("StackB",sp.val);
#endif			
            }
            if(ax.parts.hval != 0x0)
            {
#ifdef DEBUG_IAL
                printAll = MV_TRUE;
                printHexSingle ("UDMA comp Err",ax.parts.hval);
#endif
                if(mvStorageDevATASoftResetDevice(&sataAdapter.mvSataAdapter,
                                                  channelIndex, PMPort, 0) ==
                   MV_FALSE)
                {
#ifdef DEBUG_IAL
                    printHexSingle ("Softrst Err",ax.parts.hval);
#endif
                }
                flags |= 0x1;
            }
            /* 
             * Revert the Request Queue hi base address to it's value that is
             * set in initializeDriver 
             */
            if((channelIndex == 0) && (useIoBar == MV_FALSE))
            {
                mvSataWriteReg(&sataAdapter.mvSataAdapter, 0x22010, mvSataBarOffset);
            }

            break;
        }
    case 0x00:          /* Reset disk system */
    case 0x0d:          /* Alternate Fixed Disk Reset */
        if(mvStorageDevATASoftResetDevice(&sataAdapter.mvSataAdapter,
                                          channelIndex, PMPort, 0) == MV_FALSE)
        {
            flags |= 0x1;
            ax.parts.hval = 0x5; /* fixed disk reset failed */
#ifdef DEBUG_IAL
            printHexSingle ("Softrst Err",ax.parts.hval);
#endif
        } else
        {
            ax.parts.hval = 0x0; /* All went good */
        }
        break;
    case 0x04:          /* Verify Sectors */
        {
            MV_U32 lowLBAAddress;
            unsigned long track, head;
            unsigned char sector;
            MV_BOOLEAN returnStatus;
            MV_U8 ATAStatus, ATAError;
            MV_U16 timerBase, timerChanges;

            sector = (unsigned short) (cx.parts.lval & 0x3f);
            track = (((unsigned short)(cx.parts.lval & 0xc0) << 2) | cx.parts.hval);
            head = (unsigned short) dx.parts.hval;
            lowLBAAddress = 
            (unsigned long) track * HEADS * SECTORS +
            (unsigned long) head * SECTORS +
            (unsigned long) (sector - 1) ;

            /* Check if sector count is more than 128 or equals 0 */
            if((ax.parts.lval > 128) || (ax.parts.lval == 0))
            {
#ifdef DEBUG_IAL
                printHexSingle ("Sect Cnt Err",readSegmentOffset16bit(ds,si.val+2));
#endif
                flags |= 0x1;
                ax.parts.lval = 0xa;
                break;
            }
            if((lowLBAAddress + ax.parts.lval) > 
               sataAdapter.channelsInfo[channelIndex].diskSize[PMPort])
            {
#ifdef DEBUG_IAL
                printHexSingle ("LBA OOR Lo",(lowLBAAddress) & 0xffff);
                printHexSingle ("LBA OOR Hi",((lowLBAAddress)>>16) & 0xffff);
#endif
                flags |= 0x1;
                ax.parts.lval = 0x4;
                break;
            }


#ifdef DEBUG_IAL
            if(printAll == MV_TRUE)
            {
                printHexSingle ("V LBA Low",lowLBAAddress);
                printHexSingle ("V LBA Hi",lowLBAAddress >> 16);
                printHexSingle ("V Sec Count", ax.parts.lval);
            }
#endif
            returnStatus = MV_TRUE;
            /*
             * Since in this function the LBA address translation from CHS
             * can't be LBA48, then send only LBA 28 command.
             */
            mvSataWriteReg (&sataAdapter.mvSataAdapter,
                            getEdmaRegOffset (channelIndex) + MV_ATA_DEVICE_FEATURES_REG_OFFSET,
                            0);

            mvSataWriteReg (&sataAdapter.mvSataAdapter,
                            getEdmaRegOffset (channelIndex) + MV_ATA_DEVICE_SECTOR_COUNT_REG_OFFSET,
                            ax.parts.lval);

            mvSataWriteReg (&sataAdapter.mvSataAdapter,
                            getEdmaRegOffset (channelIndex) + MV_ATA_DEVICE_LBA_LOW_REG_OFFSET,
                            (MV_U32)(lowLBAAddress & 0xffUL));

            mvSataWriteReg (&sataAdapter.mvSataAdapter,
                            getEdmaRegOffset (channelIndex) + MV_ATA_DEVICE_LBA_MID_REG_OFFSET,
                            (MV_U32)((lowLBAAddress & 0xff00UL) >> 8));

            mvSataWriteReg (&sataAdapter.mvSataAdapter,
                            getEdmaRegOffset (channelIndex) + MV_ATA_DEVICE_LBA_HIGH_REG_OFFSET,
                            (MV_U32)((lowLBAAddress & 0xff0000UL) >> 16));

            mvSataWriteReg (&sataAdapter.mvSataAdapter,
                            getEdmaRegOffset (channelIndex) + MV_ATA_DEVICE_LBA_HIGH_REG_OFFSET,
                            (MV_U32)((MV_BIT6) |
                                     ((lowLBAAddress & 0x0f000000UL) >> 24)));
            mvSataWriteReg (&sataAdapter.mvSataAdapter,
                            getEdmaRegOffset (channelIndex) + MV_ATA_DEVICE_COMMAND_REG_OFFSET,
                            MV_ATA_COMMAND_READ_VERIFY_SECTORS);
            /* Command is out - wait for 10 seconds for completion */
            timerChanges = 0;
            timerBase = readSystemClockTimer();
            while(1)
            {
                MV_U16 timerCurrent;
                timerCurrent = readSystemClockTimer();
                ATAStatus = mvSataReadReg (&sataAdapter.mvSataAdapter,
                                           getEdmaRegOffset (channelIndex) +
                                           MV_ATA_DEVICE_STATUS_REG_OFFSET);
                /* Is BSY bit '0' ? */
                if(!(ATAStatus & MV_BIT7))
                {
                    break;
                }
                /* 
                 * For measuring the time, the driver is waiting for changes on the 
                 * system timer and count these changes.
                 */
                if(timerCurrent != timerBase)
                {
                    timerChanges ++;
                    timerBase = timerCurrent;
                }

                /* 
                 * Check if reached 10 seconds timeout. The 18*10 is because the system
                 * timer is ~18Hz, and the required timeout is 10seconds
                 */

                if(timerChanges == (18*10))
                {
#ifdef DEBUG_IAL
                    printHexSingle ("Timeout !!!",0);
#endif
                    returnStatus = MV_FALSE;
                    break;
                }
            }


            /* If function returned MV_TRUE, check ATA Status */
            if(returnStatus == MV_TRUE)
            {
                /* Is err bit set ? */
                if(ATAStatus & MV_BIT0)
                {
                    ATAError = mvSataReadReg (&sataAdapter.mvSataAdapter,
                                              getEdmaRegOffset (channelIndex) +
                                              MV_ATA_DEVICE_ERROR_REG_OFFSET);

#ifdef DEBUG_IAL
                    printHexSingle ("VER ERR",ATAStatus);
                    printHexSingle ("VER ERR",ATAError);
#endif

                    /* Check UNC */
                    if(ATAStatus & MV_BIT6)
                    {
                        /* Medium error */
                        ax.parts.hval = 0x2;
                        flags |= 0x1;
                    } else
                    {
                        /* HDD Error */
                        ax.parts.hval = 0xbb;
                        flags |= 0x1;
                    }
                }
            } else
            {
#ifdef DEBUG_IAL
                printHexSingle ("TimeOut",ATAStatus);
#endif

                /* Return not ready / timeout */
                ax.parts.hval = 0x80;
                flags |= 0x1;
            }
            break;
        }
    case 0x05:          /* Format Cylinder */
    case 0x09:          /* Initialize Drive Parameters */
    case 0x0c:          /* Seek to Cylinder */
    case 0x11:          /* Recalibrate Drive */
    case 0x12:          /* Controller Ram Diagnostic */
    case 0x13:          /* Controller Driver Diagnositc */
    case 0x14:          /* Controller Internal Diagnostic */
    case 0x44:          /* Extended verify sectors */
    case 0x18:          /* Set Media Type For Format */
        {
            ax.parts.hval = 0x0;    /* Fake the command as if it has passed */
            break;
        }
    case 0x47:          /* Extended seek */
        {
            ax.parts.hval = 0;  /* lie and say it all worked */
            break;
        }
    case 0x08:          /* Read Drive Parameters */
        {
            ax.val = 0;
            cx.val = 0xfeff;
            cx.parts.lval = ((CYLINDER & 0x300) >> 2) | SECTORS;
            cx.parts.hval = (CYLINDER & 0xff);
            dx.parts.hval = HEADS - 1;
            dx.parts.lval = getCurrentInstalledDrives();

            break;
        }
    case 0x10:          /* Test for Drive Ready */
        {
            ax.parts.hval = 0;
            break;
        }
    case 0x15:          /* Read Fixed Disk Type */
        {
            cx.val = 0xf9;
            dx.val = 0xc97e;
            ax.val = 0x0300;
            break;
        }
        /* Extended commands */
    case 0x41:
        {
            ax.parts.hval = 0x21; /* Major revision number */
            bx.val = 0xaa55;
            cx.val = 0x5 ; /* Device access using the packet structure */
            break;
        }
    case 0x42:  /* Extended READ */
    case 0x43:  /* Extended WRITE */
        {
            unsigned short blknumLo, blknumHi;
            MV_U32 lowLBAAddress;
            unsigned short sectorCount;
            unsigned short segment, offset;
            /* Check that packet is at least 16 bytes */
            if(((unsigned char)readSegmentOffset16bit(ds,si.val)) < 0x10)
            {
#ifdef DEBUG_IAL
                printHexSingle ("Length err",readSegmentOffset16bit(ds,si.val));
                printAll = MV_TRUE;
#endif
                flags |= 0x1;
                ax.parts.lval = 0x1;
                break;
            }
            /* Meantime, support only 32 bits of address */
            blknumLo = readSegmentOffset16bit(ds,si.val+8);
            blknumHi = readSegmentOffset16bit(ds,si.val+10);
            lowLBAAddress = (MV_U32) ((((MV_U32)blknumHi) << 16) |
                                      (((MV_U32)blknumLo) & 0x0000ffff));

            sectorCount = (unsigned char) (readSegmentOffset16bit(ds,si.val+2) & 0xff);

            /* Check if sector count is more than 128 or equals 0 */
            if((sectorCount > 128) || (sectorCount == 0))
            {
#ifdef DEBUG_IAL
                printHexSingle ("Sect Cnt Err",readSegmentOffset16bit(ds,si.val+2));
#endif
                flags |= 0x1;
                ax.parts.lval = 0xa;
                break;
            }
            if((lowLBAAddress + sectorCount) > 
               sataAdapter.channelsInfo[channelIndex].diskSize[PMPort])
            {
#ifdef DEBUG_IAL
                printHexSingle ("LBA OOR Lo",(lowLBAAddress) & 0xffff);
                printHexSingle ("LBA OOR Hi",((lowLBAAddress)>>16) & 0xffff);
#endif
                flags |= 0x1;
                ax.parts.lval = 0x4;
                break;
            }
            segment = readSegmentOffset16bit(ds,si.val + 6);
            offset = readSegmentOffset16bit(ds,si.val + 4);
#ifdef DEBUG_IAL
            if(printAll == MV_TRUE)
            {
                printHexSingle ("LBA Low",blknumLo);
                printHexSingle ("LBA Hi",blknumHi);
                printHexSingle ("LBA Low",lowLBAAddress);
                printHexSingle ("LBA Hi",lowLBAAddress>>16);
                printHexSingle ("Sec Count",sectorCount);
                printHexSingle ("Buff seg",segment);
                printHexSingle ("Buff off",offset);
                printHexSingle ("Buff Lo",generate32Addr (segment, offset));
                printHexSingle ("Buff Hi",generate32Addr (segment, offset)>>16);
                printHexSingle ("R/W",ax.parts.hval);
            }
#endif
            /* 
             * If SP < 0x200, then probably when calling requestIO it's going
             * to overflow. Add 0x200 to stack pointer and bp and substitute
             * 0x20 from the stack segment. This is the same from the int13
             * handler point of view, but when going back with iret the 
             * stack pointer, bp and stack segment must be returned to it's
             * previous value.
             */

            if(sp.val < 0x200)
            {
                incrementSP = MV_TRUE;
                _asm{
                    push si
                    add sp, 0x200
                    add bp, 0x200
                    mov si, ss
                    sub si, 0x20
                    mov ss, si
                    pop si
                }
#ifdef DEBUG_IAL
                printHexSingle ("Stack WO",sp.val);
#endif



            }

            ax.parts.hval= requestIO (channelIndex, PMPort,
                                      lowLBAAddress, 0, 
                                      sectorCount,
                                      (ax.parts.hval == 0x42) ? MV_UDMA_TYPE_READ : MV_UDMA_TYPE_WRITE,
                                      segment, offset);
            /* Revert sp, bp and ss to it's previous value */
            if(incrementSP == MV_TRUE)
            {
                _asm{
                    push si
                    sub sp, 0x200
                    sub bp, 0x200
                    mov si, ss
                    add si, 0x20
                    mov ss, si
                    pop si
                }
#ifdef DEBUG_IAL
                printHexSingle ("StackB",sp.val);
#endif			
            }

            if(ax.parts.hval != 0x0)
            {
#ifdef DEBUG_IAL
                printHexSingle ("UDMA comp Err",ax.parts.hval);
                printAll = MV_TRUE;
#endif
                if(mvStorageDevATASoftResetDevice(&sataAdapter.mvSataAdapter,
                                                  channelIndex, PMPort, 0) ==
                   MV_FALSE)
                {
#ifdef DEBUG_IAL
                    printHexSingle ("Softrst Err",ax.parts.hval);
#endif



                }
                flags |= 0x1;
            }
            /* 
             * Revert the Request Queue hi base address to it's value that is
             * set in initializeDriver 
             */
            if((channelIndex == 0) && (useIoBar == MV_FALSE))
            {
                mvSataWriteReg(&sataAdapter.mvSataAdapter, 0x22010, mvSataBarOffset);
            }
            break;
        }
    case 0x45:          /* Lock / unlock media */
    case 0x46:          /* Eject removable media */
        {
            /* Return media is not removable */
            flags |= 0x1;
            ax.parts.hval = 0xb2;
            break;
        }
    case 0x48:  /* Get drive parameters extended */
        {
            unsigned short temp;
            unsigned long size_of_disk;
            if(readSegmentOffset16bit(ds, si.val) < 26)
            {
                flags |= 0x1;
                ax.parts.hval = 1;
                break;
            }
            ax.parts.hval = 0;
            if(readSegmentOffset16bit(ds, si.val) > 30)
            {
                writeSegmentOffset16bit(ds, si.val, 30);
            }
            if((readSegmentOffset16bit(ds, si.val) > 26) &&
               (readSegmentOffset16bit(ds, si.val) < 30))
            {
                writeSegmentOffset16bit(ds, si.val, 26);
            }
            /* WORD 2 - geometry in fields 4..15 is invalid */
            writeSegmentOffset16bit(ds, si.val + 2, 0);
            /* Reset fields 4..15 */
            for(temp = 4 ; temp < 16 ; temp += 2)
            {
                writeSegmentOffset16bit(ds, si.val + temp, 0);
            }

            size_of_disk = sataAdapter.channelsInfo[channelIndex].diskSize[PMPort];
            writeSegmentOffset16bit(ds, si.val+16, (unsigned short)(size_of_disk & 0xffff));
            writeSegmentOffset16bit(ds, si.val+18, (unsigned short)((size_of_disk >> 16) & 0xffff));
            writeSegmentOffset16bit(ds, si.val+20, 0);
            writeSegmentOffset16bit(ds, si.val+22, 0);
            writeSegmentOffset16bit(ds, si.val+24, 512);    /* 512 byte per sector */

            break;
        }
    case 0x06:          /* Format Bad Track */
    case 0x07:          /* Format Drive */
    case 0x0a:          /* Read Long Sectors */
    case 0x0b:          /* Write Long Sectors */
    case 0x0e:          /* Diagnostics 1: Read Test Buffer */
    case 0x0f:          /* Diagnostics 2: Write Test Buffer */
    default:
        {
            /* function not supported */
#ifdef DEBUG_IAL
            printAll = MV_TRUE;
            printHexSingle("UNSP",ax.parts.hval);
            printHexSingle ("Press any key" ,0);
#endif
            ax.parts.hval = 0x01;
            flags |= 1;
            break;
        }
    }

#ifdef DEBUG_IAL
    if(printAll == MV_TRUE)
    {
        printHexSingle ("----",0);
        printHexSingle ("- AX",ax);
        printHexSingle ("- BX",bx);
        printHexSingle ("- CX",cx);
        printHexSingle ("- DX",dx);
        printHexSingle ("- ES",es);
        printHexSingle ("- DI",di);
        printHexSingle ("- flags",flags);
        printHexSingle ("----",0);
    }
#endif




}


/*******************************************************************************
* int13EntryPoint - Interrupt 13h entry point
*
*
* DESCRIPTION:
*   Checks if the requested interrupt 13h is for the adapter (by checking DL).
*   If so, it calls int13ExecuteRequest ; otherwise it calls the next
*   interrupt 13h handler in the chain.
*
* INPUT:
*   All CPU registers
*
* RETURN:
*   None
*
*******************************************************************************/

void int13EntryPoint (void)
{
    _asm{
        nop
        nop
        nop
        pushf
        cmp byte dl, cs:drivesNumbers + 0
        je  int13ExectueRequestWrapper
        cmp byte dl, cs:drivesNumbers + 1
        je  int13ExectueRequestWrapper
        cmp byte dl, cs:drivesNumbers + 2
        je  int13ExectueRequestWrapper
        cmp byte dl, cs:drivesNumbers + 3
        je  int13ExectueRequestWrapper
        cmp byte dl, cs:drivesNumbers + 4
        je  int13ExectueRequestWrapper
        cmp byte dl, cs:drivesNumbers + 5
        je  int13ExectueRequestWrapper
        cmp byte dl, cs:drivesNumbers + 6
        je  int13ExectueRequestWrapper
        cmp byte dl, cs:drivesNumbers + 7
        je  int13ExectueRequestWrapper

        popf
        jmp oldInt13VectorCode

        int13ExectueRequestWrapper :        
        popf
        jmp     int13ExecuteRequest






    }





}

/*******************************************************************************
* oldInt13VectorCode - Old Interrupt 13h entry point
*
*
* DESCRIPTION:
*   This function is modified in runtime (self modifying code) and has a far
*   call to the previous interrupt 13h handler before this driver hooks in
*   the interrupt 13h for the SATA channels.
*
* INPUT:
*   None
*
* RETURN:
*   None
*
*******************************************************************************/
void oldInt13VectorCode(void)
{
    printf("NULL\n");
    return;
}


/*******************************************************************************
* printf - Prints to the monitor a formatted string
*
*
* DESCRIPTION:
*   Gets a formatted string, while utilizing sprintf it prints to the monitor.
*
* INPUT:
*   None
*
* RETURN:
*   None
*
*******************************************************************************/
void printf(const char *fmt, ...)
{
    va_list args;
    int i;
    char prntBuff[100];
    va_start(args, fmt);
    i = vsprintf (prntBuff,fmt,args);
    va_end(args);
    printString(prntBuff);
    return;
}

unsigned char intToChr (unsigned char value)
{
    if((value <= 9))
    {
        return(value + '0');
    }
    if((value >= 10) && (value <= 15))
    {
        return(value + 'a'-10);
    }
    return 0;
}
#ifdef PRNTHEXTSINGLE
void printHexSingle(char *fmt, unsigned short value_)
{
    int i = 0;
    char prntBuff[20];
    unsigned short value = value_;
    for(i = 0 ; fmt[i] ; i++)
    {
        prntBuff[i] = fmt[i];
    }
    prntBuff[i] = ' ';
    prntBuff[i+1] = 0;

#ifdef DEBUG_IAL
    printStringSerial (prntBuff);
#else
    printString (prntBuff);
#endif
    for(i = 0 ; i < 4 ; i ++)
    {
        prntBuff[3-i] = intToChr(value & 0xf);
        value = value >> 4;
    }
    prntBuff[4] = '\n';
    prntBuff[5] = 0;
#ifdef DEBUG_IAL
    printStringSerial (prntBuff);
#else
    printString (prntBuff);
#endif





}
#endif

/*******************************************************************************
* mvLogMsg - Log messages printing function
*
*
* DESCRIPTION:
*   This function is part of the CORE driver OS dependent function. It gets
*   a level of the log message and accordingly decides wether to print it out
*   or not.
*
* INPUT:
*   level   -   Level of the requested message to be logged
*   fmt     -   Pointer to formatted string
*   ...     -   Parameters the are used to properly printing the log message
*
* RETURN:
*   None
*
*******************************************************************************/
#ifdef MV_DEBUG_LOG
int mvLogMsg(MV_U8 level, MV_CHAR_PTR fmt, ...)
{
    va_list args;
    int i;
    char prntBuff[150];
    if(level & 0x40)
    {
        va_start(args, fmt);
        i = vsprintf (prntBuff,fmt,args);
        va_end(args);
        printString (prntBuff);
    }
    return 0;
}
#endif



void resizeToZero(void)
{

    printf("Resizing ROM image to 0\n");
    /* Resize ROM image size to 0 - no need to recalculate checksum */
    * (unsigned char *) 0x2 = 0;
}

/*******************************************************************************
* initializeDriver - Initializes BIOS Driver parameters
*
*
* DESCRIPTION:
*   This function modifed the GDT table (global descriptor table) to be suitable
*   for accessing 88SX50xx internal registers. It then initlizes 88SX50xx part
*   and initializes all storage devices connected to it.
*
* INPUT:
*   Nonte
*
* RETURN:
*   None
*
*******************************************************************************/
void initializeDriver(void)
{
    unsigned long temp, *temp1;
    unsigned int dataSegment;
    unsigned short buff[3];

    if(alreadyInitializedChannelsFlag)
    {
        return;
    }

    mvSataDeviceId = mvSataReadConfigWord(0x2);
    mvSataRevisionId = (unsigned char)(mvSataReadConfigWord(0x8) & 0xff);
    if((mvSataDeviceId != 0x5080) &&
       (mvSataDeviceId != 0x5081) &&
       (mvSataDeviceId != 0x5040) &&
       (mvSataDeviceId != 0x5041) &&
       (mvSataDeviceId != 0x6081) &&
       (mvSataDeviceId != 0x6041))
    {
        printf("Marvell Serial ATA adapter - Wrong Device ID %x\n",mvSataDeviceId);
        return;
    }

    if((mvSataDeviceId == 0x6081) ||
       (mvSataDeviceId == 0x6041))
    {
        useIoBar = MV_TRUE;
    } else
    {
        useIoBar = MV_FALSE;
        /* Update GDT table with driver's segment location */
        temp = 0;
        dataSegment = 0;
        _asm{mov dataSegment, ds};
        temp = dataSegment;
        temp = temp << 4;

        temp += (unsigned long) GDT;
        temp1 = (unsigned long *)(((unsigned int) GDT) + 2);
        * (unsigned long *) temp1 = temp;
    }

#ifdef DEBUG_IAL
    printf("\nCode segment of BIOS extension driver is %x\n",getCodeSegment());
#endif
    printf("\nMarvell Serial ATA 88SX%04x Adapter - Driver Version %d.%d.%d.%d\n",
           (unsigned long)mvSataDeviceId,
           (* (unsigned char *) 0x40) / 10,
           (* (unsigned char *) 0x40) % 10,
           (* (unsigned char *) 0x41) / 10,
           (* (unsigned char *) 0x41) % 10);
    alreadyInitializedChannelsFlag = 1;
    if(initlizeMvSataAdapter())
    {
        printf ("Error finding / initializing adapter\n");
        resizeToZero ();
        mvMicroSecondsDelay(0, 1000000); /* One second */
        return;
    }
#ifdef DEBUG_IAL
    printf("BAR offset is %lx\n",mvSataBarOffset);
#endif

    /* If 88SX50xx BAR offset bit 20 is '1', then A20 line should be enabled */
    if((mvSataBarOffset & 0x00100000) && (useIoBar == MV_FALSE))
    {
        activateA20Gate();
    }
    if(findAndInitChannels() == MV_FALSE)
    {
        printf ("Error finding or initializing Serial ATA Channels\n");
        resizeToZero ();
        mvMicroSecondsDelay(0, 1000000); /* One second */
        return;
    }
    /* Fix for 88SX50xx FEr PCI#10*/
    /* Fix for 88SX60xx FEr PCI#3*/
    if((sataAdapter.mvSataAdapter.implement50XXB0Workarounds == MV_TRUE) ||
       (sataAdapter.mvSataAdapter.implement60X1A1Workarounds == MV_TRUE))
    {
        requestedAllignment = 8; /* Needed allignment is 8 byte for buffer */
    } else
    {
        requestedAllignment = 2; /* Needed allignment for 2 bytes only */
    }

    /* Check if the BIOS is plug & play BIOS */
    buff[0] = readSegmentOffset16bit(pnpCheckSegment, pnpCheckOffset);
    buff[1] = readSegmentOffset16bit(pnpCheckSegment, pnpCheckOffset + 2);
    buff[2] = 0;
    if(!strcmp ((const char *)buff, "$PnP"))
    {
        pnpCheckResult = 1;
#ifdef DEBUG_IAL
        printf("Detected PnP BIOS\n");
#endif



    } else
    {
        pnpCheckResult = 0;
#ifdef DEBUG_IAL
        printf("Detected non-PnP BIOS\n");
#endif



    }
    mvMicroSecondsDelay(0, 1000000); /* One second */
#ifdef DEBUG_IAL
    waitForKeystroke();
#endif

    if(useIoBar == MV_FALSE)
    {
        /* 
         * Write BAR0 offset to Channel #0 EDMA request queue high base address.
         * This is used as a signature for the int13 runtime routine to check if
         * the A20 GATE is needed to be activated or not.
         */
        mvSataWriteReg(&sataAdapter.mvSataAdapter, 0x22010, mvSataBarOffset);
    }
}

/*******************************************************************************
* hookChannel - Hooks storage device to int13
*
*
* DESCRIPTION:
*   This function hooks a specific storage device connected to serial ATA
*   channel by reserving a device number in the 0040:0075 BDA address and
*   saving then hooking old interrupt 13.
*
* INPUT:
*   channel -   Serial ATA channel number required to be hooked
*
* RETURN:
*   None
*
*******************************************************************************/
int hookChannel(MV_U8 driveIndex)
{
    unsigned char drive;
    unsigned char channelIndex;

    channelIndex = sataChannelNumber[driveIndex];
    /* Check if there is actually a hard drive connected */
    if(channelIndex == 0xff)
    {
        return 0;
    }
    if(!sataAdapter.mvSataAdapter.sataChannel[channelIndex])
    {
        return 0;
    }
    if(drivesNumbers[driveIndex] != 0xff)
    {
        printf("Error - called by BIOS to hook channel %d twice !!!\n",
               channelIndex);
        return 0;
    }

    if(channelIndex >= sataAdapter.mvSataAdapter.numberOfChannels)
    {
        return 0;
    }
    drive = getCurrentInstalledDrives();

    /*
     * If we are first to install, then copy INT13h verctor over to INT 40h
     * so the floopy services are handled properly
     */

    drive = drive & 0x7f;
    if(drive == 0)
    {
#ifdef DEBUG_IAL
        printf("Copying int 13h vector to int 40h\n");
        mvMicroSecondsDelay(0, 1000000);
#endif
        writeSegmentOffset16bit (0x0, 0x40*4, readSegmentOffset16bit(0x0,0x13*4));
        writeSegmentOffset16bit (0x0, (0x40*4)+2, readSegmentOffset16bit(0x0,(0x13*4)+2));
    }

    drivesNumbers[driveIndex] = drive + 0x80;

    setCurrentInstalledDrives(drive + 1);
    if(alreadySavedInt13Vector == 0)
    {
        savedInt13CodeSegment = getCurrentInt13CodeSegment();
        savedInt13InstructionPointer = getCurrentInt13InstructionPointer();
        /* Setup our int13 handling routine */
        setCurrentInt13CodeSegment(getCodeSegment());
        setCurrentInt13InstructionPointer((unsigned short) &int13EntryPoint + 0x8);
        _asm{
            push ax

            mov al, 0xea
            mov cs:oldInt13VectorCode, al
            mov ax, savedInt13InstructionPointer
            mov cs:oldInt13VectorCode + 1, ax
            mov ax, savedInt13CodeSegment
            mov cs:oldInt13VectorCode + 3, ax

            pop ax





        }
#ifdef DEBUG_IAL
        printf("Old int #13 - %04x : %04x\n",(unsigned long)savedInt13CodeSegment,(unsigned long)savedInt13InstructionPointer);
        printf("New int #13 - %04x : %04x\n",(unsigned long)getCurrentInt13CodeSegment(),(unsigned long)getCurrentInt13InstructionPointer());
#endif        




    }
#ifdef DEBUG_IAL
    if(portMultiplierPort[driveIndex] != 0xff)
    {
        printf("%d : Hooking drive %x to channel %d PMPort %d (adapter bus %d device %d)\n",
               driveIndex, (unsigned long)(drive+0x80), channelIndex,
               portMultiplierPort[driveIndex], mvSataBusNum,mvSataDeviceNum>>3);
    } else
    {
        printf("%d : Hooking drive %x to channel %d (adapter bus %d device %d)\n",
               driveIndex, (unsigned long)(drive+0x80), channelIndex,
               mvSataBusNum,mvSataDeviceNum>>3);
    }
    waitForKeystroke();
#endif
    alreadySavedInt13Vector = 1;
    return 1;
}


/*******************************************************************************
* readSystemClockTimer - 
*
*
* DESCRIPTION:
*   Returns the low 16bit of the system timer
*
* INPUT:
*   None
*
* RETURN:
*   Low 16bit of the system timer
*
*******************************************************************************/
static MV_U16 readSystemClockTimer(void)
{
    MV_U16 returnValue = 0;
    _asm{
        mov ah, 0x0
        int 0x1a
        mov returnValue, dx
    }
    return returnValue;
}

/*******************************************************************************
* mvMicroSecondsDelay - Delays for a required uSeconds
*
*
* DESCRIPTION:
*   Delays for a required uSeconds. This function is part of the CORE driver
*   OS dependent functions.
*
* INPUT:
*   x   -   Ammount of uSeconds to delay
*
* RETURN:
*   None
*
*******************************************************************************/
void mvMicroSecondsDelay(MV_VOID_PTR not_used, unsigned long x)
{
    MV_U32 numOfChanges;
    MV_U16 oldValue, newValue;
    /* Using system timer, the tick is ~18.206, which is 54.926mSec */
    numOfChanges = (x/54926UL) + 1;
    newValue = readSystemClockTimer();
    oldValue = newValue;
    while(numOfChanges)
    {
        newValue = readSystemClockTimer();
        if(oldValue != newValue)
        {
            numOfChanges--;
        }
        oldValue = newValue;
    }
}

/*******************************************************************************
* mvSataEventNotify - Receives event notifcations from CORE driver
*
*
* DESCRIPTION:
*   This function is part of the CORE driver OS dependent functions. It receives
*   event notifications from CORE driver upon SATA cabling state change or
*   PCI error events
*
* INPUT:
*   pSataAdapter    -   Pointer to MV_SATA_ADAPTER adapter data structure
*   eventType       -   Type of event
*   param1, param2  -   Parameters that are specific for the event.
*
* RETURN:
*   None
*
* NOTE : Refer to the CORE driver user manual for reference withregards the 
*   content of the eventType, param1 and param2 variables.
*
*******************************************************************************/
MV_BOOLEAN
mvSataEventNotify(MV_SATA_ADAPTER *pSataAdapter, MV_EVENT_TYPE eventType,
                  MV_U32 param1, MV_U32 param2)
{
    MV_PRINTD("IAL Event received: adapter %d event %s param1= %x param2 = %x\n",
              pSataAdapter->adapterId, (eventType == MV_EVENT_TYPE_ADAPTER_ERROR) ?
              "ADAPTER ERROR" : (param1 == 1)?"CONNECT":"DISCONNECT", param1, param2);
    return MV_TRUE;
}

/*******************************************************************************
* getCurrentInstalledDrives
*
*
* DESCRIPTION:
*   Gets the current installed drive number in the 0040:0075 BDA address
*
* INPUT:
*   Nona
*
* RETURN:
*   1 byte value of the 0040:0075 BDA address
*
*******************************************************************************/
unsigned char getCurrentInstalledDrives(void)
{
    unsigned char drive = 0;
    _asm{
        push es
        push ax
        mov ax, 0x0040
        mov es, ax
        mov al, es:0x0075
        mov drive, al
        pop es
        pop ax
    }
    return drive;
}

/*******************************************************************************
* setCurrentInstalledDrives
*
*
* DESCRIPTION:
*   Sets the current installed drive number in the 0040:0075 BDA address
*
* INPUT:
*   drive - 1 byte value to be written on the 0040:0075 BDA address
*
* RETURN:
*   Nona
*
*******************************************************************************/
void setCurrentInstalledDrives(unsigned char drive)
{
    _asm{
        push es
        push dx
        mov dx, 0x0040
        mov es, dx
        mov byte es:0x0075, al
        pop es
        pop dx
    }
    return;
}

/*******************************************************************************
* getCurrentInt13CodeSegment
*
*
* DESCRIPTION:
*   Gets the current code segment in entry 13 of the interrupt vector
*   table (int 13)
*
* INPUT:
*   Nona
*
* RETURN:
*   2 bytes
*
*******************************************************************************/
unsigned int    getCurrentInt13CodeSegment(void)
{
    unsigned short codeSegment = 0;
    _asm{
        push es
        push ax
        mov ax, 0x0000
        mov es, ax
        mov ax, es:0x4e ; 0x13 * 4 + 2
        mov codeSegment, ax
        pop es
        pop ax
    }
    return codeSegment;





}
/*******************************************************************************
* getCurrentInt13InstructionPointer
*
*
* DESCRIPTION:
*   Gets the current instruction pointer in entry 13 of the interrupt vector
*   table (int 13)
*
* INPUT:
*   Nona
*
* RETURN:
*   2 bytes
*
*******************************************************************************/
unsigned int    getCurrentInt13InstructionPointer(void)
{
    unsigned short instructionPointer = 0;
    _asm{
        push es
        push ax
        mov ax, 0x0000
        mov es, ax
        mov ax, es:0x4c ; 0x13 * 4
        mov instructionPointer, ax
        pop es
        pop ax
    }
    return instructionPointer;





}
/*******************************************************************************
* setCurrentInt13CodeSegment
*
*
* DESCRIPTION:
*   Sets the code segment value in entry 13 of the interrupt vector
*   table (int 13)
*
* INPUT:
*   2 bytes
*
* RETURN:
*   None
*
*******************************************************************************/
void    setCurrentInt13CodeSegment(unsigned short value)
{
    _asm{
        push di
        push es
        mov di, 0x0000
        mov es, di
        mov di, value
        mov es:0x4e, di ; 0x13 * 4 + 2
        pop es
        pop di
    }
}
/*******************************************************************************
* setCurrentInt13InstructionPointer
*
*
* DESCRIPTION:
*   Sets the instruction pointer in entry 13 of the interrupt vector
*   table (int 13)
*
* INPUT:
*   2 bytes
*
* RETURN:
*   None
*
*******************************************************************************/
void    setCurrentInt13InstructionPointer(unsigned short value)
{
    _asm{
        push di
        push es
        mov di, 0x0000
        mov es, di
        mov di, value
        mov es:0x4c, di ; 0x13 * 4
        pop es
        pop di
    }
}


/*******************************************************************************
* getCodeSegment
*
*
* DESCRIPTION:
*   Gets the code segment value of the function in the place it is currently
*   running from.
*
* INPUT:
*   None
*
* RETURN:
*   2 bytes
*
*******************************************************************************/
unsigned short  getCodeSegment(void)
{
    unsigned short value = 0;
    _asm{
        push di
        push cs
        pop di
        mov value, di
        pop di
    }
    return value;
}

/*******************************************************************************
* findAndInitChannels
*
*
* DESCRIPTION:
*   Initlizes mvSata adapter and all the SATA channels it is and all storage
*   devices connected to them.
*
* INPUT:
*   None
*
* RETURN:
*   MV_TRUE upon succes, MV_FALSE upon failure
*
*******************************************************************************/
MV_BOOLEAN findAndInitChannels(void)
{
    int numOfDrives = 0;
    MV_SATA_ADAPTER *pMvSataAdapter;
    unsigned char checkSum;
    MV_U8 channelIndex;

    memset (&sataAdapter, 0, sizeof(HW_ADAPTER_DESCRIPTION));

    sataAdapter.adapterBusNumber = mvSataBusNum;
    sataAdapter.adapterDeviceNumber = mvSataDeviceNum >> 3;
    sataAdapter.mvSataAdapter.adapterIoBaseAddress = mvSataBarOffset;
    sataAdapter.mvSataAdapter.adapterId = 0;
    sataAdapter.mvSataAdapter.pciConfigDeviceId = mvSataDeviceId;
    sataAdapter.mvSataAdapter.pciConfigRevisionId = mvSataRevisionId;
    sataAdapter.mvSataAdapter.intCoalThre[0]= 4;
    sataAdapter.mvSataAdapter.intCoalThre[1]= 4;
    sataAdapter.mvSataAdapter.intTimeThre[0] = 150*50 ;
    sataAdapter.mvSataAdapter.intTimeThre[1] = 150*50;
    sataAdapter.mvSataAdapter.mvSataEventNotify = mvSataEventNotify;
    sataAdapter.mvSataAdapter.IALData = &sataAdapter;
    sataAdapter.mvSataAdapter.pciSerrMask = MV_PCI_SERR_MASK_REG_ENABLE_ALL;
    sataAdapter.mvSataAdapter.pciInterruptMask = 0;
    sataAdapter.mvSataAdapter.pciCommand = MV_PCI_COMMAND_REG_DEFAULT;
    sataAdapter.requestsArrayBaseAddr = 0;

    pMvSataAdapter = &sataAdapter.mvSataAdapter;

    if(mvSataInitAdapter(&sataAdapter.mvSataAdapter) == MV_FALSE)
    {
        MV_ERROR ("Error Initializing SATA Adapter\n");
        return MV_FALSE;
    }

    /* Enable staggered spin-up of all SATA channels */
    mvSataEnableStaggeredSpinUpAll(&sataAdapter.mvSataAdapter);

    /*
     * The following is faking the adapter data structure as if we have
     * multipler channel data structure.
     */
    for(channelIndex = 0 ; channelIndex < sataAdapter.mvSataAdapter.numberOfChannels ; channelIndex++)
    {
        if(mvSataIsStorageDeviceConnected(&sataAdapter.mvSataAdapter, channelIndex))
        {
            InitChannel(&sataAdapter, channelIndex);
            if(mvSataConfigureChannel(&sataAdapter.mvSataAdapter, channelIndex) == MV_FALSE)
            {
                MV_ERROR ("Error in mvSataConfigureChannel on channel %d\n",channelIndex);
                sataAdapter.mvSataAdapter.sataChannel[channelIndex] = NULL;
            }
        } else
        {
            sataAdapter.mvSataAdapter.sataChannel[channelIndex] = NULL;
        }
    }

    /*
     * Issue SRST to all channels to port multiplier control port. If the
     * there is a port multiplier then it will send it's signature. In any
     * other case the PM port will be disregarded and normal signature will
     * be sent.
     */
    for(channelIndex = 0 ; channelIndex < sataAdapter.mvSataAdapter.numberOfChannels ; channelIndex++)
    {
        if(sataAdapter.mvSataAdapter.sataChannel[channelIndex])
        {
            if(mvStorageDevATAStartSoftResetDevice(&sataAdapter.mvSataAdapter,
                                                   channelIndex, 
                                                   MV_SATA_PM_CONTROL_PORT) == 
               MV_FALSE)
            {
                MV_ERROR ("Error - mvStorageDevATAStartSoftResetDevice failed on "
                          "channel %d\n", channelIndex);
                return MV_FALSE;
            }




        }
    }
    /* Check if SRST has finished and if signature (FIS 34) arrived */

    for(channelIndex = 0 ; channelIndex < sataAdapter.mvSataAdapter.numberOfChannels ; channelIndex++)
    {
        MV_STORAGE_DEVICE_REGISTERS registerStruct;
        MV_SATA_DEVICE_TYPE deviceType;
        MV_U16 timeout;

        /* This BIOS extension driver detects and hooks only first 8 drives */
        if(numOfDrives == MV_SATA_BIOS_MAX_DRIVES)
        {
            break;
        }
        if(sataAdapter.mvSataAdapter.sataChannel[channelIndex] == NULL)
        {
            continue;
        }
        memset (&registerStruct, 0, sizeof(MV_STORAGE_DEVICE_REGISTERS));
        timeout = 50; /* 10seconds */
        while(timeout)
        {
            timeout --;
            if(mvStorageIsDeviceBsyBitOff(&sataAdapter.mvSataAdapter,
                                          channelIndex,
                                          &registerStruct) == MV_TRUE)
            {
                break;
            }
            mvMicroSecondsDelay (0, 10000000/50);
        }
        if(timeout == 0)
        {
            MV_ERROR("Error - Did not get FIS 34 on channel %d. Disregarding "
                     " channel\n",channelIndex);
            continue;
        }
        deviceType = mvGetSataDeviceType(&registerStruct);
        mvStorageDevSetDeviceType (&sataAdapter.mvSataAdapter,
                                   channelIndex, deviceType);
        if(deviceType == MV_SATA_DEVICE_TYPE_ATA_DISK)
        {
            if(StartChannel(&sataAdapter,channelIndex,MV_FALSE,0) == MV_FALSE)
            {
                printf ("Failed initializing storage deivce connected to SATA channel %d\n",channelIndex);
                sataAdapter.mvSataAdapter.sataChannel[channelIndex] = 0;
            } else
            {
                /*
                 * StartChannel successed. Now save the corresponding SATA channel
                 * in the sataChannelNumber array to be used back again when
                 * hooking the drive to the int13 chain.
                 */
                sataChannelNumber[numOfDrives] = channelIndex;
                /* Direct connection, no PM port */
                portMultiplierPort[numOfDrives] = 0xff;
                numOfDrives ++;
            }
        } else if(deviceType == MV_SATA_DEVICE_TYPE_PM)
        {
            MV_BOOLEAN driveList[15];
            MV_U8 PMPort;
            printf("SATA %d - Found port multiplier\n", channelIndex);
            if(StartPM(&sataAdapter,channelIndex,driveList) == MV_FALSE)
            {
                printf("Error accessing port multipler - skipping channel\n");
                sataAdapter.mvSataAdapter.sataChannel[channelIndex] = NULL;
                continue;
            }
            for(PMPort = 0 ; PMPort < MV_SATA_PM_MAX_PORTS ; PMPort ++)
            {
                if((driveList[PMPort] == MV_TRUE) && 
                   (numOfDrives < MV_SATA_BIOS_MAX_DRIVES) &&
                   (sataAdapter.mvSataAdapter.sataChannel[channelIndex]))
                {
                    /* clear S Error*/
                    if(mvPMDevWriteReg(&sataAdapter.mvSataAdapter, channelIndex, PMPort,
                                       MV_SATA_PSCR_SERROR_REG_NUM, 0xffffffff, NULL) ==
                       MV_FALSE)
                    {
                        MV_ERROR("MV IAL[%d %d %d]: mvPMDevWriteReg Failed\n",
                                 &sataAdapter.mvSataAdapter.adapterId, channelIndex, PMPort);

                        sataAdapter.mvSataAdapter.sataChannel[channelIndex] = 0;
                        continue;
                    }

                    if(StartChannel(&sataAdapter,channelIndex,MV_TRUE,PMPort) == MV_FALSE)
                    {
                        printf ("Failed initializing storage deivce connected to "
                                "SATA channel %d PM Port %d\n",channelIndex, PMPort);
                        sataAdapter.mvSataAdapter.sataChannel[channelIndex] = 0;
                    } else
                    {
                        /*
                         * StartChannel successed. Now save the corresponding SATA channel
                         * in the sataChannelNumber array to be used back again when
                         * hooking the drive to the int13 chain.
                         */
                        sataChannelNumber[numOfDrives] = channelIndex;
                        /* temp variable is the desenation PM Port for the drive */
                        portMultiplierPort[numOfDrives] = PMPort;
                        numOfDrives ++;




                    }
                }
            }
        } else
        {
            printf("SATA %d - Received unrecognized device signature (%d)\n",
                   channelIndex, deviceType);
        }
    }

    mvSataMaskAdapterInterrupt (&sataAdapter.mvSataAdapter);
    for(channelIndex = 0 ; channelIndex < numOfDrives ; channelIndex++)
    {
        unsigned char *pPnpHeader;
        unsigned char *productNameString;
        unsigned char deviceNum;
        int temp;
        pPnpHeader = (unsigned char *) (0x100 + channelIndex*32);

#ifdef DEBUG_IAL
        printf("Modifying BCV #%d properties\n",channelIndex);
#endif

        /* Set pointer to product name string in PnP Header */
        productNameString = (unsigned char *)* (unsigned short *) (pPnpHeader + 0x10);
        /* Set bus number in product name string */
        productNameString[5] = ((mvSataBusNum / 10) % 10) + '0';
        productNameString[6] = (mvSataBusNum % 10) + '0';
        productNameString[7] = '-';
        /* Set devoce number in product name string */
        deviceNum = mvSataDeviceNum >> 3;
        productNameString[8] = ((deviceNum / 10) % 10) + '0';
        productNameString[9] = (deviceNum % 10) + '0';
        /* Set SATA Channel number */
        productNameString[11] = sataChannelNumber[channelIndex] + '0';
        if(portMultiplierPort[channelIndex] != 0xff)
        {
            productNameString[12] = '-';
            productNameString[13] = intToChr(portMultiplierPort[channelIndex]);
            productNameString[14] = 0;
        } else
        {
            productNameString[12] = 0;




        }
        /* Is this last BCV Entry ? */
        if(channelIndex == (numOfDrives - 1))
        {
            /* Set last PnPHeader to point to NULL in it's next PnPHeader field */
            * (unsigned short *)(pPnpHeader + 0x6) = 0;
        }

        /* Recalculate checksum of pnp expansiom ROM header */
        checkSum = 0;
        /* Reset checksum on current PnP header */
        pPnpHeader[0x9] = 0;
        for(temp = 0 ; temp < 32 ; temp ++)
        {
            checkSum += * (unsigned char *)(pPnpHeader + temp);
        }
        pPnpHeader[0x9] = (~checkSum) + 1;




    }
    if(numOfDrives == 0)
    {
        printf("No connected drives found\n");
        resizeToZero ();
    }
    return MV_TRUE;
}


/*******************************************************************************
* outCharacterSerial
*
*
* DESCRIPTION:
*   Prints a single character to the serial port
*
* INPUT:
*   character - The character to be printed
*
* RETURN:
*   None
*
*******************************************************************************/
void outCharacterSerial(unsigned char character)
{
#ifdef DEBUG_IAL_SERIAL
    _asm{
        nop
        mov al, character
        mov ah, 0x1
        mov dx, 0
        int 0x14
    }
#endif



}

/*******************************************************************************
* initializeSerial
*
*
* DESCRIPTION:
*   Initializes COM1 of the PC
*
* INPUT:
*   None
*
* RETURN:
*   None
*
*******************************************************************************/
void initializeSerial(void)
{
#ifdef DEBUG_IAL_SERIAL

    /* 9600bps 8N1 */
    _asm{
        nop
        mov ah, 0x0
        mov al, 0xe3
        mov dx, 0
        int 0x14
    }
#endif



}

/*******************************************************************************
* outCharacter
*
*
* DESCRIPTION:
*   Prints a single character on the console (video adapter or other)
*
* INPUT:
*   character - The character to be printed
*
* RETURN:
*   None
*
*******************************************************************************/
void outCharacter(unsigned char character)
{
    _asm{
        nop
        mov al, character
        mov ah, 0xe
        mov bh, 0
        mov bl, 0x7
        int 0x10
    }
}

/*******************************************************************************
* printStringSerial
*
*
* DESCRIPTION:
*   Prints a string to the serial port using outCharacterSerial function
*
* INPUT:
*   msg - A pointer to string
*
* RETURN:
*   None
*
*******************************************************************************/
int printStringSerial(char *msg)
{
    int i;
    unsigned short savedDataSegment = 0;
    _asm{
        push cx
        mov cx, ds
        mov savedDataSegment, cx
        mov cx, ss
        mov ds, cx
        pop cx





    }
    for(i = 0 ; msg[i] ; i++)
    {
        outCharacterSerial(msg[i]);
        if(msg[i] == 10)
        {
            outCharacterSerial(13);
        }
    }
    _asm{
        push cx
        mov cx, savedDataSegment
        mov ds, cx
        pop cx
    }

    return 0;
}

/*******************************************************************************
* printString
*
*
* DESCRIPTION:
*   Prints a string on the console using outCharacter function
*
* INPUT:
*   msg - A pointer to string
*
* RETURN:
*   None
*
*******************************************************************************/
int printString(char *msg)
{
    int i;
    unsigned short savedDataSegment = 0;
    _asm{
        push cx
        mov cx, ds
        mov savedDataSegment, cx
        mov cx, ss
        mov ds, cx
        pop cx





    }
    for(i = 0 ; msg[i] ; i++)
    {
        outCharacter(msg[i]);
        if(msg[i] == 10)
        {
            outCharacter(13);
        }
    }
    _asm{
        push cx
        mov cx, savedDataSegment
        mov ds, cx
        pop cx
    }

    return 0;
}


/*******************************************************************************
* readRegisterIoBar
*
*
* DESCRIPTION:
*   Reads a 32 bit register from mvSata adapter through IO BAR
*
* INPUT:
*   base    - A 16bit base address of the IO BAR
*   offset  - A 32bit offset of the register
*
* RETURN:
*   32bit value of the register
*
* NOTE:
*   This function utilizes the PCI IO BAR
*******************************************************************************/
static unsigned long readRegisterIoBar (unsigned short base_, unsigned long offset_)
{
    unsigned long value = 0;
    _asm{
        push eax
        push dx

        mov dx, base_
        mov eax, offset_
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        in eax, dx

        mov value, eax
        pop dx
        pop eax
    };
    return value;
}


/*******************************************************************************
* writeRegisterIoBar
*
*
* DESCRIPTION:
*   Writes a 32 bit value to a register in a mvSata adapter through IO BAR
*
* INPUT:
*   base    - A 16bit base address of the IO BAR
*   offset  - A 32bit offset of the register
*   value   - A 32bit of the value to be written to the register
*
* RETURN:
*   None
*
* NOTE:
*   This function utilizes the PCI IO BAR
*******************************************************************************/
static void writeRegisterIoBar (unsigned short base_,
                                unsigned long offset_,
                                unsigned long value_)
{
    _asm{
        push eax
        push dx
        mov dx, base_
        mov eax, offset_
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        mov eax, value_
        out dx, eax
        pop dx
        pop eax
    };
}

/*******************************************************************************
* readRegisterProtectedMode
*
*
* DESCRIPTION:
*   Reads a 32 bit register from mvSata adapter in protected mode (Mem BAR)
*
* INPUT:
*   offset  - A 32bit offset of the register (absolute offset)
*
* RETURN:
*   32bit of the value of the register
*
* NOTE:
*   This function utilizes the protected mode of the processor in order to
*   access a 32bit region.
*******************************************************************************/
static unsigned long readRegisterProtectedMode (register unsigned long offset)
{
    unsigned long value = 0;
    _asm{
        push ecx
        push edx
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  ecx, gs:[edx]


        ;       You can add the code in protected mode code here

        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        mov  value, ecx
        pop edx
        pop ecx
    };
    return value;
}


/*******************************************************************************
* writeRegisterProtectedMode
*
*
* DESCRIPTION:
*   Writes a 32 bit value to a register in a mvSata adapter in protected mode
*   (Mem BAR).
*
* INPUT:
*   offset  - A 32bit offset of the register (absolute offset)
*   value   - A 32bit of the value to be written in the register
*
* RETURN:
*   None
*
* NOTE:
*   This function utilizes the protected mode of the processor in order to
*   access a 32bit region.
*******************************************************************************/
static void writeRegisterProtectedMode (unsigned long offset, unsigned long value)
{
    _asm{
        push ecx
        push edx
        ; value
        shl ecx, 16
        mov cx, bx
        ; offset
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  gs:[edx], ecx


        ;       You can add the code in protected mode code here

        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        pop edx
        pop ecx
    };
}

unsigned long readRegister (unsigned long base, unsigned long offset)
{
    if(useIoBar == MV_TRUE)
    {
        return readRegisterIoBar ((unsigned short)base, offset);
    } else
    {
        return readRegisterProtectedMode(base+offset);
    }
}

void writeRegister (unsigned long base, unsigned long offset, unsigned long value)
{
    if(useIoBar == MV_TRUE)
    {
        writeRegisterIoBar ((unsigned short)base,offset,value);
    } else
    {
        writeRegisterProtectedMode (base+offset,value);
    }
    return;
}


/*******************************************************************************
* activateA20Gate
*
*
* DESCRIPTION:
*   Queries keyboard controller to activate A20 gate line
*
* INPUT:
*   None
* RETURN:
*   None
*
*******************************************************************************/
static void activateA20Gate(void)
{
    _asm{
        cli

        call    a20wait
        mov     al,0xAD
        out     0x64,al

        call    a20wait
        mov     al,0xD0
        out     0x64,al

        call    a20wait2
        in      al,0x60
        push    eax

        call    a20wait
        mov     al,0xD1
        out     0x64,al

        call    a20wait
        pop     eax
        or      al,2
        out     0x60,al

        call    a20wait
        mov     al,0xAE
        out     0x64,al

        call    a20wait
        jmp go_out

        a20wait:
        l0: mov     ecx,65536
        l1: in      al,0x64
        test    al,2
        jz      l2
        loop    l1
        jmp     l0
        l2: ret


        a20wait2:
        c0: mov     ecx,65536
        c1: in      al,0x64
        test    al,1
        jnz     c2
        loop    c1
        jmp     c0
        c2: ret

        go_out :
        nop
    }
}
/*******************************************************************************
* mvSataReadConfigWord
*
*
* DESCRIPTION:
*   Utilizing the BIOS32 PCI services, this function reads 16 bit from the
*   configuration space of a specific mvSata adapter
*
* INPUT:
*   offset  -   Offset of the required register within the configuration space
*
* RETURN:
*   16bit value of the requested register
*
*******************************************************************************/
unsigned short mvSataReadConfigWord(unsigned short offset)
{
    unsigned short temp = 0;
    _asm{
        pusha
        mov di, ax ; The offset is in ax */
        mov ax, 0xb109
        mov bh, mvSataBusNum
        mov bl, mvSataDeviceNum
        int 0x1a
        mov temp, cx
        popa
    }
    return temp;





}

/*******************************************************************************
* mvSataWriteConfigWord
*
*
* DESCRIPTION:
*   Utilizing the BIOS32 PCI services, this function writes 16 bit value to
*   the configuration space of a specific mvSata adapter
*
* INPUT:
*   offset  -   Offset of the required register within the configuration space
*   data    -   The value to be written on the configuration space register
*
* RETURN:
*   None
*
*******************************************************************************/
void mvSataWriteConfigWord(unsigned short offset, unsigned short data)
{
    _asm{
        pusha
        mov di, ax ; The offset is in ax */
        mov ax, 0xb10c
        mov bh, mvSataBusNum
        mov bl, mvSataDeviceNum
        mov cx, dx
        int 0x1a
        popa
    }
}

/*******************************************************************************
* initlizeMvSataAdapter
*
*
* DESCRIPTION:
*   This function queries the mvSata adapter PCI configuration space by reading,
*   writing and setting parameters.
*
* INPUT:
*   None
*
* RETURN:
*   None
*
*******************************************************************************/
int initlizeMvSataAdapter (void)
{
    unsigned short pciCommand;
    if(mvSataResult)
    {
        printf ("Couldn't find adapter\n");
        return -1;
    }
    printf ("Found adapter at bus %d, device %d\n",mvSataBusNum,
            (mvSataDeviceNum >> 3));


    if(useIoBar == MV_TRUE)
    {
        mvSataBarOffset = mvSataReadConfigWord(0x1a);
        mvSataBarOffset = (mvSataBarOffset << 16) | mvSataReadConfigWord(0x18);
        mvSataBarOffset = mvSataBarOffset & 0xfffffff0;

        if(mvSataBarOffset == 0x0)
        {
            printf("Error - BIOS did not initialize mvSata IO BAR\n");
            return -1;
        }
        if(mvSataBarOffset & 0xffff0000)
        {
            printf("Error - BIOS initialized IO BAR high 16bit address with non zero value (%lx)\n",mvSataBarOffset);
            return -1;
        }
    } else
    {
        mvSataBarOffset = mvSataReadConfigWord(0x12);
        mvSataBarOffset = (mvSataBarOffset << 16) | mvSataReadConfigWord(0x10);
        mvSataBarOffset = mvSataBarOffset & 0xfffffff0;
        if(mvSataBarOffset == 0x14000000)
        {
            printf("Error - BIOS did not initialize mvSata BAR 0\n");
            return -1;
        }
    }


    /* Enable master / IO / Memory accesses */
    /* Read PCI Command regsiter */
    pciCommand = mvSataReadConfigWord(0x4);

    /* No need to enable the bits */
    if(pciCommand & 0x7)
    {
        return 0;
    }

    pciCommand = pciCommand | 0x7;

    mvSataWriteConfigWord(0x4, pciCommand);

    /* Read command register to verify that command is written */
    pciCommand = mvSataReadConfigWord(0x4);

    /* Check if bits are enabled */
    if((pciCommand & 0x7) != 0x7)
    {
        return -1;
    }

    return 0;






}

/*******************************************************************************
* InitChannel
*
*
* DESCRIPTION:
*   Initializes specifc SATA channel parameters such as request / response
*   queue pointers
*
* INPUT:
*   HwDeviceExtension   -   Pointer to HwDeviceExtension data structure
*   channelIndex        -   Number of SATA channel
*
* RETURN:
*   None
*
*******************************************************************************/
void
InitChannel(
           PHW_ADAPTER_DESCRIPTION HwDeviceExtension,
           MV_U8 channelIndex)
{
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_SATA_CHANNEL     *pSataChannel;

    pSataChannel = &HwDeviceExtension->mvSataChannels[channelIndex];
    pSataAdapter->sataChannel[channelIndex] = pSataChannel;
    pSataChannel->channelNumber = channelIndex;
    pSataChannel->requestQueue = 0;
    pSataChannel->requestQueuePciLowAddress = 0;
    pSataChannel->requestQueuePciHiAddress = 0;
    pSataChannel->responseQueue = 0;
    pSataChannel->responseQueuePciLowAddress = 0;
    pSataChannel->responseQueuePciHiAddress = 0;
}

static void
Ata2HostString(MV_U16 *source,
               MV_U16 *target,
               MV_U16 wordsCount
              )
{
    MV_U16 i;
    for(i=0 ; i < wordsCount; i++)
    {
        target[i] = (source[i] >> 8) | ((source[i] & 0xff) << 8);
    }
}

/*******************************************************************************
* StartChannel
*
*
* DESCRIPTION:
*   Initializes specifc SATA channel by having few diagnostics tests then
*   querying storage device connected to the SATA channel and configuring both
*   storage device and SATA channel to have maximum supported interface speed.
*
* INPUT:
*   HwDeviceExtension   -   Pointer to HwDeviceExtension data structure
*   channelIndex        -   Number of SATA channel
*
* RETURN:
*   None
*
*******************************************************************************/
MV_BOOLEAN
StartChannel(
            PHW_ADAPTER_DESCRIPTION HwDeviceExtension,
            MV_U8 channelIndex, MV_BOOLEAN PMEnabled, MV_U8 PMPort)
{
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_SATA_CHANNEL     *pSataChannel;
    SATA_CHANNEL_INFO   *channelInfo = &(HwDeviceExtension->channelsInfo[channelIndex]);
    MV_U16              identifyBuffer[MV_ATA_IDENTIFY_DEV_DATA_LENGTH];
    char                driveModel[30];
    ATA_IDENTIFY_INFO   identifyInfo;

    pSataChannel = pSataAdapter->sataChannel[channelIndex];
    if(!pSataChannel)
    {
        MV_ERROR ("Error in StartChannel - Channel data structure is null\n");
        return MV_FALSE;
    }

    if(mvStorageDevATASoftResetDevice(pSataAdapter, channelIndex, PMPort, 0) == MV_FALSE)
    {
        printf("Error - Failed initializing(SRST) drive on channel %d, PMPort %d\n",
               channelIndex, PMPort);
        return MV_FALSE;
    }

    if(mvInitSataDisk (pSataAdapter, channelIndex, PMPort, &identifyInfo,
                       identifyBuffer) == MV_FALSE)
    {
        printf("Error - Failed initializing drive on channel %d, PMPort %d\n",
               channelIndex, PMPort);
        return MV_FALSE;
    }
    channelInfo->diskSize[PMPort] = identifyInfo.ATADiskSize;
    channelInfo->supportsLBA48Address[PMPort] = identifyInfo.LBA48Supported;
    Ata2HostString(identifyBuffer + IDEN_MODEL_OFFSET, (MV_U16_PTR)driveModel,
                   13);
    driveModel[25] = '\0';
    if(PMEnabled == MV_TRUE)
    {
        printf("SATA %d (PM %d) : ",channelIndex, PMPort);
    } else
    {
        printf("SATA %d : ",channelIndex);
    }
    printf("MODEL [%25s]", driveModel);

    if(channelInfo->diskSize[PMPort] / (2048UL))
    {
        printf (" SIZE %ld MByte\n",channelInfo->diskSize[PMPort] / (2048UL));
    } else
    {
        printf ("\n");
    }

    return MV_TRUE;
}






