/*******************************************************************************
*
*                   Copyright 2003,MARVELL SEMICONDUCTOR ISRAEL, LTD.
* THIS CODE CONTAINS CONFIDENTIAL INFORMATION OF MARVELL.
* NO RIGHTS ARE GRANTED HEREIN UNDER ANY PATENT, MASK WORK RIGHT OR COPYRIGHT
* OF MARVELL OR ANY THIRD PARTY. MARVELL RESERVES THE RIGHT AT ITS SOLE
* DISCRETION TO REQUEST THAT THIS CODE BE IMMEDIATELY RETURNED TO MARVELL.
* THIS CODE IS PROVIDED "AS IS". MARVELL MAKES NO WARRANTIES, EXPRESSED,
* IMPLIED OR OTHERWISE, REGARDING ITS ACCURACY, COMPLETENESS OR PERFORMANCE.
*
* MARVELL COMPRISES MARVELL TECHNOLOGY GROUP LTD. (MTGL) AND ITS SUBSIDIARIES,
* MARVELL INTERNATIONAL LTD. (MIL), MARVELL TECHNOLOGY, INC. (MTI), MARVELL
* SEMICONDUCTOR, INC. (MSI), MARVELL ASIA PTE LTD. (MAPL), MARVELL JAPAN K.K.
* (MJKK), MARVELL SEMICONDUCTOR ISRAEL. (MSIL),  MARVELL TAIWAN, LTD. AND
* SYSKONNECT GMBH.
*
*******************************************************************************/
#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include "mvFlash.h"
#include "mvRegs.h"
#include <string.h>



/* Defines */
#undef DEBUG
#define PCI_DATA_STRUCTURE_OFFSET	0x18
#define DEVICE_ID_OFFSET_IN_PCIDS	0x6
/* Global variables */
/* Global descriptors table */
unsigned int GDT[] = {  0xf, 0, 0, 0, 
    0xffff, 0, 0x9200, 0x8f};

int useIOBar = 0;

union REGS regs;
struct mvPciSataAdapter
{
    int valid;
    unsigned char bus;
    unsigned char devFunc;
    unsigned long bar;
    unsigned char version;
    unsigned char subVersion;

};

/*******************************************************************************
* activateA20Gate
*
*
* DESCRIPTION:
*   Queries keyboard controller to activate A20 gate line
*
* INPUT:
*	None
* RETURN:
*	None
*
*******************************************************************************/
static void activateA20Gate(void)
{
    _asm {
        cli

        call    a20wait
        mov     al,0xAD
        out     0x64,al

        call    a20wait
        mov     al,0xD0
        out     0x64,al

        call    a20wait2
        in      al,0x60
        push    eax

        call    a20wait
        mov     al,0xD1
        out     0x64,al

        call    a20wait
        pop     eax
        or      al,2
        out     0x60,al

        call    a20wait
        mov     al,0xAE
        out     0x64,al

        call    a20wait
        jmp go_out

        a20wait:
        l0: mov     ecx,65536
        l1: in      al,0x64
        test    al,2
        jz      l2
        loop    l1
        jmp     l0
        l2: ret


        a20wait2:
        c0: mov     ecx,65536
        c1: in      al,0x64
        test    al,1
        jnz     c2
        loop    c1
        jmp     c0
        c2: ret

        go_out :
        nop
    }
}

unsigned char read8bitIoBar (unsigned short offst, register unsigned long address)
{
    unsigned char value = 0;
    _asm{
        pushad
        mov dx, offst
        mov eax, address
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        mov eax, address
        and ax, 0x3
        add dx, ax
        in al, dx

        mov value, al
        popad
    };
#ifdef DEBUG
    printf("Reading from IOBar %08lx : %x\n",offst, value);
#endif
    return value;
}

unsigned short read16bitIoBar (unsigned short offst, register unsigned long address)
{
    unsigned short value = 0;
    _asm{
        pushad
        mov dx, offst
        mov eax, address
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        mov eax, address
        and ax, 0x3
        add dx, ax
        in ax, dx

        mov value, ax
        popad
    };
#ifdef DEBUG
    printf("Reading from IOBar %08lx : %x\n",offst, value);
#endif
    return value;
}


unsigned long read32bitIoBar (unsigned short offst, register unsigned long address)
{
    unsigned long value = 0;
    _asm{
        pushad
        mov dx, offst
        mov eax, address
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        mov eax, address
        and ax, 0x3
        add dx, ax
        in eax, dx

        mov value, eax
        popad
    };
    return value;
}

void write8bitIoBar (unsigned short offst, register unsigned long address, unsigned char value)
{
    _asm{
        pushad
        mov dx, offst
        mov eax, address
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        mov eax, address
        and ax, 0x3
        add dx, ax
        mov al, value
        out dx, al
        popad
    };
}

void write16bitIoBar (unsigned short offst, register unsigned long address, unsigned short value)
{
    _asm{
        pushad
        mov dx, offst
        mov eax, address
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        mov eax, address
        and ax, 0x3
        add dx, ax
        mov ax, value
        out dx, ax
        popad
    };
}

void write32bitIoBar (unsigned short offst, register unsigned long address, unsigned long value)
{
    _asm{
        pushad
        mov dx, offst
        mov eax, address
        and al, 0xfc
        out dx, eax

        add dx, 0x80
        mov eax, address
        and ax, 0x3
        add dx, ax
        mov eax, value
        out dx, eax
        popad
    };
}

/*******************************************************************************
* read8bitMemBar
*
*
* DESCRIPTION:
*   Reads 8 bit from 32bit address space
*
* INPUT:
*   offset	- A 32bit offset 
*
* RETURN:
*   8 bit
*
* NOTE:
*	This function utilizes the protected mode of the processor in order to
*	access a 32bit region.
*******************************************************************************/
unsigned char read8bitMemBar (register unsigned long offset)
{
    /* Use protected mode */
    unsigned char value = 0;
    _asm{
        pushad
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  cl, gs:[edx]
        ;       You can add the code in protected mode code here
        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        mov  value, cl
        popad
    };
#ifdef DEBUG
    printf("Reading from %08lx : %x\n",offset, value);
#endif
    return value;
}


/*******************************************************************************
* read16bitMemBar
*
*
* DESCRIPTION:
*   Reads 16 bit from 32bit address space
*
* INPUT:
*   offset	- A 32bit offset 
*
* RETURN:
*   8 bit
*
* NOTE:
*	This function utilizes the protected mode of the processor in order to
*	access a 32bit region.
*******************************************************************************/
unsigned short read16bitMemBar (register unsigned long offset)
{
    /* Use protected mode */
    unsigned short value = 0;
    _asm{
        pushad
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  cx, gs:[edx]
        ;       You can add the code in protected mode code here
        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        mov  value, cx
        popad
    };
#ifdef DEBUG
    printf("Reading from %08lx : %x\n",offset, value);
#endif
    return value;
}

/*******************************************************************************
* read32bitMemBar
*
*
* DESCRIPTION:
*   Reads 32 bit from 32bit address space
*
* INPUT:
*   offset	- A 32bit offset 
*
* RETURN:
*   8 bit
*
* NOTE:
*	This function utilizes the protected mode of the processor in order to
*	access a 32bit region.
*******************************************************************************/
unsigned long read32bitMemBar (register unsigned long offset)
{
    /* Use protected mode */
    unsigned long value = 0;
    _asm{
        pushad
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  ecx, gs:[edx]
        ;       You can add the code in protected mode code here
        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        mov  value, ecx
        popad
    };
#ifdef DEBUG
    printf("Reading from %08lx : %08lx\n",offset, value);
#endif
    return value;
}

/*******************************************************************************
* write8bitMemBar
*
*
* DESCRIPTION:
*   Writes 8 bit to 32bit address space
*
* INPUT:
*   offset	- A 32bit offset 
*	value	- 8 bit value
*
* RETURN:
*   None
*
* NOTE:
*	This function utilizes the protected mode of the processor in order to
*	access a 32bit region.
*******************************************************************************/
void write8bitMemBar (unsigned long offset, unsigned char value)
{
    /* Use protected mode */
    _asm{
        pushad
        ; value
        shl ecx, 16
        mov cx, bx
        ; offset
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  gs:[edx], cl


        ;       You can add the code in protected mode code here

        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        popad
    };
#ifdef DEBUG
    printf("Writing to %08lx : %x\n",offset, value);
#endif
    return;
}

/*******************************************************************************
* write16bitMemBar
*
*
* DESCRIPTION:
*   Writes 16 bit to 32bit address space
*
* INPUT:
*   offset	- A 32bit offset 
*	value	- 16 bit value
*
* RETURN:
*   None
*
* NOTE:
*	This function utilizes the protected mode of the processor in order to
*	access a 32bit region.
*******************************************************************************/
void write16bitMemBar (unsigned long offset, unsigned short value)
{
    /* Use protected mode */
    _asm{
        pushad
        ; value
        shl ecx, 16
        mov cx, bx
        ; offset
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  gs:[edx], cx


        ;       You can add the code in protected mode code here

        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        popad
    };
#ifdef DEBUG
    printf("Writing to %08lx : %x\n",offset, value);
#endif
    return;
}
/*******************************************************************************
* write32bitMemBar
*
*
* DESCRIPTION:
*   Writes 32 bit to 32bit address space
*
* INPUT:
*   offset	- A 32bit offset 
*	value	- 32 bit value
*
* RETURN:
*   None
*
* NOTE:
*	This function utilizes the protected mode of the processor in order to
*	access a 32bit region.
*******************************************************************************/
void write32bitMemBar (unsigned long offset, unsigned long value)
{
    /* Use protected mode */
    _asm{
        pushad
        ; value
        shl ecx, 16
        mov cx, bx
        ; offset
        shl edx, 16
        mov dx, ax

        nop
        nop
        cli                ;Don't allow interrupts during this
        push gs            ;Save real mode selectors
        push eax

        lgdt [GDT]         ;Load the GDT
        mov  eax, cr0      ;Switch to pmode
        inc  ax
        mov  cr0, eax
        ; Now we are in protected mode
        mov  ax, 0x8 ;Our only pmode selector
        mov  gs, ax        ;Install 4Gb limits (warning)

        mov  gs:[edx], ecx


        ;       You can add the code in protected mode code here

        mov eax, cr0
        dec  ax            ;Switch back to real mode
        mov  cr0, eax

        pop  eax
        pop  gs
        sti
        popad
    };
#ifdef DEBUG
    printf("Writing to %08lx : %08lx\n",offset, value);
#endif
    return;
}

unsigned char read8bit (unsigned long offset, unsigned long address)
{
    if (useIOBar)
    {
        return read8bitIoBar ((unsigned short)offset,address | 0x80000000UL);
    } else
    {
        return read8bitMemBar (offset + address);
    }
}
unsigned short read16bit (unsigned long offset, unsigned long address)
{
    if (useIOBar)
    {
        return read16bitIoBar ((unsigned short)offset,address | 0x80000000UL);
    } else
    {
        return read16bitMemBar (offset + address);
    }
}
unsigned long read32bit (unsigned long offset, unsigned long address)
{
    if (useIOBar)
    {
        return read32bitIoBar ((unsigned short)offset,address | 0x80000000UL);
    } else
    {
        return read32bitMemBar (offset + address);
    }
}


void write8bit (unsigned long offset, unsigned long address, unsigned char value)
{
    if (useIOBar)
    {
        write8bitIoBar (offset, address | 0x80000000UL, value);
    } else
    {
        write8bitMemBar (offset + address, value);
    }
}
void write16bit (unsigned long offset, unsigned long address, unsigned short value)
{
    if (useIOBar)
    {
        write8bitIoBar (offset, address | 0x80000000UL, value);
    } else
    {
        write16bitMemBar (offset + address, value);
    }
}
void write32bit (unsigned long offset, unsigned long address, unsigned long value)
{
    if (useIOBar)
    {
        write32bitIoBar (offset, address | 0x80000000UL, value);
    } else
    {
        write32bitMemBar (offset + address, value);
    }
}


int readConfig16bit(unsigned char bus, unsigned char devFunc, 
                    unsigned short offset, unsigned short *value)
{

    regs.h.ah = 0xb1;
    regs.h.al = 0x9;
    regs.h.bh = bus;
    regs.h.bl = devFunc;
    regs.w.di = offset;
    int86 (0x1a, &regs,&regs);
    * value = regs.w.cx;

    if (regs.w.cflag & 0x1)
    {
        return regs.h.ah;
    }
    return 0;
}

int writeConfig16bit(unsigned char bus, unsigned char devFunc, 
                     unsigned short offset, unsigned short value)
{

    regs.h.ah = 0xb1;
    regs.h.al = 0xc;
    regs.h.bh = bus;
    regs.h.bl = devFunc;
    regs.w.di = offset;
    regs.w.cx = value;
    int86 (0x1a, &regs,&regs);
    if (regs.w.cflag & 0x1)
    {
        return regs.h.ah;
    }
    return 0;
}

void switchBackBetweenBar0andExpRom(unsigned char bus, unsigned char devFunc,
                                    unsigned long bar0)
{
    if (!useIOBar)
    {
        /* Switch between BAR0 and EXP ROM BAR */
        writeConfig16bit(bus, devFunc, 0x30, 0);
        writeConfig16bit(bus, devFunc, 0x32, 0);
        writeConfig16bit(bus, devFunc, 0x10, bar0 & 0xffff);
        writeConfig16bit(bus, devFunc, 0x12, (bar0 >> 16) & 0xffff);
    }
}

void main (int argc, char **argv)
{
    int i, numOfAdapters;
    struct mvPciSataAdapter adapterList[10] = {0,};
    FILE *romImage;
    unsigned long imageSize;
    unsigned char ch;
    unsigned long temp, *temp1;
    unsigned int dataSegment;
    unsigned long flashSize, offset;
    int sectorsToErase;
    unsigned char imageVersion, imageSubversion;
    unsigned short deviceId = 0, PCIDS = 0;


    if (argc != 2)
    {
        printf("Usage - %s <rom image>\n",argv[0]);
        exit (-1);
    }
    romImage = fopen (argv[1], "rob");
    if (!romImage)
    {
        printf("ROM Image %s not found\n",argv[1]);
        exit (-1);
    }

    /* Read the device ID of the adapter from the requested ROM image */
    /*
     * First look in PnP Option ROM header file for offset of PCI data
     * structure.
     */
    i = PCI_DATA_STRUCTURE_OFFSET;
    while (i)
    {
        int result = fgetc(romImage);
        if (result == EOF)
        {
            break;
        }
        i--;
    }
    if (i != 0)
    {
        printf("Error seeking for Device ID in requested ROM image\n");
        exit(-1);
    }
    for (i = 0 ; i < 2 ; i ++)
    {
        int result;
        result = fgetc(romImage);
        if (result == EOF)
        {
            printf("Error seeking for Device ID in requested ROM image\n");
            exit(-1);
        }
        PCIDS = (PCIDS >> 8) | (((unsigned char) result) << 8);
    }
    /* 
     * Seek to Device ID. This is offset 0x6 in PCI Data structure minus
     * what was already read from the image.
     */
    i = PCIDS+ DEVICE_ID_OFFSET_IN_PCIDS -
        (PCI_DATA_STRUCTURE_OFFSET + 0x2);
    while (i)
    {
        int result = fgetc(romImage);
        if (result == EOF)
        {
            break;
        }
        i--;
    }
    if (i != 0)
    {
        printf("Error seeking for Device ID in requested ROM image\n");
        exit(-1);
    }

    for (i = 0 ; i < 2 ; i ++)
    {
        int result;
        result = fgetc(romImage);
        if (result == EOF)
        {
            printf("Error seeking for Device ID in requested ROM image\n");
            exit(-1);
        }
        deviceId = (deviceId >> 8) | (((unsigned char) result) << 8);
    }
    if ((deviceId != 0x5081) && (deviceId != 0x5041) &&
        (deviceId != 0x6081) && (deviceId != 0x6041))
    {
        printf("Error - Device ID in requested ROM image is invalid (%04x)\n",
               (unsigned long)deviceId);
        exit (-1);
    }
    if ((deviceId == 0x6081) ||
        (deviceId == 0x6041))
    {
        useIOBar = 1;
    } else
    {
        useIOBar = 0;
    }

    fclose (romImage);
    romImage = fopen (argv[1], "rob");
    if (!romImage)
    {
        printf("ROM Image %s not found\n",argv[1]);
        exit (-1);
    }
    if (!useIOBar)
    {
        activateA20Gate();
        {
            temp = 0;
            dataSegment = 0;
            _asm{mov dataSegment, ds};
            temp = dataSegment;
            temp = temp << 4;

            temp += (unsigned long) GDT;
            temp1 = (unsigned long *)(((unsigned int) GDT) + 2);
            * (unsigned long *) temp1 = temp;
        }
    }

    /* Fill adapterList data structure with all mvSata adapters on PCI */
    i = 0;
    printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
    printf("###############################################################\n");
    printf("##                                                           ##\n");
    printf("## Marvell Serial ATA BIOS upgrade utility                   ##\n");
    printf("##                                                           ##\n");
    printf("###############################################################\n");
    printf("##                                                           ##\n");
    printf("## Detecting 88SX%04x Adapters ...                           ##\n",(unsigned long)deviceId);
    printf("##                                                           ##\n");
    while (1)
    {
        unsigned short temp;
        regs.h.ah = 0xb1;
        regs.h.al = 0x2;
        regs.w.cx = deviceId;
        regs.w.dx = 0x11ab;
        regs.w.si = i;
        int86 (0x1a, &regs,&regs);
        if (regs.w.cflag & 0x1)
        {
            break;
        }
        adapterList[i].bus = regs.h.bh;
        adapterList[i].devFunc = regs.h.bl;


        if (useIOBar)
        {
            if (readConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x18, &temp))
            {
                break;
            }
            adapterList[i].bar = temp & 0xfff0;


            if (readConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x1a, &temp))
            {
                break;
            }
            adapterList[i].bar = adapterList[i].bar | ((unsigned long)((unsigned long)(temp) << 16));
            if (adapterList[i].bar & 0xffff0000UL)
            {
                printf ("Error - Can't access IO Bar (offset %08lx)\n",
                        adapterList[i].bar);
                exit (-1);
            }
            if (adapterList[i].bar & 0xffffff00UL == 0)
            {
                printf ("Error - BIOS did not inialize IO BAR PCI resources\n");
                exit (-1);
            }
        } else
        {
            if (readConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x10, &temp))
            {
                break;
            }
            adapterList[i].bar = temp & 0xfff0;


            if (readConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x12, &temp))
            {
                break;
            }
            adapterList[i].bar = adapterList[i].bar | ((unsigned long)((unsigned long)(temp) << 16));
            /* 
             * In case BIOS did not allocate memory space to BAR 0, then disregard
             * the adapter
             */


            if (adapterList[i].bar == 0x14000000)
            {
                printf ("Error - BIOS did not inialize PCI resources\n");
                exit (-1);
            }

        }

        /* Print the device found */
        printf("## Found card %d at bus %03d   : dev %03d                       ##\n",i + 1,
               adapterList[i].bus, adapterList[i].devFunc >> 3);

        if ((deviceId == 0x5081) ||
            (deviceId == 0x5041))
        {
            unsigned long regVal;
            /* enabled flash controller clock */
            regVal = read32bitMemBar (adapterList[i].bar + 
                                      MV_PCI_EXPANSION_ROM_CONTROL_REG_OFFSET);
            regVal |= 0x1;
            write32bitMemBar (adapterList[i].bar + 
                              MV_PCI_EXPANSION_ROM_CONTROL_REG_OFFSET,regVal);
            /* Make sure the register is written */
            regVal = read32bitMemBar (adapterList[i].bar + 0xd2c);
            /* Set the GPIO mode to flash interface */
            write32bitMemBar (adapterList[i].bar +
                              MV_FLASH_GPIO_PORT_CONTROL_OFFSET, 0x2);
        }
        if ((deviceId == 0x6081) ||
            (deviceId == 0x6041))
        {
            unsigned long regVal;
            /* Enable flash */
            write32bitIoBar (adapterList[i].bar, 0x000104f0, 0x00020062);
            /* Enable BAR2 mapping to flash */
            regVal = read32bitIoBar (adapterList[i].bar, 0xc08);
            regVal &= 0xffffffc0UL;
            regVal |= 0x0000001fUL;
            write32bitIoBar (adapterList[i].bar, 0xc08, regVal);
        }


        if (!useIOBar)
        {
            /* Switch between BAR0 and EXP ROM BAR */
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x10, 0);
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x12, 0x1400);
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x30, (adapterList[i].bar & 0xffff) | 0x1);
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x32, ((adapterList[i].bar >> 16) & 0xffff));
        }

        /* Check if previous BIOS code exists */
        if (read16bit (adapterList[i].bar, 0) == 0xaa55)
        {
            unsigned char buff[50];
            int temp;
            for (temp = 0x60 ; temp <= 0x76 ; temp ++)
            {
                buff[temp - 0x60] = read8bit(adapterList[i].bar, temp);
            }
            if (strcmp(buff, "Marvell Semiconductors"))
            {
                adapterList[i].version = 0xff;
                adapterList[i].subVersion = 0xff;
            } else
            {
                /* Valid exp. ROM - read the old version which is in bytes 0x40 */
                /* and 0x41 of the image */
                adapterList[i].version = read8bit (adapterList[i].bar, 0x40);
                adapterList[i].subVersion = read8bit (adapterList[i].bar, 0x41);
            }
        } else
        {
            adapterList[i].version = 0xff;
            adapterList[i].subVersion = 0xff;
        }
        if (!useIOBar)
        {
            /* Switch back between BAR0 and EXP ROM BAR */
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x30, 0);
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x32, 0);
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x10, (adapterList[i].bar & 0xffff));
            writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x12, ((adapterList[i].bar >> 16) & 0xffff));
        }

        i++;
    }
    printf("##                                                           ##\n");
    printf("###############################################################\n");

    if (i == 0)
    {
        printf("No Marvell SATA devices found - exiting\n");
        exit(0);
    }
    numOfAdapters = i;

    printf("##                                                           ##\n");
    printf("##  Verifying requested image checksum                       ##\n");
    ch = 0 ;
    imageSize = 0;
    while (1)
    {
        int result = fgetc(romImage);
        if (result == EOF)
        {
            break;
        }
        ch += (unsigned char)result;
        if (imageSize == 0x40)
        {
            imageVersion = result;
        }
        if (imageSize == 0x41)
        {
            imageSubversion = result;
        }
        imageSize ++;
    }
    fclose (romImage);
    if (ch)
    {
        printf("Error - Image <%s> checksum is non-zero\n",argv[1]);
        exit (-1);
    }

    /* Image bigger than 256KByte */
    if (imageSize > 0x00080000)
    {
        printf("Error - Image too big\n");
    }
    printf("##  Checksum OK - Version %d.%d.%d.%d                            ##\n",
           imageVersion/10,imageVersion%10, imageSubversion/10,imageSubversion%10);
    printf("##                                                           ##\n");
    printf("###############################################################\n");
    while (1)
    {
        printf("##                                                           ##\n");
        printf("##  Select Adapter :                                         ##\n");
        for (i = 0 ; i < numOfAdapters ; i ++)
        {
            if ((adapterList[i].version == 0xff) && (adapterList[i].subVersion == 0xff))
            {
                printf("## [%1d] 88SX%04x On Bus %03d, device %03d                       ##\n",
                       i+1, deviceId, adapterList[i].bus,adapterList[i].devFunc>> 3);
            } else
            {
                printf("## [%1d] 88SX%04x On Bus %03d, device %03d (Current Ver. %d.%d.%d.%d)##\n",
                       i+1, deviceId,adapterList[i].bus,adapterList[i].devFunc>> 3,
                       adapterList[i].version / 10,
                       adapterList[i].version % 10,
                       adapterList[i].subVersion / 10,
                       adapterList[i].subVersion % 10);
            }
        }
        printf("##                                                           ##\n");
        printf("###############################################################\n");
        ch = getchar();
        if ((ch >= '0') && (ch <= '9'))
        {
            ch = ch - '0';
        } else
        {
            printf("Press a valid adapter number\n");
            continue;
        }
        if (ch > numOfAdapters)
        {
            printf("Press a valid adapter number\n");
            continue;
        } else
        {
            printf("Choosen adapter %d\n",ch);
            break;
        }
    }
    i = ch - 1;

    if (!useIOBar)
    {
        /* Switch between BAR0 and EXP ROM BAR */
        writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x10, 0);
        writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x12, 0x1400);
        writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x30, (adapterList[i].bar & 0xffff) | 0x1);
        writeConfig16bit(adapterList[i].bus, adapterList[i].devFunc, 0x32, ((adapterList[i].bar >> 16) & 0xffff));
    }
    printf("###############################################################\n");
    printf("##                                                           ##\n");
    flashSize = mvFlashInit (adapterList[i].bar , 1, PURE8);
    printf("##  Detected flash size 0x%08lx                           ##\n",flashSize);
    if (!flashSize)
    {
        printf("Error - flash size detected is 0\n");
        switchBackBetweenBar0andExpRom (adapterList[i].bus, adapterList[i].devFunc,
                                        adapterList[i].bar);
        exit (-1);
    }
    if (flashSize > 0x00080000)
    {
        printf("Error - Flash size too big\n");
    }
    if (flashSize < imageSize)
    {
        printf("Error - Image size (%08lx bytes) is bigger than the flash size (%08lx bytes)\n",
               imageSize, flashSize);
    }

    sectorsToErase = 0;
    while (1)
    {
        /* TODO - Rabeeh - Check if sector reached last sector - if so erase all flash */
        if (mvFlashGetSectorOffset(sectorsToErase + 1) > imageSize)
        {
            break;
        }
        sectorsToErase ++;

    }

    printf("##  Erasing %2d sector(s)                                     ##\n",
           sectorsToErase + 1);
    while (1)
    {
        if (mvFlashEraseSector(sectorsToErase) == false)
        {
            printf("Error erasing sector %d\n",sectorsToErase);
            switchBackBetweenBar0andExpRom (adapterList[i].bus, adapterList[i].devFunc,
                                            adapterList[i].bar);
            exit (-1);
        }
        if (sectorsToErase == 0)
        {
            break;
        }
        sectorsToErase --;
    }
    printf("##  Writing image     ");
    romImage = fopen (argv[1], "rob");
    if (!romImage)
    {
        printf("Error reopening image %s\n",argv[1]);
        switchBackBetweenBar0andExpRom (adapterList[i].bus, adapterList[i].devFunc,
                                        adapterList[i].bar);
        exit (-1);
    }
    offset = 0;
    while (1)
    {
        int result = fgetc(romImage);
        static int progress = 0;
        if (result == EOF)
        {
            break;
        }
        if (mvFlashWriteChar(offset, result) == false)
        {
            break;
        }
        offset ++;
        imageSize --;
        progress ++;
        if (!(progress % 1000))
        {
            progress = 0;
            printf(".");
            fflush (stdout);
        }
    }
    for (temp = 39 ; temp > (offset / 1000) ; temp --)
    {
        printf(" ");
    }
    printf("##\n");
    if (imageSize)
    {
        printf("Error - while writing image - couldn't write whole image (%d bytes left)\n",
               imageSize);
        switchBackBetweenBar0andExpRom (adapterList[i].bus, adapterList[i].devFunc,
                                        adapterList[i].bar);
        exit (-1);
    }
    fclose (romImage);

    for (imageSize = 0 ; imageSize < 512 ; imageSize ++)
    {
        mvFlashWriteChar(offset, 0);
        offset ++;
    }

    printf("##  Verifying image   ");
    romImage = fopen (argv[1], "rob");
    if (!romImage)
    {
        printf("\nError reopening image %s\n",argv[1]);
        switchBackBetweenBar0andExpRom (adapterList[i].bus, adapterList[i].devFunc,
                                        adapterList[i].bar);
        exit (-1);
    }
    offset = 0;
    ch = 0;
    while (1)
    {
        int result = fgetc(romImage);
        static int progress = 0;
        if (result == EOF)
        {
            break;
        }
        if (mvFlashReadChar(offset) != result)
        {
            printf("\nError - images are not identical (offset 0x%08lx)\n",offset);
            break;
        }
        ch += mvFlashReadChar(offset);
        offset ++;
        progress ++;
        if (!(progress % 1000))
        {
            progress = 0;
            printf(".");
            fflush (stdout);
        }
    }
    for (temp = 39 ; temp > (offset / 1000) ; temp --)
    {
        printf(" ");
    }
    printf("##\n");
    if (ch)
    {
        printf("\nError - Image on flash checksum is not zero\n");
    }
    fclose (romImage);
    printf("##                                                           ##\n");
    printf("##   *****  Flash Upgrade Finished Successfully  *****       ##\n");
    printf("##                                                           ##\n");
    printf("###############################################################\n");
    /* Switch between BAR0 and EXP ROM BAR */
    switchBackBetweenBar0andExpRom (adapterList[i].bus, adapterList[i].devFunc,
                                    adapterList[i].bar);

}