/*******************************************************************************
Copyright (C) Marvell International Ltd. and its affiliates

This software file (the "File") is owned and distributed by Marvell 
International Ltd. and/or its affiliates ("Marvell") under the following
alternative licensing terms.  Once you have made an election to distribute the
File under one of the following license alternatives, please (i) delete this
introductory statement regarding license alternatives, (ii) delete the two
license alternatives that you have not elected to use and (iii) preserve the
Marvell copyright notice above.

********************************************************************************
Marvell Commercial License Option

If you received this File from Marvell and you have entered into a commercial
license agreement (a "Commercial License") with Marvell, the File is licensed
to you under the terms of the applicable Commercial License.

********************************************************************************
Marvell GPL License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File in accordance with the terms and conditions of the General 
Public License Version 2, June 1991 (the "GPL License"), a copy of which is 
available along with the File in the license.txt file or by writing to the Free 
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 or 
on the worldwide web at http://www.gnu.org/licenses/gpl.txt. 

THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE IMPLIED 
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY 
DISCLAIMED.  The GPL License provides additional details about this warranty 
disclaimer.
********************************************************************************
Marvell BSD License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File under the following licensing terms. 
Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

    *   Redistributions of source code must retain the above copyright notice,
	    this list of conditions and the following disclaimer. 

    *   Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution. 

    *   Neither the name of Marvell nor the names of its contributors may be 
        used to endorse or promote products derived from this software without 
        specific prior written permission. 
    
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/
/*******************************************************************************
* mvOsWin2k.h - O.S. interface header file for Win2k
*
* DESCRIPTION:
*       This header file contains OS dependent definition under Win2k
*
* DEPENDENCIES:
*       None.
*
*******************************************************************************/

#ifndef __INCmvOsWin2kh
#define __INCmvOsWin2kh

/* Includes */
#include "miniport.h"
#include "scsi.h"
#include <stdio.h>
#include <stdarg.h>

/* Definitions */

#define MV_DEFAULT_QUEUE_DEPTH 2
#define MV_SATA_SUPPORT_GEN2E_128_QUEUE_LEN
#define MV_SATA_STORE_COMMANDS_INFO_ON_IAL_STACK
#define MV_SATA_SUPPORT_EDMA_SINGLE_DATA_REGION

/* READ Ahead enable */
#define ENABLE_READ_AHEAD
/* Force non usage of CTQ */
#undef FORCE_NONE_QUEUED_MODE

/* IO Control interface (for GUI) */
#define IOCTL

/* Perform interrupt service in DPC level */
#undef DPC_INT


/* When drive is connected, the following defines how much time (and if) the
    initialzation of the drive should start */
#define DELAY_SUDDEN_CONNECTION
#define DELAY_SUDDEN_CONNECTION_TIME (10*1000000) /*10 seconds*/

/* System dependent macro for flushing CPU write cache */
#define MV_CPU_WRITE_BUFFER_FLUSH()

/* System dependent little endian from / to CPU conversions */
#define MV_CPU_TO_LE16(x)   (x)
#define MV_CPU_TO_LE32(x)   (x)

#define MV_LE16_TO_CPU(x)   (x)
#define MV_LE32_TO_CPU(x)   (x)

/* System dependent register read / write in byte/word/dword variants */

#define MV_REG_WRITE_BYTE(base, offset, val)    \
    ScsiPortWriteRegisterUchar(((PUCHAR)base) + offset, (UCHAR)(val))
#define MV_REG_WRITE_WORD(base, offset, val)    \
    ScsiPortWriteRegisterUshort((PUSHORT)((PUCHAR)(base) + offset), MV_CPU_TO_LE16((USHORT)(val)))
#define MV_REG_WRITE_DWORD(base, offset, val)    \
    ScsiPortWriteRegisterUlong((PULONG)((PUCHAR)(base) + offset), MV_CPU_TO_LE32((ULONG)(val)))

#define MV_REG_READ_BYTE(base, offset)  ScsiPortReadRegisterUchar(((PUCHAR)base) + offset)
#define MV_REG_READ_WORD(base, offset)  ScsiPortReadRegisterUshort((PUSHORT)((PUCHAR)(base) + offset))
#define MV_REG_READ_DWORD(base, offset) ScsiPortReadRegisterUlong((PULONG)((PUCHAR)(base) + offset))

/* Typedefs    */
typedef UCHAR       MV_BOOLEAN;
#define MV_FALSE        0
#define MV_TRUE         1
/* System dependant typedefs */
typedef VOID            MV_VOID, *MV_VOID_PTR;
typedef UCHAR           MV_U8,   *MV_U8_PTR;
typedef USHORT          MV_U16,  *MV_U16_PTR;
typedef ULONG           MV_U32,  *MV_U32_PTR;
typedef PCHAR           MV_CHAR_PTR;
typedef PVOID           MV_BUS_ADDR_T;

/* Structures  */
/* System dependent structure */
typedef struct mvOsSemaphore
{
    int     notUsed;
} MV_OS_SEMAPHORE;


/* Functions (User implemented)*/

/* Semaphore init, take and release */
#define mvOsSemInit(p)      (MV_TRUE)
#define mvOsSemTake(p)      (MV_TRUE)
#define mvOsSemRelease(p)   (MV_TRUE)

/* Delay function in micro seconds resolution */

void mvMicroSecondsDelay(MV_VOID_PTR,  MV_U32);

/* System logging function */
#include "mvLog.h"

/* Enable READ/WRITE Long SCSI command only when driver is compiled for debugging */
#ifdef MV_LOGGER
    #define MV_SATA_SUPPORT_READ_WRITE_LONG
#endif

#define MV_IAL_LOG_ID       3

#endif /* __INCmvOsWin2kh */

