/*******************************************************************************
Copyright (C) Marvell International Ltd. and its affiliates

This software file (the "File") is owned and distributed by Marvell 
International Ltd. and/or its affiliates ("Marvell") under the following
alternative licensing terms.  Once you have made an election to distribute the
File under one of the following license alternatives, please (i) delete this
introductory statement regarding license alternatives, (ii) delete the two
license alternatives that you have not elected to use and (iii) preserve the
Marvell copyright notice above.

********************************************************************************
Marvell Commercial License Option

If you received this File from Marvell and you have entered into a commercial
license agreement (a "Commercial License") with Marvell, the File is licensed
to you under the terms of the applicable Commercial License.

********************************************************************************
Marvell GPL License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File in accordance with the terms and conditions of the General 
Public License Version 2, June 1991 (the "GPL License"), a copy of which is 
available along with the File in the license.txt file or by writing to the Free 
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 or 
on the worldwide web at http://www.gnu.org/licenses/gpl.txt. 

THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE IMPLIED 
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY 
DISCLAIMED.  The GPL License provides additional details about this warranty 
disclaimer.
********************************************************************************
Marvell BSD License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File under the following licensing terms. 
Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

    *   Redistributions of source code must retain the above copyright notice,
	    this list of conditions and the following disclaimer. 

    *   Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution. 

    *   Neither the name of Marvell nor the names of its contributors may be 
        used to endorse or promote products derived from this software without 
        specific prior written permission. 
    
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/
/*******************************************************************************
* mvWin2kScsiMiniPort.c - C File for implementation of the Scsi MiniPort fot
*   win2k.
*
* DESCRIPTION:
*       None.
*
* DEPENDENCIES:
*   miniport.h
*   mvOs.h
*   mvSata.h.
*   mvStorageDev.h
*   mvRegs.h
*
*******************************************************************************/

/* includes */
#include "excpt.h"
#include "initguid.h"
#include "mvWin2kScsiMiniPort.h"
#include "..\CoreDriver\mvStorageDev.h"
#include "mvIALCommon.h"
#include "miniport.h"
#include "devioctl.h"
#include "ntdddisk.h"
#include "ntddscsi.h"
#include <stdio.h>

#ifdef IOCTL

    #include "mvSataUserApi.h"

typedef struct
{
    SRB_IO_CONTROL sic;
    UCHAR          ucDataBuffer[512];
} SRB_BUFFER, *PSRB_BUFFER;

#endif

#ifdef MV_LOGGER
char *szModules[] = {"Core Driver",
"SAL",
"Common IAL",
"Windows IAL"
};
extern char* mvLogMsgType[MV_MAX_MESSAGE_TYPE];
#endif
VOID myKeQuerySystemTime (PLARGE_INTEGER pSystemTime);

#ifdef DPC_INT
BOOLEAN
HwScsiEnableInterruptsCallback(
                              IN PVOID DeviceExtension
                              );
BOOLEAN
HwScsiDisableInterruptsCallback(
                               IN PVOID DeviceExtension
                               );
#endif



/* Bits for HD_ERROR */
#define NM_ERR          0x02    /* media present */
#define ABRT_ERR        0x04    /* Command aborted */
#define MCR_ERR         0x08    /* media change request */
#define IDNF_ERR        0x10    /* ID field not found */
#define MC_ERR          0x20    /* media changed */
#define UNC_ERR         0x40    /* Uncorrect data */
#define WP_ERR          0x40    /* write protect */
#define ICRC_ERR        0x80    /* new meaning:  CRC error during transfer */


#ifdef MV_LOGGER
    #define MV_IAL_DEBUG_CDB            0x1
MV_U8   ialDebugLevel = 0;
#endif

BOOLEAN mvSataInitializedForDump = FALSE;

BOOLEAN
initAdapterExtension( IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                      IN PCHAR ArgumentString);

BOOLEAN
setAdapterConfigSpaceParams( IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                             IN PPORT_CONFIGURATION_INFORMATION ConfigInfo);

BOOLEAN
AllocateEdmaQueues(
                  IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                  IN PPORT_CONFIGURATION_INFORMATION ConfigInfo
                  );

VOID
InitChannel(
           IN PHW_DEVICE_EXTENSION HwDeviceExtension,
           IN UCHAR channelIndex
           );

VOID
HwScsiTimer(
           IN PVOID DeviceExtension
           );
BOOLEAN
ResetSataChannel(
                IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                UCHAR   channelIndex
                );

BOOLEAN
sendScsiCommand(
               IN PHW_DEVICE_EXTENSION HwDeviceExtension,
               IN PSCSI_REQUEST_BLOCK Srb
               );

MV_BOOLEAN
IOControlIdentifyDeviceCommandCompletionCB(MV_SATA_ADAPTER *pSataAdapter,
                                           MV_U8 channelNum,
                                           MV_COMPLETION_TYPE comp_type,
                                           void *commandId,
                                           MV_U16 responseFlags,
                                           MV_U32 timeStamp,
                                           MV_STORAGE_DEVICE_REGISTERS *registerStruct);

MV_BOOLEAN SmartATACommandCompletionCB(MV_SATA_ADAPTER *pSataAdapter,
                                       MV_U8 channelNum,
                                       MV_COMPLETION_TYPE comp_type,
                                       MV_VOID_PTR commandId,
                                       MV_U16 responseFlags,
                                       MV_U32 timeStamp,
                                       MV_STORAGE_DEVICE_REGISTERS *registerStruct);



MV_BOOLEAN
CommandCompletionCB(MV_SATA_ADAPTER *pSataAdapter,
                    MV_U8 channelNum,
                    MV_COMPLETION_TYPE comp_type,
                    void *commandId,
                    MV_U16 responseFlags,
                    MV_U32 timeStamp,
                    MV_STORAGE_DEVICE_REGISTERS *registerStruct);


/* Functions that start  are entry points           */
/* for the OS port driver.                          */

ULONG
DriverEntry(
           IN PVOID DriverObject,
           IN PVOID Argument2
           );

ULONG
mvFindAdapter(
             IN PVOID HwDeviceExtension,
             IN PVOID Context,
             IN PVOID BusInformation,
             IN PCHAR ArgumentString,
             IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo,
             OUT PBOOLEAN Again
             );

#ifdef MV_WINDOWS_2000
BOOLEAN
mvAdapterState(
              IN PVOID HwDeviceExtension,
              IN PVOID Context,
              IN BOOLEAN SaveState
              );
#endif

BOOLEAN
mvHwInitialize(
              IN PVOID DeviceExtension
              );


BOOLEAN
mvStartIo(
         IN PVOID DeviceExtension,
         IN PSCSI_REQUEST_BLOCK Srb
         );

BOOLEAN
mvInterrupt(
           IN PVOID DeviceExtension
           );

BOOLEAN
mvResetBus(
          IN PVOID HwDeviceExtension,
          IN ULONG PathId
          );
#ifdef MV_WINDOWS_2000
SCSI_ADAPTER_CONTROL_STATUS
mvAdapterControl(
                IN PVOID HwDeviceExtension,
                IN SCSI_ADAPTER_CONTROL_TYPE ControlType,
                IN PVOID Parameters
                );
#endif
MV_BOOLEAN IALCompletion(struct mvSataAdapter *pSataAdapter,
                         MV_SATA_SCSI_CMD_BLOCK *pCmdBlock);

MV_BOOLEAN canQueueCmnds(PHW_DEVICE_EXTENSION HwDeviceExtension,
                         MV_SATA_ADAPTER *pSataAdapter,
                         MV_U8  channelIndex, MV_U8 PMPort);

BOOLEAN buildPRDTable(PHW_DEVICE_EXTENSION HwDeviceExtension, PSCSI_REQUEST_BLOCK Srb);

MV_BOOLEAN canQueueCmnds(PHW_DEVICE_EXTENSION HwDeviceExtension,
                         MV_SATA_ADAPTER *pSataAdapter,
                         MV_U8  channelIndex, MV_U8 PMPort)
{
    MV_U8   commandsPerChannel;
    if (mvSataGetNumOfPortQueuedCommands(pSataAdapter, channelIndex, PMPort,
                                         &commandsPerChannel) <
        HwDeviceExtension->channelsInfo[channelIndex].queueDepth)
    {
        if (commandsPerChannel < HwDeviceExtension->channelsInfo[channelIndex].channelQueueDepth)
        {
            return MV_TRUE;
        }
    }
    return MV_FALSE;
}
#ifdef MV_LOGGER
VOID
reportCDBCompletion(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                    PSCSI_REQUEST_BLOCK Srb);
#endif

#ifdef MV_LOGGER
VOID
reportCDBCompletion(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                    PSCSI_REQUEST_BLOCK Srb)
{
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    if ((ialDebugLevel & MV_IAL_DEBUG_CDB) ||
        (mvLogGetModuleFilter(MV_IAL_LOG_ID) & MV_DEBUG))
    {
        if (Srb->Function == SRB_FUNCTION_EXECUTE_SCSI)
        {
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "------\\/----\\/---Completion Info-----\n");
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Adapter %d: SCSI cmd completed (PathId %d, TargetId %d, LUN %d, length = %d)\n",
                     pSataAdapter->adapterId, Srb->PathId, Srb->TargetId, Srb->Lun, Srb->Length);
            if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_SUCCESS)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_SUCCESS\n");
            }
            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_ERROR)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_ERROR\n");
            }
            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_INVALID_TARGET_ID)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_INVALID_TARGET_ID\n");
            }
            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_INVALID_LUN)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_INVALID_LUN\n");
            }
            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_ABORTED)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_ABORTED\n");
            }
            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_NO_DEVICE)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_NO_DEVICE\n");
            }
            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_INVALID_REQUEST)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_INVALID_REQUEST\n");
            }

            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_DATA_OVERRUN)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_DATA_OVERRUN\n");
            }
            else if ((Srb->SrbStatus & 0x3f) == SRB_STATUS_BUS_RESET)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with SRB_STATUS_BUS_RESET\n");
            }
            else
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Command completed with status 0x%x\n",Srb->SrbStatus);
            }

            if (Srb->ScsiStatus == SCSISTAT_CHECK_CONDITION)
            {
                MV_U8 i;
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Scsi status is SCSISTAT_CHECK_CONDITION - dumping sense info buffer\n");
                for (i = 0 ; i < Srb->SenseInfoBufferLength ; i++)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- %02x\n", ((PUCHAR)Srb->SenseInfoBuffer)[i]);
                }
            }
            if ((Srb->Cdb[0] == SCSIOP_MODE_SENSE) ||
                (Srb->Cdb[0] == SCSIOP_INQUIRY) ||
                (Srb->Cdb[0] == SCSIOP_READ_CAPACITY) ||
                (Srb->Cdb[0] == SCSIOP_REQUEST_SENSE))
            {
                MV_U32 i;
                if (Srb->Cdb[0] == SCSIOP_MODE_SENSE)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Dumping MODE SENSE result (returning %d bytes)\n",
                             Srb->DataTransferLength);
                }
                else if (Srb->Cdb[0] == SCSIOP_INQUIRY)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Dumping INQUIRY result (returning %d bytes)\n",
                             Srb->DataTransferLength);
                }
                else if (Srb->Cdb[0] == SCSIOP_READ_CAPACITY)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Dumping READ CAPACITY result (returning %d bytes)\n",
                             Srb->DataTransferLength);
                }
                else if (Srb->Cdb[0] == SCSIOP_REQUEST_SENSE)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Dumping REQUEST SENSE result (returning %d bytes)\n",
                             Srb->DataTransferLength);
                }
                else
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "---- Dumping command %02x result (returning %d bytes)\n",
                             Srb->Cdb[0],Srb->DataTransferLength);
                }

                for (i = 0 ; i < Srb->DataTransferLength ; i += 8)
                {
                    char buffer[256];
                    int len = 0, j;
                    len += sprintf (&buffer[len], "---- %03d :",i);
                    for (j = 0 ; j < 8 ; j++)
                    {
                        if ((i + j) < Srb->DataTransferLength)
                        {
                            len += sprintf (&buffer[len], " %02x",
                                            ((UCHAR *)Srb->DataBuffer)[i + j]);
                        }
                    }

                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   buffer);
                }
            }

            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "------/\\----/\\---Completion Info-----\n");
        }
    }

}
#endif



#ifdef IOCTL


BOOLEAN mvAdapterSmartATACommand(IN MV_SATA_ADAPTER     *pSataAdapter,
                                 IN MV_SATA_CHANNEL      *pSataChannel,
                                 IN MV_U8 pmPort,
                                 IN PSCSI_REQUEST_BLOCK Srb,
                                 IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                                 IN PSENDCMDINPARAMS cmdInParams,
                                 IN MV_NON_UDMA_PROTOCOL protocolType,
                                 IN ULONG  buffersize,
                                 IN MV_U8 channelIndex)
{

#ifdef MV_SATA_STORE_COMMANDS_INFO_ON_IAL_STACK
    MV_QUEUE_COMMAND_INFO   *pCommandInfo = &((PSRB_EXTENSION)Srb->SrbExtension)->commandInfo;
#else
    MV_QUEUE_COMMAND_INFO   commandInfo;
    MV_QUEUE_COMMAND_INFO   *pCommandInfo = &commandInfo;
#endif
    MV_QUEUE_COMMAND_RESULT result;
    PSENDCMDOUTPARAMS cmdOutParams = (PSENDCMDOUTPARAMS)cmdInParams;
    ((PSRB_EXTENSION)Srb->SrbExtension)->commandType =
    MV_QUEUED_COMMAND_TYPE_NONE_UDMA;
    memset(pCommandInfo, 0, sizeof(MV_QUEUE_COMMAND_INFO));

    pCommandInfo->type = MV_QUEUED_COMMAND_TYPE_NONE_UDMA;

    pCommandInfo->commandParams.NoneUdmaCommand.protocolType = protocolType;
    pCommandInfo->commandParams.NoneUdmaCommand.isEXT = MV_FALSE;

    if (protocolType == MV_NON_UDMA_PROTOCOL_NON_DATA)
    {

        pCommandInfo->commandParams.NoneUdmaCommand.bufPtr = NULL;
    }
    else
    {
        if (buffersize == 0)
        {
            Srb->SrbStatus = SRB_STATUS_ERROR;
            return FALSE;
        }
        if (protocolType == MV_NON_UDMA_PROTOCOL_PIO_DATA_IN)
        {
            pCommandInfo->commandParams.NoneUdmaCommand.bufPtr =
            (MV_U16*)cmdOutParams->bBuffer;
            cmdOutParams->cBufferSize = buffersize;
        }
        else
        {
            pCommandInfo->commandParams.NoneUdmaCommand.bufPtr =
            (MV_U16*)cmdInParams->bBuffer;
        }
    }
    pCommandInfo->PMPort = pmPort;
    pCommandInfo->commandParams.NoneUdmaCommand.count = buffersize/2;
    pCommandInfo->commandParams.NoneUdmaCommand.features =
    cmdInParams->irDriveRegs.bFeaturesReg;
    pCommandInfo->commandParams.NoneUdmaCommand.sectorCount =
    cmdInParams->irDriveRegs.bSectorCountReg;
    pCommandInfo->commandParams.NoneUdmaCommand.lbaLow =
    cmdInParams->irDriveRegs.bSectorNumberReg;
    pCommandInfo->commandParams.NoneUdmaCommand.lbaMid =
    cmdInParams->irDriveRegs.bCylLowReg;
    pCommandInfo->commandParams.NoneUdmaCommand.lbaHigh =
    cmdInParams->irDriveRegs.bCylHighReg;
    pCommandInfo->commandParams.NoneUdmaCommand.device = (MV_U8)MV_BIT6;
    pCommandInfo->commandParams.NoneUdmaCommand.command =
    cmdInParams->irDriveRegs.bCommandReg;
    pCommandInfo->commandParams.NoneUdmaCommand.callBack =
    SmartATACommandCompletionCB;
    pCommandInfo->commandParams.NoneUdmaCommand.commandId = (MV_VOID_PTR) Srb;

    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  "Sending SMART command: features 0x%02X\n",
             pCommandInfo->commandParams.NoneUdmaCommand.features);
    __try
    {
        result = mvSataQueueCommand(pSataAdapter, pSataChannel->channelNumber,
                                    pCommandInfo);
    }
    __except (EXCEPTION_EXECUTE_HANDLER)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "mvSataQueueCommand raised an exception\n");
        Srb->SrbStatus = SRB_STATUS_ERROR;
        return FALSE;
    }
    if (result != MV_QUEUE_COMMAND_RESULT_OK)
    {
        switch (result)
        {
        case MV_QUEUE_COMMAND_RESULT_BAD_LBA_ADDRESS:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue SMART command failed. Bad LBA "
                     "\n");
            break;
        case MV_QUEUE_COMMAND_RESULT_QUEUED_MODE_DISABLED:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue SMART command failed. EDMA"
                     " disabled adapter %d channel %d\n",
                     pSataAdapter->adapterId, channelIndex);
            break;
        case MV_QUEUE_COMMAND_RESULT_FULL:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue SMART command failed. Queue is"
                     " Full adapter %d channel %d\n",
                     pSataAdapter->adapterId, channelIndex);

            break;
        case MV_QUEUE_COMMAND_RESULT_BAD_PARAMS:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue SMART command failed. (Bad "
                     "Params), pMvSataAdapter: %p,  pSataChannel: %p.\n",
                     pSataAdapter, pSataAdapter->sataChannel[channelIndex]);
            break;
        default:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Bad result value (%d) from queue"
                     " command\n", result);
        }
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "mvSataQueueCommand SMART Failed\n");
        Srb->SrbStatus = SRB_STATUS_ERROR;
        return FALSE;
    }
    else
    {
        ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
    }
    return TRUE;
}

static BOOLEAN hasPMConnected(MV_SATA_ADAPTER *pSataAdapter)
{
    int i;
    for (i = 0; i < pSataAdapter->numberOfChannels; i++)
    {
        if ((pSataAdapter->sataChannel[i] != NULL) &&
            (pSataAdapter->sataChannel[i]->deviceType == MV_SATA_DEVICE_TYPE_PM))
        {
            return TRUE;
        }
    }
    return FALSE;
}

BOOLEAN mvAdapterSmartCommand(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                              IN PSCSI_REQUEST_BLOCK Srb)
{

    PSRB_IO_CONTROL psrbIoControl = (PSRB_IO_CONTROL)Srb->DataBuffer;
    PSENDCMDINPARAMS cmdInParams = (PSENDCMDINPARAMS)
                                   &((PUCHAR)Srb->DataBuffer)[psrbIoControl->HeaderLength];
    PSENDCMDOUTPARAMS cmdOutParams = (PSENDCMDOUTPARAMS)
                                     &((PUCHAR)Srb->DataBuffer)[psrbIoControl->HeaderLength];

    PSRB_EXTENSION      SrbExtension = (PSRB_EXTENSION)Srb->SrbExtension;
    UCHAR               channelIndex = Srb->PathId;
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_SATA_CHANNEL     *pSataChannel;
    BOOLEAN             bRequestQueued = FALSE;
    UCHAR               PathId = Srb->PathId;
    UCHAR               TargetId = Srb->TargetId;
    UCHAR               Lun = Srb->Lun;
    MV_U8               pmPort = TargetId;


    if (psrbIoControl->ControlCode != IOCTL_SCSI_MINIPORT_SMART_VERSION)
    {
        /*if SENDCMDINPARAMS is used in IO control
        and no Port multipier is connected
        use cmdInParams->bDriveNumber as channel index*/
        if (hasPMConnected(pSataAdapter) == FALSE)
        {
            if (cmdInParams->bDriveNumber > channelIndex)
            {
                channelIndex = cmdInParams->bDriveNumber;
            }
        }
    }

    psrbIoControl->ReturnCode = 0xFFFFFFFF;
    if (channelIndex >= MV_SATA_CHANNELS_NUM)
    {
        Srb->SrbStatus = SRB_STATUS_INVALID_PATH_ID;
        ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
        ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
        return TRUE;
    }

    pSataChannel = pSataAdapter->sataChannel[channelIndex];

    if (pSataAdapter->sataChannel[channelIndex] == NULL)
    {
        mvLogMsg(MV_IAL_LOG_ID, MV_DEBUG_ERROR,
                 "[%d %d] Cannot execute SMART command: disk is disconnected.",
                 pSataAdapter->adapterId, channelIndex);
        Srb->SrbStatus = SRB_STATUS_NO_DEVICE;
        ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
        ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
        return TRUE;
    }

    if (MV_FALSE == canQueueCmnds(HwDeviceExtension, pSataAdapter,
                                  channelIndex, pmPort))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "[%d %d] Cannot execute SMART command: "
                 "channel is busy\n", pSataAdapter->adapterId,  channelIndex);

        Srb->SrbStatus = SRB_STATUS_BUSY;
        ScsiPortNotification(RequestComplete, (PVOID)HwDeviceExtension, Srb);
        ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
        return TRUE;
    }

    SrbExtension->HwDeviceExtension = (PVOID)HwDeviceExtension;
    Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
    switch (psrbIoControl->ControlCode)
    {
    case IOCTL_SCSI_MINIPORT_SMART_VERSION:
        {
            PGETVERSIONINPARAMS versionParams = (PGETVERSIONINPARAMS)
                                                &((PUCHAR)Srb->DataBuffer)[psrbIoControl->HeaderLength];
            if (psrbIoControl->Length < sizeof(GETVERSIONINPARAMS))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "IOCTL_SCSI_MINIPORT_SMART_VERSION failed:"
                         " invalid buffer length =%d\n",
                         psrbIoControl->Length);
                break;
            }
            versionParams->bVersion = 1;
            versionParams->bRevision = 0;
            versionParams->bIDEDeviceMap = 0;
            versionParams->fCapabilities = CAP_SMART_CMD | CAP_ATA_ID_CMD;
            Srb->SrbStatus = SRB_STATUS_SUCCESS;
            psrbIoControl->ReturnCode = 0;
        }
        break;
    case IOCTL_SCSI_MINIPORT_ENABLE_SMART:
    case IOCTL_SCSI_MINIPORT_DISABLE_SMART:
    case IOCTL_SCSI_MINIPORT_ENABLE_DISABLE_AUTOSAVE:
    case IOCTL_SCSI_MINIPORT_EXECUTE_OFFLINE_DIAGS:
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_NON_DATA,
                                             0,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
        }
        break;
    case IOCTL_SCSI_MINIPORT_RETURN_STATUS:
        if (psrbIoControl->Length < sizeof(SENDCMDOUTPARAMS)
            + sizeof(IDEREGS) - 1)
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError =
                SMART_INVALID_BUFFER;
            }
            break;
        }
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_NON_DATA,
                                             0,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
            }
        }
        break;
    case IOCTL_SCSI_MINIPORT_SAVE_ATTRIBUTE_VALUES:
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_NON_DATA,
                                             0,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
            }
        }
        break;
    case IOCTL_SCSI_MINIPORT_READ_SMART_ATTRIBS:
        if (psrbIoControl->Length < READ_ATTRIBUTE_BUFFER_SIZE +
            sizeof(SENDCMDOUTPARAMS) - 1)
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError =
                SMART_INVALID_BUFFER;
            }
            break;
        }
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_PIO_DATA_IN,
                                             READ_ATTRIBUTE_BUFFER_SIZE,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
            }
        }
        break;
    case IOCTL_SCSI_MINIPORT_READ_SMART_THRESHOLDS:
        if (psrbIoControl->Length < READ_THRESHOLD_BUFFER_SIZE +
            sizeof(SENDCMDOUTPARAMS) - 1)
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError =
                SMART_INVALID_BUFFER;
            }
            break;
        }
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_PIO_DATA_IN,
                                             READ_THRESHOLD_BUFFER_SIZE,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
            }
        }
        break;
    case IOCTL_SCSI_MINIPORT_IDENTIFY:
        if (psrbIoControl->Length < IDENTIFY_BUFFER_SIZE +
            sizeof(SENDCMDOUTPARAMS) - 1)
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError =
                SMART_INVALID_BUFFER;
            }
            break;
        }
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_PIO_DATA_IN,
                                             IDENTIFY_BUFFER_SIZE,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
            }
        }
        break;

    case IOCTL_SCSI_MINIPORT_ENABLE_DISABLE_AUTO_OFFLINE:
        cmdOutParams->DriverStatus.bDriverError = SMART_NOT_SUPPORTED;
        break;
#ifdef IOCTL_SCSI_MINIPORT_READ_SMART_LOG
    case IOCTL_SCSI_MINIPORT_READ_SMART_LOG:
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_PIO_DATA_IN,
                                             cmdInParams->irDriveRegs.bSectorCountReg *
                                             SMART_LOG_SECTOR_SIZE,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
            }
        }
        break;
#endif

#ifdef IOCTL_SCSI_MINIPORT_WRITE_SMART_LOG
    case IOCTL_SCSI_MINIPORT_WRITE_SMART_LOG:
        if (TRUE == mvAdapterSmartATACommand(pSataAdapter,
                                             pSataChannel,
                                             pmPort,
                                             Srb,
                                             HwDeviceExtension,
                                             cmdInParams,
                                             MV_NON_UDMA_PROTOCOL_PIO_DATA_OUT,
                                             cmdInParams->irDriveRegs.bSectorCountReg *
                                             SMART_LOG_SECTOR_SIZE,
                                             channelIndex))
        {
            bRequestQueued = TRUE;
        }
        else
        {
            if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            {
                cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
            }
        }
        break;
#endif
    default:
        if (psrbIoControl->Length >= sizeof(SENDCMDOUTPARAMS))
            cmdOutParams->DriverStatus.bDriverError = SMART_INVALID_COMMAND;
        channelIndex = Srb->PathId;
        break;
    }

    if (bRequestQueued == FALSE)
    {
#ifdef MV_DEBUG_IAL
        reportCDBCompletion(HwDeviceExtension,Srb);
#endif

        ScsiPortNotification(RequestComplete,
                             HwDeviceExtension,
                             Srb);
        ScsiPortNotification(NextRequest,
                             HwDeviceExtension,
                             NULL);
    }
    return TRUE;
}



BOOLEAN mvIoctlProcessAdapterMessage(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                                     IN PSCSI_REQUEST_BLOCK Srb)
{
    PSRB_IO_CONTROL psrbIoControl = (PSRB_IO_CONTROL)Srb->DataBuffer;
    MV_SATA_ADAPTER   *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_U8   *dataPtr = &((MV_U8*)(Srb->DataBuffer))[sizeof(SRB_IO_CONTROL)];
    MV_U32 controlCode;


    Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
    psrbIoControl->ReturnCode = 0xFFFFFFFF;

    if (psrbIoControl->Length < sizeof(MV_U32))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "[%d] mvIoctlProcessAdapterMessage(): invalid buffer length.\n",
                 pSataAdapter->adapterId);
        goto finishAdapterMessage;
    }
    controlCode = *(MV_U32*)dataPtr;

    switch (controlCode)
    {
    case MV_ADAPTER_MSG_READ_REGISTER:
        {
            MV_ADAPTER_MSG_READ_REGISTER_STRUCT* rrStruct =
            (MV_ADAPTER_MSG_READ_REGISTER_STRUCT*)dataPtr;
            if (psrbIoControl->Length < sizeof(MV_ADAPTER_MSG_READ_REGISTER_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessAdapterMessage"
                         "(MV_ADAPTER_MSG_READ_REGISTER): invalid buffer length.\n",
                         pSataAdapter->adapterId);
                break;
            }
            if (rrStruct->regAddr & 0x3)
            {
                break;
            }
            rrStruct->result = mvSataReadReg(pSataAdapter,
                                             rrStruct->regAddr);
            Srb->SrbStatus = SRB_STATUS_SUCCESS;
            psrbIoControl->ReturnCode = 0;
            break;
        }
    case MV_ADAPTER_MSG_WRITE_REGISTER:
        {
            MV_ADAPTER_MSG_WRITE_REGISTER_STRUCT* wrStruct =
            (MV_ADAPTER_MSG_WRITE_REGISTER_STRUCT*)dataPtr;
            if (psrbIoControl->Length < sizeof(MV_ADAPTER_MSG_WRITE_REGISTER_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessAdapterMessage"
                         "(MV_ADAPTER_MSG_WRITE_REGISTER): invalid buffer length.\n",
                         pSataAdapter->adapterId);

                break;
            }
            if (wrStruct->regAddr & 0x3)
            {
                break;
            }
            mvSataWriteReg (pSataAdapter, wrStruct->regAddr, wrStruct->result);
            Srb->SrbStatus = SRB_STATUS_SUCCESS;
            psrbIoControl->ReturnCode = 0;
            break;
        }
    }

    finishAdapterMessage:
#ifdef MV_DEBUG_IAL
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
    ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
    return TRUE;
}

BOOLEAN mvIoctlProcessDriverMessage(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                                    IN PSCSI_REQUEST_BLOCK Srb)
{

    PSRB_IO_CONTROL psrbIoControl = (PSRB_IO_CONTROL)Srb->DataBuffer;
    MV_SATA_ADAPTER   *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_U8   *dataPtr = &((MV_U8*)(Srb->DataBuffer))[sizeof(SRB_IO_CONTROL)];
    MV_U32 controlCode;

    Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
    psrbIoControl->ReturnCode = 0xFFFFFFFF;

    if (psrbIoControl->Length < sizeof(MV_U32))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "[%d] mvIoctlProcessDriverMessage():"
                 " invalid buffer length.\n",
                 pSataAdapter->adapterId);
        goto finishDriverMessage;
    }
    controlCode = *(MV_U32*)dataPtr;
    switch (controlCode)
    {
    case MV_DRIVER_MSG_GET_STATS:
        {
            MV_U8 channelIndex, temp;
            MV_DRIVER_MSG_GET_STATS_STRUCT* stats =
            (MV_DRIVER_MSG_GET_STATS_STRUCT*) dataPtr;
            LARGE_INTEGER systemTime;

            if (psrbIoControl->Length < sizeof(MV_DRIVER_MSG_GET_STATS_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessDriverMessage"
                         "(MV_DRIVER_MSG_GET_STATS):"
                         " invalid buffer length.\n",
                         pSataAdapter->adapterId);
                break;
            }
            myKeQuerySystemTime (&systemTime);
            /* Convert to mSec */
            systemTime.QuadPart /= 10000;
            stats->currentTimeLo = systemTime.LowPart;
            stats->currentTimeHi = systemTime.HighPart;
            for (channelIndex = 0 ; channelIndex < 8 ; channelIndex ++)
            {
                stats->channelStats[channelIndex].pmVector = 0x0;
                if (HwDeviceExtension->mvSataAdapter.sataChannel[channelIndex])
                {
                    stats->channelStats[channelIndex].totalAccumulatedOutstanding =
                    HwDeviceExtension->ataScsiAdapterExt.totalAccumulatedOutstanding[channelIndex];
                    stats->channelStats[channelIndex].channelIndex = channelIndex;
                    if (mvStorageDevGetDeviceType (&HwDeviceExtension->mvSataAdapter,
                                                   channelIndex) == MV_SATA_DEVICE_TYPE_PM)
                    {
                        stats->channelStats[channelIndex].pmVector |= (1<<15);
                    }
                    for (temp = 0 ; temp < MV_SATA_PM_MAX_PORTS ; temp++)
                    {
                        MV_SATA_SCSI_DRIVE_DATA *driveData =
                        &HwDeviceExtension->ataScsiAdapterExt.ataDriveData[channelIndex][temp];
                        if (driveData->driveReady == MV_TRUE)
                        {
                            stats->channelStats[channelIndex].pmVector |= (1 << temp);
                            stats->channelStats[channelIndex].totalIosPerPmPort[temp] =
                            driveData->stats.totalIOs;
                            stats->channelStats[channelIndex].totalSectorsPerPmPort[temp] =
                            driveData->stats.totalSectorsTransferred;

                        }
                        else
                        {

                            stats->channelStats[channelIndex].totalIosPerPmPort[temp] = 0;
                            stats->channelStats[channelIndex].totalSectorsPerPmPort[temp] = 0;
                        }
                    }

                }
                else
                {
                    stats->channelStats[channelIndex].channelIndex = 0xffffffff; /* Channel not connected */
                }
            }
            psrbIoControl->ReturnCode = 0;
            Srb->SrbStatus = SRB_STATUS_SUCCESS;

            break;
        }
    case MV_DRIVER_MSG_GET_PCI_ATTR:
        {
            MV_U8 channelIndex;
            MV_SATA_DEVICE_TYPE devType;
            MV_DRIVER_MSG_GET_PCI_ATTR_STRUCT* pciAttr =
            (MV_DRIVER_MSG_GET_PCI_ATTR_STRUCT*)dataPtr;
            if (psrbIoControl->Length < sizeof(MV_DRIVER_MSG_GET_PCI_ATTR_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessDriverMessage"
                         "(MV_DRIVER_MSG_GET_PCI_ATTR):"
                         " invalid buffer length.\n",
                         pSataAdapter->adapterId);
                break;
            }
            pciAttr->pciBusNum = (HwDeviceExtension->adapterBusNumber & 0xff);
            pciAttr->pciDevFuncNum = HwDeviceExtension->adapterDeviceNumber;
            pciAttr->COREDriverInstanceId = HwDeviceExtension->mvSataAdapter.adapterId;
            pciAttr->deviceId = HwDeviceExtension->mvSataAdapter.pciConfigDeviceId;
            pciAttr->revId = HwDeviceExtension->mvSataAdapter.pciConfigRevisionId;
            pciAttr->baseAddress = HwDeviceExtension->BAR0 ;
             
            for (channelIndex = 0 ; channelIndex < MV_SATA_CHANNELS_NUM ; channelIndex ++)
            {
                if (HwDeviceExtension->mvSataAdapter.sataChannel[channelIndex])
                {

                    devType = mvStorageDevGetDeviceType (&HwDeviceExtension->mvSataAdapter,
                                                         channelIndex);
                }
                else
                {
                    devType = MV_SATA_DEVICE_TYPE_UNKNOWN;
                }
                if (devType == MV_SATA_DEVICE_TYPE_UNKNOWN) pciAttr->attachedDrives[channelIndex] = 0;
                if (devType == MV_SATA_DEVICE_TYPE_ATA_DISK) pciAttr->attachedDrives[channelIndex] = 1;
                if (devType == MV_SATA_DEVICE_TYPE_ATAPI_DISK) pciAttr->attachedDrives[channelIndex] = 2;
                if (devType == MV_SATA_DEVICE_TYPE_PM) pciAttr->attachedDrives[channelIndex] = 3;
                pciAttr->queueDepth[channelIndex] = HwDeviceExtension->channelsInfo[channelIndex].queueDepth;
            }
            psrbIoControl->ReturnCode = 0;
            Srb->SrbStatus = SRB_STATUS_SUCCESS;
            break;
        }
    case MV_DRIVER_MSG_GET_DEBUG_FILTERS:
        {
            MV_U8 count;
            MV_DRIVER_MSG_GET_DEBUG_FILTERS_STRUCT* debugFilters =
            (MV_DRIVER_MSG_GET_DEBUG_FILTERS_STRUCT*)dataPtr;
            if (psrbIoControl->Length < sizeof(MV_DRIVER_MSG_GET_DEBUG_FILTERS_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessDriverMessage"
                         "(MV_DRIVER_MSG_GET_DEBUG_FILTERS_STRUCT):"
                         " invalid buffer length.\n",
                         pSataAdapter->adapterId);
                break;
            }
#ifdef MV_LOGGER
            debugFilters->numOfDebugLevels = 9;
            debugFilters->numOfDebugeeModules = 4;
            for (count = 0 ; count < debugFilters->numOfDebugeeModules ; count ++)
            {
                strncpy (debugFilters->debugeeModules[count], szModules[count], 39);
                debugFilters->debugLevelsFilterVector[count] =
                mvLogGetModuleFilter (count);
            }
            for (count = 0 ; count < debugFilters->numOfDebugLevels ; count ++)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG, "Pointer %p->%p\n",
                         mvLogMsgType, mvLogMsgType[count]);

                strncpy (debugFilters->debugLevelsNames[count], mvLogMsgType[count], 39);
            }
#else
            debugFilters->numOfDebugLevels = 0;
            debugFilters->numOfDebugeeModules = 0;
#endif
            psrbIoControl->ReturnCode = 0;
            Srb->SrbStatus = SRB_STATUS_SUCCESS;
            break;
        }
    case MV_DRIVER_MSG_SET_MODULE_DEBUG_FILTER:
        {
            MV_DRIVER_MSG_SET_MODULE_DEBUG_FILTER_STRUCT* debugFilter =
            (MV_DRIVER_MSG_SET_MODULE_DEBUG_FILTER_STRUCT*)dataPtr;
            if (psrbIoControl->Length < sizeof(MV_DRIVER_MSG_SET_MODULE_DEBUG_FILTER_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessDriverMessage"
                         "(MV_DRIVER_MSG_SET_MODULE_DEBUG_FILTER):"
                         " invalid buffer length.\n",
                         pSataAdapter->adapterId);
                break;
            }
#ifdef MV_LOGGER
            mvLogMsg(MV_IAL_LOG_ID,MV_DEBUG,"Modifying debug level of module %d to %x\n",
                     debugFilter->moduleNum , debugFilter->filterVector);
            if ((debugFilter->moduleNum < 4) &&
                (mvLogSetModuleFilter((MV_U8)debugFilter->moduleNum,
                                      debugFilter->filterVector) == MV_TRUE))
            {
                psrbIoControl->ReturnCode = 0;
                Srb->SrbStatus = SRB_STATUS_SUCCESS;
            }
#endif
            break;
        }
    case MV_DRIVER_MSG_DISABLE_QUEUING:
        {
            MV_DRIVER_MSG_DISABLE_QUEUING_STRUCT* queuing =
            (MV_DRIVER_MSG_DISABLE_QUEUING_STRUCT*)dataPtr;
            MV_U8 channelIndex;
            if (psrbIoControl->Length < sizeof(MV_DRIVER_MSG_DISABLE_QUEUING_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessDriverMessage"
                         "(MV_DRIVER_MSG_DISABLE_QUEUING):"
                         " invalid buffer length.\n",
                         pSataAdapter->adapterId);
                break;
            }
            if (queuing->channelIndex >= MV_SATA_CHANNELS_NUM)
            {
                Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
                break;
            }
            channelIndex = (MV_U8)queuing->channelIndex;
            mvLogMsg(MV_IAL_LOG_ID,MV_DEBUG,"Channel %d - Modifying queuing mode to %s\n",
                     channelIndex, (queuing->onOff_ == 1) ? "Permitted" : "Inhibited");
            HwDeviceExtension->channelsInfo[channelIndex].forceNonQueued =
            (queuing->onOff_ == 1) ? MV_TRUE:MV_FALSE;
            ResetSataChannel(HwDeviceExtension, channelIndex);
            ScsiPortNotification(ResetDetected, HwDeviceExtension, NULL);
            Srb->SrbStatus = SRB_STATUS_BUS_RESET;
            break;
        }
    case MV_DRIVER_MSG_MODIFY_QUEUE_DEPTH:
        {
            MV_DRIVER_MSG_MODIFY_QUEUE_DEPTH_STRUCT* queueDepthStruct =
            (MV_DRIVER_MSG_MODIFY_QUEUE_DEPTH_STRUCT*)dataPtr;
            MV_U8 channelIndex;
            if (psrbIoControl->Length < sizeof(MV_DRIVER_MSG_MODIFY_QUEUE_DEPTH_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessDriverMessage"
                         "(MV_DRIVER_MSG_DISABLE_QUEUING):"
                         " invalid buffer length.\n",
                         pSataAdapter->adapterId);
                break;
            }
            if ((queueDepthStruct->channelIndex >= MV_SATA_CHANNELS_NUM) ||
                (queueDepthStruct->queueDepth > 31))
            {
                Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
                break;
            }
            channelIndex = (MV_U8)queueDepthStruct->channelIndex;
            mvLogMsg(MV_IAL_LOG_ID,MV_DEBUG,"Channel %d - Modifying queue depth to %d\n",
                     channelIndex, queueDepthStruct->queueDepth);
            HwDeviceExtension->channelsInfo[channelIndex].queueDepth = (MV_U8)queueDepthStruct->queueDepth;
            HwDeviceExtension->channelsInfo[channelIndex].forceQueueDepth = MV_TRUE;
            psrbIoControl->ReturnCode = 0;
            Srb->SrbStatus = SRB_STATUS_SUCCESS;
            break;
        }
    default:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "mvIoctlProcessDriverMessage"
                 "Unkown control code %x",
                 controlCode);

    }


    finishDriverMessage:
#ifdef MV_DEBUG_IAL
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
    ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
    return TRUE;

}



MV_BOOLEAN
IOControlVersion2PIOCommandCompletionCB(MV_SATA_ADAPTER *pSataAdapter,
                                        MV_U8 channelNum,
                                        MV_COMPLETION_TYPE comp_type,
                                        MV_VOID_PTR commandId,
                                        MV_U16 responseFlags,
                                        MV_U32 timeStamp,
                                        MV_STORAGE_DEVICE_REGISTERS *registerStruct)
{
    PSCSI_REQUEST_BLOCK Srb = commandId;
    PVOID HwDeviceExtension;
    PSRB_EXTENSION SrbExt;
    PHW_DEVICE_EXTENSION deviceExtension;
    MV_U8   sataChannel;
    MV_U8   *dataPtr;
    MV_DEVICE_MSG_QUEUE_PIO_COMMAND_STRUCT* pioStruct;
    PSRB_IO_CONTROL psrbIoControl;

    if (commandId == NULL)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "[%d %d] commandId is NULL, cannot handle this !!!\n",
                 pSataAdapter->adapterId, channelNum);
        return MV_FALSE;
    }

    dataPtr = &((MV_U8*)(Srb->DataBuffer))[sizeof(SRB_IO_CONTROL)];
    pioStruct = (MV_DEVICE_MSG_QUEUE_PIO_COMMAND_STRUCT*)dataPtr;
    psrbIoControl = (PSRB_IO_CONTROL)Srb->DataBuffer;
    SrbExt = Srb->SrbExtension;
    HwDeviceExtension = SrbExt->HwDeviceExtension;
    deviceExtension = HwDeviceExtension;
    sataChannel =  Srb->PathId;

    pioStruct->sataRegs.errorRegister = registerStruct->errorRegister;
    pioStruct->sataRegs.statusRegister = registerStruct->statusRegister;
    pioStruct->sataRegs.sectorCountRegister = registerStruct->sectorCountRegister;
    pioStruct->sataRegs.lbaLowRegister = registerStruct->lbaLowRegister;
    pioStruct->sataRegs.lbaMidRegister = registerStruct->lbaMidRegister;
    pioStruct->sataRegs.lbaHighRegister = registerStruct->lbaHighRegister;
    pioStruct->sataRegs.deviceRegister = registerStruct->deviceRegister;
    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,
             "IOControlVersion2PIOCommandCompletionCB completion register dump -\n"
             "error register = %x, ATA status = %x sector count = %x, "
             "lba low = %x, lba mid = %x, lba high = %x.\n",
             pioStruct->sataRegs.errorRegister,
             pioStruct->sataRegs.statusRegister,
             pioStruct->sataRegs.sectorCountRegister,
             pioStruct->sataRegs.lbaLowRegister,
             pioStruct->sataRegs.lbaMidRegister,
             pioStruct->sataRegs.lbaHighRegister,
             pioStruct->sataRegs.deviceRegister);


    switch (comp_type)
    {
    case MV_COMPLETION_TYPE_NORMAL:
        psrbIoControl->ReturnCode = 0;
        Srb->SrbStatus = SRB_STATUS_SUCCESS;
        Srb->ScsiStatus = SCSISTAT_GOOD;
        break;
    case MV_COMPLETION_TYPE_ABORT:

        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "command Aborted. IOControl v.2 command\n");
        Srb->SrbStatus =  SRB_STATUS_ABORTED;
        break;
    case MV_COMPLETION_TYPE_ERROR:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "command Failed. IOControl v.2 command\n");
        Srb->SrbStatus = SRB_STATUS_ERROR;
        break;
    default:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "Unknown completion type (%d)\n", comp_type);
        return MV_FALSE;
    }
#ifdef MV_LOGGER
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete,
                         HwDeviceExtension,
                         Srb);

    ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
    return MV_TRUE;
}


BOOLEAN mvIoctlProcessDeviceMessage(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                                    IN PSCSI_REQUEST_BLOCK Srb)
{

    PSRB_IO_CONTROL psrbIoControl = (PSRB_IO_CONTROL)Srb->DataBuffer;
    MV_SATA_ADAPTER   *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_U8   *dataPtr = &((MV_U8*)(Srb->DataBuffer))[sizeof(SRB_IO_CONTROL)];
    MV_U32 controlCode;

    Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
    psrbIoControl->ReturnCode = 0xFFFFFFFF;
    if (psrbIoControl->Length < sizeof(MV_U32))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "[%d] mvIoctlProcessDeviceMessage()"
                 " invalid buffer length.\n",
                 pSataAdapter->adapterId);
        goto finishDeviceMessage;
    }
    controlCode = *(MV_U32*)dataPtr;
    switch (controlCode)
    {
    case MV_DEVICE_MSG_PIO_COMMAND:
        {
            MV_DEVICE_MSG_QUEUE_PIO_COMMAND_STRUCT* pioStruct =
            (MV_DEVICE_MSG_QUEUE_PIO_COMMAND_STRUCT*)dataPtr;
#ifdef MV_SATA_STORE_COMMANDS_INFO_ON_IAL_STACK
            MV_QUEUE_COMMAND_INFO   *pCommandInfo = &((PSRB_EXTENSION)Srb->SrbExtension)->commandInfo;
#else
            MV_QUEUE_COMMAND_INFO   commandInfo;
            MV_QUEUE_COMMAND_INFO   *pCommandInfo = &commandInfo;
#endif      
            MV_QUEUE_COMMAND_RESULT result;

            if (psrbIoControl->Length < sizeof(MV_DEVICE_MSG_QUEUE_PIO_COMMAND_STRUCT))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d] mvIoctlProcessDeviceMessage(MV_DEVICE_MSG_PIO_COMMAND)"
                         " Buffer too small (0x%x).\n", pSataAdapter->adapterId,
                         psrbIoControl->Length);
                break;
            }
            /*Sanity checks*/
            if (MV_NON_UDMA_PROTOCOL_NON_DATA == pioStruct->protocolType)
            {
                if (pioStruct->bufferSize > 0)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d] mvIoctlProcessDeviceMessage(MV_DEVICE_MSG_PIO_COMMAND)"
                             " invalid buffer length for non data command.\n",
                             pSataAdapter->adapterId);
                    break;
                }
            }
            else
            {
                if (pioStruct->bufferSize < ATA_SECTOR_SIZE)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d] mvIoctlProcessDeviceMessage(MV_DEVICE_MSG_PIO_COMMAND)"
                             " invalid buffer length for data command.\n",
                             pSataAdapter->adapterId);
                    break;
                }
                if (psrbIoControl->Length < (sizeof(MV_DEVICE_MSG_QUEUE_PIO_COMMAND_STRUCT) +
                                             pioStruct->bufferSize))
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d] mvIoctlProcessDeviceMessage(MV_DEVICE_MSG_PIO_COMMAND)"
                             " invalid buffer length.\n",
                             pSataAdapter->adapterId);
                    break;
                }
            }
            if (NULL == HwDeviceExtension->mvSataAdapter.sataChannel[pioStruct->channelIndex])
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                         "[%d %d]Trying to access disk info from a disconnected"
                         " channel\n",
                         pSataAdapter->adapterId,
                         pioStruct->channelIndex);
                break;
            }
            /* Check if outstanding commands is bigger or equal to 31.*/
            if (MV_FALSE == canQueueCmnds(HwDeviceExtension,
                                          pSataAdapter,
                                          (MV_U8)pioStruct->channelIndex,
                                          (MV_U8)pioStruct->PMPort))
            {
                Srb->SrbStatus = SRB_STATUS_BUSY;
                break;
            }
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,
                     "Queuing command - %x",pioStruct->sataRegs.commandRegister);
            memset (pCommandInfo, 0, sizeof(MV_QUEUE_COMMAND_INFO));
            ((PSRB_EXTENSION)Srb->SrbExtension)->commandType =
            MV_QUEUED_COMMAND_TYPE_NONE_UDMA;
            pCommandInfo->type = MV_QUEUED_COMMAND_TYPE_NONE_UDMA;
            pCommandInfo->PMPort = (MV_U8)pioStruct->PMPort;
            pCommandInfo->commandParams.NoneUdmaCommand.bufPtr =
            (MV_U16_PTR) &pioStruct->bufferStart;
            pCommandInfo->commandParams.NoneUdmaCommand.count =
            pioStruct->bufferSize/sizeof(MV_U16);
            pCommandInfo->commandParams.NoneUdmaCommand.callBack =
            IOControlVersion2PIOCommandCompletionCB;
            pCommandInfo->commandParams.NoneUdmaCommand.command =
            pioStruct->sataRegs.commandRegister;
            pCommandInfo->commandParams.NoneUdmaCommand.commandId =
            (MV_VOID_PTR) Srb;
            pCommandInfo->commandParams.NoneUdmaCommand.isEXT = MV_FALSE;
            pCommandInfo->commandParams.NoneUdmaCommand.protocolType =
            pioStruct->protocolType;
            pCommandInfo->commandParams.NoneUdmaCommand.device =
            pioStruct->sataRegs.deviceRegister;
            pCommandInfo->commandParams.NoneUdmaCommand.features =
            pioStruct->sataRegs.featuresRegister;
            pCommandInfo->commandParams.NoneUdmaCommand.sectorCount =
            pioStruct->sataRegs.sectorCountRegister;
            pCommandInfo->commandParams.NoneUdmaCommand.lbaLow =
            pioStruct->sataRegs.lbaLowRegister;
            pCommandInfo->commandParams.NoneUdmaCommand.lbaMid =
            pioStruct->sataRegs.lbaMidRegister;
            pCommandInfo->commandParams.NoneUdmaCommand.lbaHigh =
            pioStruct->sataRegs.lbaHighRegister;
            mvLogMsg(MV_IAL_LOG_ID, MV_DEBUG,
                     "[%d %d] Sending PIO command 0x%X: Srb %p\n",
                     pSataAdapter->adapterId, pioStruct->channelIndex,
                     pioStruct->sataRegs.commandRegister, Srb);
            result = mvSataQueueCommand(pSataAdapter,
                                        (MV_U8)pioStruct->channelIndex,
                                        pCommandInfo);
            if (result != MV_QUEUE_COMMAND_RESULT_OK)
            {
                switch (result)
                {
                case MV_QUEUE_COMMAND_RESULT_BAD_LBA_ADDRESS:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d %d] Queue 0x%X command failed. Bad LBA ??!!",
                             pSataAdapter->adapterId, pioStruct->channelIndex,
                             pioStruct->sataRegs.commandRegister);
                    break;
                case MV_QUEUE_COMMAND_RESULT_QUEUED_MODE_DISABLED:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d %d] Queue 0x%X command failed. EDMA"
                             " disabled\n",
                             pSataAdapter->adapterId, pioStruct->channelIndex,
                             pioStruct->sataRegs.commandRegister);
                    break;
                case MV_QUEUE_COMMAND_RESULT_FULL:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d %d] Queue 0x%X command failed. Queue is full.\n",
                             pSataAdapter->adapterId, pioStruct->channelIndex,
                             pioStruct->sataRegs.commandRegister);
                    break;
                case MV_QUEUE_COMMAND_RESULT_BAD_PARAMS:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d %d] Queue Identify command failed. (Bad "
                             "Params).\n",
                             pSataAdapter->adapterId, pioStruct->channelIndex,
                             pioStruct->sataRegs.commandRegister);
                    break;
                default:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                             "[%d %d] Bad result value 0x%X from queue"
                             " command\n",
                             pSataAdapter->adapterId, pioStruct->channelIndex,
                             result);
                }
                Srb->SrbStatus = SRB_STATUS_ERROR;
                break;
            }
            else
            {
                ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
                return TRUE;
            }
        }
    }
    finishDeviceMessage:
#ifdef MV_DEBUG_IAL
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
    ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
    return TRUE;

}



BOOLEAN mvIoctlVersion2(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                        IN PSCSI_REQUEST_BLOCK Srb)
{
    PSRB_IO_CONTROL psrbIoControl = (PSRB_IO_CONTROL)Srb->DataBuffer;
    switch (psrbIoControl->ControlCode)
    {
    case MV_ADAPTER_MSG:
        return mvIoctlProcessAdapterMessage(HwDeviceExtension,
                                            Srb);
        break;
    case MV_DEVICE_MSG:
        return mvIoctlProcessDeviceMessage(HwDeviceExtension,
                                           Srb);
        break;
    case MV_DRIVER_MSG:
        return mvIoctlProcessDriverMessage(HwDeviceExtension,
                                           Srb);
    default:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 "IO Control with control code 0x%08x is invalid\n",
                 psrbIoControl->ControlCode);
    }

    Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
#ifdef MV_DEBUG_IAL
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete,
                         HwDeviceExtension,
                         Srb);
    ScsiPortNotification(NextRequest, HwDeviceExtension,  NULL);
    return TRUE;
}




BOOLEAN
mvAdapterIoControl(IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                   IN PSCSI_REQUEST_BLOCK Srb)
{

    PSRB_IO_CONTROL psrbIoControl = (PSRB_IO_CONTROL)Srb->DataBuffer;
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    UCHAR   channelIndex;
    MV_U8   PMPort = 0;
    ULONG   length = 0;
    PSRB_EXTENSION      SrbExtension = (PSRB_EXTENSION)Srb->SrbExtension;
    MV_U8   sataChannel = Srb->PathId;


    SrbExtension->HwDeviceExtension = (PVOID)HwDeviceExtension;
    /* Default is success */
    Srb->SrbStatus = SRB_STATUS_SUCCESS;
    /*S.M.A.R.T. Command Handler*/
    if (!memcmp(psrbIoControl->Signature,
                "SCSIDISK",
                8))
    {
        return mvAdapterSmartCommand(HwDeviceExtension, Srb);
    }
    if (!memcmp(psrbIoControl->Signature,
                MV_IOCTL_VERSION2_SIGNATURE,
                MV_IOCTL_VERSION2_SIGNATURE_LENGTH))
    {
        return mvIoctlVersion2(HwDeviceExtension, Srb);
    }
    if (memcmp(psrbIoControl->Signature,
               MV_IOCTL_VERSION_SIGNATURE,
               MV_IOCTL_VERSION_SIGNATURE_LENGTH))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "IO Control signature not correct (%s) and"
                 " ControlCode %08x\n", psrbIoControl->Signature,
                 psrbIoControl->ControlCode);

        Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
#ifdef MV_DEBUG_IAL
        reportCDBCompletion(HwDeviceExtension,Srb);
#endif
        ScsiPortNotification(RequestComplete,
                             HwDeviceExtension,
                             Srb);
        ScsiPortNotification(NextRequest,
                             HwDeviceExtension,
                             NULL);
        return TRUE;
    }
    switch (psrbIoControl->ControlCode)
    {
    case MV_IOCTL_GET_ADAPTER_INFO:
        {
            psrbIoControl->ReturnCode = (HwDeviceExtension->adapterBusNumber & 0xff) |
                                        ((HwDeviceExtension->adapterDeviceNumber <<8) & 0xff00) |
                                        ((HwDeviceExtension->mvSataAdapter.pciConfigDeviceId <<16) & 0xffff0000);
            break;
        }

    case MV_IOCTL_GET_STATS:
        /* Be carefull that total data transferred to user should not */
        /* 512 bytes */
        if (Srb->DataTransferLength < MV_IOCTL_BUFFER_LENGTH)
        {
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR, "IO Control length error - %d (expected %d)\n",
                     Srb->DataTransferLength, MV_IOCTL_BUFFER_LENGTH);
            Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
            break;
        }
        length = 0;
        memset (Srb->DataBuffer, 0, Srb->Length);
        for (channelIndex = 0 ; channelIndex < MV_SATA_CHANNELS_NUM ; channelIndex ++)
        {
            if ((HwDeviceExtension->mvSataAdapter.sataChannel[channelIndex]) &&
                (mvStorageDevGetDeviceType(&HwDeviceExtension->mvSataAdapter,
                                           channelIndex)
                 == MV_SATA_DEVICE_TYPE_ATA_DISK))
            {
                length += sprintf (
                                  &((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer[length],
                                  "%01d %010d %010d %010d %010d %010d %010d %010d %010d "
                                  "%010d %010d %010d %010d %010d %010d\n",
                                  channelIndex,
                                  HwDeviceExtension->ataScsiAdapterExt.ataDriveData[channelIndex][PMPort].stats.totalIOs,
                                  HwDeviceExtension->ataScsiAdapterExt.totalAccumulatedOutstanding[channelIndex],
                                  HwDeviceExtension->ataScsiAdapterExt.ataDriveData[channelIndex][PMPort].stats.totalSectorsTransferred,
                                  HwDeviceExtension->stats[channelIndex].totalReadIOs,
                                  HwDeviceExtension->stats[channelIndex].totalWriteIOs,
                                  HwDeviceExtension->stats[channelIndex].totalVerifyIOs,
                                  HwDeviceExtension->stats[channelIndex].totalReadsSectorsTransferred,
                                  HwDeviceExtension->stats[channelIndex].totalWritesSectorsTransferred,
                                  HwDeviceExtension->stats[channelIndex].totalVerifiesSectorsTransferred,
                                  HwDeviceExtension->stats[channelIndex].totalMediumErrorsOccured,
                                  HwDeviceExtension->stats[channelIndex].totalCRCErrorsOccured,
                                  HwDeviceExtension->stats[channelIndex].totalWriteProtectErrorsOccured,
                                  HwDeviceExtension->stats[channelIndex].totalIllegalRequestsOccured,
                                  HwDeviceExtension->stats[channelIndex].lastErrorOccuredTimeStamp
                                  );
            }
        }
        length += sprintf (&((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer[length],
                           "%010d\n",HwDeviceExtension->numOfInterrupts);
        psrbIoControl->ReturnCode = length;
        break;
    case MV_IOCTL_GET_DRIVE_PARAMS:
        {
#ifdef MV_SATA_STORE_COMMANDS_INFO_ON_IAL_STACK
            MV_QUEUE_COMMAND_INFO   *pCommandInfo = &((PSRB_EXTENSION)Srb->SrbExtension)->commandInfo;
#else
            MV_QUEUE_COMMAND_INFO   commandInfo;
            MV_QUEUE_COMMAND_INFO   *pCommandInfo = &commandInfo;
#endif
            MV_QUEUE_COMMAND_RESULT result;
            channelIndex = (MV_U8)psrbIoControl->ReturnCode;

            /* Sanity check of parameters */
            if ((Srb->DataTransferLength != MV_IOCTL_BUFFER_LENGTH) ||
                (channelIndex >= MV_SATA_CHANNELS_NUM))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   "- IO Control - MV_IOCTL_GET_DRIVE_PARAMS error"
                         " on channel %d\n", channelIndex);
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   " Error - Possible cause -\n");
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   " Error - 1.. Channel number not correct\n");
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   " Error - 2.. IO Control data buffer length not"
                         " correct (%d)\n",Srb->DataTransferLength);

                Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
                break;
            }
            if (!HwDeviceExtension->mvSataAdapter.sataChannel[channelIndex])
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   " - Trying to read disk info from a disconnect"
                         " channel %d\n",channelIndex);
                Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
                break;
            }
            /* Check if outstanding commands is bigger or equal to 31.*/
            if (mvSataNumOfDmaCommands(pSataAdapter,channelIndex) >=
                HwDeviceExtension->channelsInfo[channelIndex].channelQueueDepth)
            {
                Srb->SrbStatus = SRB_STATUS_BUSY;
                break;
            }
            /* Report if queued DMA is enabled - there is no sanity check if */
            /* Queued DMA is supported because this will be filtered by GUI */
            ((char *) ((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer)[512] =
            (HwDeviceExtension->channelsInfo[channelIndex].forceNonQueued ==
             MV_FALSE) ? 1:0;

            /* Update used queue depth */
            ((char *) ((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer)[513] =
            HwDeviceExtension->channelsInfo[channelIndex].queueDepth;

            memset (pCommandInfo, 0, sizeof(MV_QUEUE_COMMAND_INFO));
            ((PSRB_EXTENSION)Srb->SrbExtension)->commandType = MV_QUEUED_COMMAND_TYPE_NONE_UDMA;
            pCommandInfo->type = MV_QUEUED_COMMAND_TYPE_NONE_UDMA;
            pCommandInfo->commandParams.NoneUdmaCommand.bufPtr =
            (MV_U16_PTR) ((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer;
            pCommandInfo->commandParams.NoneUdmaCommand.count = 256; /* 512 bytes */
            pCommandInfo->commandParams.NoneUdmaCommand.callBack =
            IOControlIdentifyDeviceCommandCompletionCB;
            pCommandInfo->commandParams.NoneUdmaCommand.command = MV_ATA_COMMAND_IDENTIFY;
            pCommandInfo->commandParams.NoneUdmaCommand.commandId = (MV_VOID_PTR) Srb;
            pCommandInfo->commandParams.NoneUdmaCommand.isEXT = MV_FALSE;
            pCommandInfo->commandParams.NoneUdmaCommand.protocolType = MV_NON_UDMA_PROTOCOL_PIO_DATA_IN;
            pCommandInfo->commandParams.NoneUdmaCommand.device = (MV_U8)(MV_BIT6);
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  "Sending Identify command: channel %d, Srb %p\n",
                     channelIndex, Srb);
            result = mvSataQueueCommand(pSataAdapter, channelIndex, pCommandInfo);
            if (result != MV_QUEUE_COMMAND_RESULT_OK)
            {
                switch (result)
                {
                case MV_QUEUE_COMMAND_RESULT_BAD_LBA_ADDRESS:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue Identify command failed. Bad LBA ??!!");
                    break;
                case MV_QUEUE_COMMAND_RESULT_QUEUED_MODE_DISABLED:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue Identify command failed. EDMA"
                             " disabled adapter %d channel %d\n",
                             pSataAdapter->adapterId, channelIndex);
                    break;
                case MV_QUEUE_COMMAND_RESULT_FULL:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue Identify command failed. Queue is"
                             " Full adapter %d channel %d\n",
                             pSataAdapter->adapterId, channelIndex);
                    break;
                case MV_QUEUE_COMMAND_RESULT_BAD_PARAMS:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Queue Identify command failed. (Bad "
                             "Params), pMvSataAdapter: %p,  pSataChannel: %p.\n",
                             pSataAdapter, pSataAdapter->sataChannel[channelIndex]);
                    break;
                default:
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Bad result value (%d) from queue"
                             " command\n", result);
                }
                Srb->SrbStatus = SRB_STATUS_ERROR;
                break;
            }
            else
            {
                ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
                return TRUE;
            }
            break;
        }

    case MV_IOCTL_SET_SATA_PHY_STATE :
        {
            unsigned int action =  (psrbIoControl->ReturnCode & 0xf0) >> 4;
            channelIndex = (MV_U8)(psrbIoControl->ReturnCode & 0xf);

            /* Sanity check if parameters are OK */
            if ((channelIndex >= MV_SATA_CHANNELS_NUM) || (action > 1) ||
                (psrbIoControl->ReturnCode != ((action << 4) & (channelIndex)) ) )
            {
                Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
                break;
            }
            if (action == 0) mvSataChannelPhyShutdown(&HwDeviceExtension->mvSataAdapter,
                                                      (MV_U8)channelIndex);
            else mvSataChannelPhyPowerOn(&HwDeviceExtension->mvSataAdapter,
                                         (MV_U8)channelIndex);
            break;
        }
    case MV_IOCTL_DUMP_PERFORMANCE_COUNTERS :
        {
            break;
        }
    case MV_IOCTL_READ_REGISTER :
        {
            psrbIoControl->ReturnCode = mvSataReadReg(
                                                     &HwDeviceExtension->mvSataAdapter,psrbIoControl->ReturnCode);
            break;
        }
    case MV_IOCTL_WRITE_REGISTER :
        {
            MV_U32 offset = 0, value = 0;
            offset = psrbIoControl->ReturnCode;
            value = (value << 8) | (MV_U8) ((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer[3];
            value = (value << 8) | (MV_U8) ((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer[2];
            value = (value << 8) | (MV_U8) ((PSRB_BUFFER)(Srb->DataBuffer))->ucDataBuffer[1];
            value    =    (   value    <<    8)    |    (   MV_U8)    (   (   PSRB_BUFFER)(   Srb->DataBuffer))->ucDataBuffer[   0];
            /* Check if offset is DWords alligned */
            if (offset & 0x3) break;
            mvSataWriteReg (&HwDeviceExtension->mvSataAdapter, offset, value);
            break;
        }

    case MV_IOCTL_MODIFY_QUEUED_DMA_MODE :
        {
            MV_BOOLEAN force;

            channelIndex = (MV_U8)(psrbIoControl->ReturnCode & 0xf);
            if ((channelIndex >= MV_SATA_CHANNELS_NUM))
            {
                Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
                break;
            }
            force =  ((psrbIoControl->ReturnCode & 0x10) >> 4) ? MV_TRUE:MV_FALSE;
            HwDeviceExtension->channelsInfo[channelIndex].forceNonQueued = force;
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "Setting force to %s\n",(force == MV_TRUE) ? "TRUE":"FALSE");
            ResetSataChannel(HwDeviceExtension, channelIndex);
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "Notify Reset detected\n");
            ScsiPortNotification(ResetDetected, HwDeviceExtension, NULL);
            Srb->SrbStatus = SRB_STATUS_BUS_RESET;
            break;
        }
    case MV_IOCTL_READ_CORE_DEBUG_LEVEL :
        {
#ifdef MV_LOGGER
            psrbIoControl->ReturnCode =
            mvLogGetModuleFilter(MV_CORE_DRIVER_LOG_ID);
#else
            psrbIoControl->ReturnCode = 0xffffffff;
#endif
            break;
        }
    case MV_IOCTL_WRITE_CORE_DEBUG_LEVEL :
        {
#ifdef MV_LOGGER
            MV_U32 debugLevel = psrbIoControl->ReturnCode;
            mvLogSetModuleFilter(MV_CORE_DRIVER_LOG_ID, debugLevel);
            mvLogSetModuleFilter(MV_SAL_LOG_ID, debugLevel);
            mvLogSetModuleFilter(MV_IAL_COMMON_LOG_ID, debugLevel);
            mvLogSetModuleFilter(MV_IAL_LOG_ID, debugLevel);
#endif
            break;
        }
    case MV_IOCTL_READ_IAL_DEBUG_LEVEL :
        {
#ifdef MV_LOGGER
            psrbIoControl->ReturnCode = ialDebugLevel;
#else
            psrbIoControl->ReturnCode = 0xffffffff;
#endif
            break;
        }
    case MV_IOCTL_WRITE_IAL_DEBUG_LEVEL :
        {
#ifdef MV_LOGGER
            ialDebugLevel = (MV_U8) psrbIoControl->ReturnCode;
#endif
            break;
        }
    case MV_IOCTL_SET_MAX_OUTSTANDING_COMMANDS :
        {
            MV_U16 maxOut = (MV_U16)((psrbIoControl->ReturnCode >> 16) & 0xff);

            channelIndex = (MV_U8)(psrbIoControl->ReturnCode & 0xf);
            if ((channelIndex >= MV_SATA_CHANNELS_NUM))
            {
                Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
                break;
            }
            HwDeviceExtension->channelsInfo[channelIndex].queueDepth = (MV_U8)maxOut;
            HwDeviceExtension->channelsInfo[channelIndex].forceQueueDepth = MV_TRUE;
            break;
        }
    default :
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   "IO Control with control code 0x%08x is invalid\n",
                 psrbIoControl->ControlCode);
        Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
#ifdef MV_LOGGER
        reportCDBCompletion(HwDeviceExtension,Srb);
#endif
        ScsiPortNotification(RequestComplete,
                             HwDeviceExtension,
                             Srb);

        ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
        return FALSE;
    }
#ifdef MV_LOGGER
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete,
                         HwDeviceExtension,
                         Srb);
    ScsiPortNotification(NextRequest,
                         HwDeviceExtension,
                         NULL);
    return TRUE;
}
#endif

#ifdef DPC_INT
BOOLEAN
HwScsiEnableInterruptsCallback(
                              IN PVOID DeviceExtension
                              )
{
    PHW_DEVICE_EXTENSION deviceExtension = DeviceExtension;
    MV_SATA_ADAPTER     *pSataAdapter = &(deviceExtension->mvSataAdapter);
    MV_U32 channel;

    if (mvSataInterruptServiceRoutine(pSataAdapter) == MV_TRUE)
    {
        deviceExtension->numOfInterrupts ++;
    }
    ScsiPortNotification(CallDisableInterrupts, DeviceExtension, HwScsiDisableInterruptsCallback);
    return TRUE;
}

BOOLEAN
HwScsiDisableInterruptsCallback(
                               IN PVOID DeviceExtension
                               )
{
    PHW_DEVICE_EXTENSION deviceExtension = DeviceExtension;
    MV_SATA_ADAPTER     *pSataAdapter = &(deviceExtension->mvSataAdapter);

    mvSataUnmaskAdapterInterrupt (pSataAdapter);

    return TRUE;
}
#endif

/* Command Completion and Event Notification (user implemented) */
MV_BOOLEAN
mvSataEventNotify(MV_SATA_ADAPTER *pSataAdapter, MV_EVENT_TYPE eventType,
                  MV_U32 param1, MV_U32 param2)
{
    PHW_DEVICE_EXTENSION deviceExtension =
    (PHW_DEVICE_EXTENSION)pSataAdapter->IALData;
    MV_U8   channelIndex = (MV_U8)param2;


    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  " Event received: adapter %d event %d"
             " param1= %x param2 = %x\n", pSataAdapter->adapterId,
             eventType - MV_EVENT_TYPE_ADAPTER_ERROR, param1, param2);

    switch (eventType)
    {
    case MV_EVENT_TYPE_ADAPTER_ERROR:
        ScsiPortLogError(deviceExtension,
                         NULL,
                         0,
                         0,
                         0,
                         SP_BUS_PARITY_ERROR,
                         __LINE__);
        /*TBD*/
        break;
    case MV_EVENT_TYPE_SATA_CABLE:
        {
            if (param1 == MV_SATA_CABLE_EVENT_CONNECT)/* connect*/
            {
                mvRestartChannel(&deviceExtension->ialCommonExt,
                                 channelIndex,
                                 &deviceExtension->ataScsiAdapterExt,
                                 MV_FALSE);
            }
            else
            { /*disconnect*/
                if (param1 == MV_SATA_CABLE_EVENT_DISCONNECT)
                {
                    if (pSataAdapter->sataChannel[channelIndex])
                    {
                        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,
                                 " [%d,%d]: channel removed\n",
                                 pSataAdapter->adapterId, channelIndex);
                    }
                    else
                    {
                        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                                 " [%d,%d]: channel already removed!!\n",
                                 pSataAdapter->adapterId, channelIndex);

                    }
                    if (mvSataIsStorageDeviceConnected(pSataAdapter, channelIndex) ==
                        MV_FALSE)
                    {
                        mvStopChannel(&deviceExtension->ialCommonExt,
                                      channelIndex,
                                      &deviceExtension->ataScsiAdapterExt);
                        deviceExtension->channelsInfo[channelIndex].forceNonQueued = MV_FALSE;
                    }
                }
                else
                {
                    if (param1 == MV_SATA_CABLE_EVENT_PM_HOT_PLUG)
                    {
                        mvPMHotPlugDetected(&deviceExtension->ialCommonExt,
                                            channelIndex,
                                            &deviceExtension->ataScsiAdapterExt);
                    }
                }
            }
            /*else*/
        }/*case*/
        break;
    case MV_EVENT_TYPE_SATA_ERROR:
        switch (param1)
        {
        case MV_SATA_RECOVERABLE_COMMUNICATION_ERROR:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                     " [%d %d] sata recoverable error occured\n",
                     pSataAdapter->adapterId, channelIndex);
            break;
        case MV_SATA_UNRECOVERABLE_COMMUNICATION_ERROR:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                     " [%d %d] sata unrecoverable error occured, restart channel\n",
                     pSataAdapter->adapterId, channelIndex);
            mvSataChannelHardReset(pSataAdapter, channelIndex);
            mvRestartChannel(&deviceExtension->ialCommonExt, channelIndex,
                             &deviceExtension->ataScsiAdapterExt, MV_TRUE);
            break;
        case MV_SATA_DEVICE_ERROR:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                     " [%d %d] device error occured\n",
                     pSataAdapter->adapterId, channelIndex);
            break;
        }
        break;
    default:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  " adapter %d unknown event %d"
                 " param1= %x param2 = %x\n", pSataAdapter->adapterId,
                 eventType - MV_EVENT_TYPE_ADAPTER_ERROR, param1, param2);

    }/*switch*/
    return MV_TRUE;
}

BOOLEAN strToInt(PCHAR String, int * value)
{
    if (String[0] == '-')
    {
        *value = -1;
        return TRUE;
    }
    if ((String[0] < '0') || (String[0] > '9'))
    {
        return FALSE;
    }
    if ((String[1] < '0') || (String[1] > '9'))
    {
        return FALSE;
    }
    * value = ((String[0] - '0') * 10) + (String[1] - '0');
    return TRUE;
}

void
parseQueueDepth (IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                 IN PCHAR ArgumentString)
{
    int length;
    int i, j;
    MV_BOOLEAN found = MV_FALSE;
    int values[MV_SATA_CHANNELS_NUM];

    if (ArgumentString == NULL ) return;
    length = strlen (ArgumentString);
    for (i = 0 ; i < (length - (int)strlen("QueueDepth")) ; i ++)
    {
        if (!strncmp (&ArgumentString[i], "QueueDepth", strlen ("QueueDepth")))
        {
            found = MV_TRUE;
            break;
        }
    }
    if (found == MV_FALSE)
    {
        return;
    }
    j = strlen("QueueDepth 00 00 00 00 00 00 00 00");

    if (((int)strlen (ArgumentString) - i) < j)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   "parseQueueDepth : Found QueueDepth key, but parameters "
                 "length mismatch (%d < %d)\n", (strlen (ArgumentString) - i),
                 j);
        return;
    }
    i+= strlen ("QueueDepth ");
    for (j = 0 ; j < MV_SATA_CHANNELS_NUM ; j++)
    {
        if (strToInt (&ArgumentString[i + j*3], &values[j]) == FALSE)
        {
            return;
        }
    }
    for (j = 0 ; j < MV_SATA_CHANNELS_NUM ; j++)
    {
        if (values[j] != -1)
        {
            HwDeviceExtension->channelsInfo[j].forceQueueDepth = MV_TRUE;
            HwDeviceExtension->channelsInfo[j].queueDepth = (MV_U8) values[j];
        }

    }
    return;
}


BOOLEAN
initAdapterExtension( IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                      IN PCHAR ArgumentString)
{
    MV_SATA_ADAPTER *pSataAdapter = &HwDeviceExtension->mvSataAdapter;
    MV_U8   i;
    BOOLEAN badArgs = FALSE;

#ifdef IOCTL
    HwDeviceExtension->numOfInterrupts = 0;
#endif
    for (i = 0 ; i < MV_SATA_CHANNELS_NUM; i ++)
    {
        HwDeviceExtension->channelsInfo[i].forceNonQueued = MV_FALSE;
        HwDeviceExtension->channelsInfo[i].forceQueueDepth = MV_FALSE;
        HwDeviceExtension->channelsInfo[i].channelQueueDepth = MV_SATA_SW_QUEUE_SIZE;
    }
    parseQueueDepth (HwDeviceExtension, ArgumentString);

    return TRUE;
}
BOOLEAN
setAdapterConfigSpaceParams( IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                             IN PPORT_CONFIGURATION_INFORMATION ConfigInfo)
{
    UCHAR   buffer[64];
    if (ScsiPortGetBusData( HwDeviceExtension,
                            PCIConfiguration,
                            ConfigInfo->SystemIoBusNumber,
                            ConfigInfo->SlotNumber,
                            buffer,
                            64) != 64)
    {
        return FALSE;
    }
    HwDeviceExtension->mvSataAdapter.pciConfigDeviceId = * ((MV_U16_PTR) &buffer[2]);
    HwDeviceExtension->mvSataAdapter.pciConfigRevisionId = buffer[8];
    HwDeviceExtension->BAR0 = *((MV_U32 *)(&buffer[0x10]));
    return TRUE;
}

BOOLEAN
AllocateEdmaQueues(
                  IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                  IN PPORT_CONFIGURATION_INFORMATION ConfigInfo
                  )
{
    ULONG   length;
    PUCHAR pArrayBaseAddr;
    SCSI_PHYSICAL_ADDRESS   pArrayBaseDmaAddr;
    ULONG   requestQSize;
    ULONG   responseQSize;
    ULONG   requestsArraySize;
    ULONG   responsesArraySize;

    if ((HwDeviceExtension->mvSataAdapter.pciConfigDeviceId == 0x6042) ||
        (HwDeviceExtension->mvSataAdapter.pciConfigDeviceId == 0x7042))
    {
        requestQSize = MV_EDMA_GEN2E_REQUEST_QUEUE_SIZE;
        responseQSize = MV_EDMA_GEN2E_RESPONSE_QUEUE_SIZE;
    }
    else
    {
        requestQSize = MV_EDMA_REQUEST_QUEUE_SIZE;
        responseQSize = MV_EDMA_RESPONSE_QUEUE_SIZE;
    }
    HwDeviceExtension->requestQSize = requestQSize;
    HwDeviceExtension->responseQSize = responseQSize;

    /*allocate spare queue for alignment*/
    requestsArraySize = (ConfigInfo->NumberOfBuses + 1) * requestQSize;
    responsesArraySize = ConfigInfo->NumberOfBuses * responseQSize;

    pArrayBaseAddr =
    ScsiPortGetUncachedExtension(HwDeviceExtension, ConfigInfo,
                                 requestsArraySize + responsesArraySize);

    if (pArrayBaseAddr == NULL)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d]: Failed to allocate memory for EDMA request"
                 " queues\n", HwDeviceExtension->mvSataAdapter.adapterId);
        ScsiPortLogError(HwDeviceExtension,
                         NULL,
                         0,
                         0,
                         0,
                         SP_INTERNAL_ADAPTER_ERROR,
                         __LINE__);

        return FALSE;
    }
    /*Requests array*/
    HwDeviceExtension->requestsArrayBaseAlignedAddr =
    pArrayBaseAddr + requestQSize;
    (ULONG_PTR)(HwDeviceExtension->requestsArrayBaseAlignedAddr) &=
    ~((ULONG_PTR)requestQSize - 1);
    pArrayBaseDmaAddr =  ScsiPortGetPhysicalAddress(HwDeviceExtension,
                                                    NULL,
                                                    pArrayBaseAddr,
                                                    &length);

    HwDeviceExtension->requestsArrayBaseDmaAlignedAddr =
    ScsiPortGetPhysicalAddress(HwDeviceExtension,
                               NULL,
                               HwDeviceExtension->requestsArrayBaseAlignedAddr,
                               &length);

    if ((HwDeviceExtension->requestsArrayBaseDmaAlignedAddr.QuadPart - pArrayBaseDmaAddr.QuadPart) !=
        (HwDeviceExtension->requestsArrayBaseAlignedAddr - pArrayBaseAddr))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d]: Error in Request Queues Alignment\n",
                 HwDeviceExtension->mvSataAdapter.adapterId);
        ScsiPortLogError(HwDeviceExtension,
                         NULL,
                         0,
                         0,
                         0,
                         SP_INTERNAL_ADAPTER_ERROR,
                         __LINE__);
        return FALSE;
    }
    /*Responses array*/
    HwDeviceExtension->responsesArrayBaseAlignedAddr =
    HwDeviceExtension->requestsArrayBaseAlignedAddr +
    (ConfigInfo->NumberOfBuses * requestQSize);
    pArrayBaseDmaAddr = ScsiPortGetPhysicalAddress(HwDeviceExtension,
                                                   NULL,
                                                   pArrayBaseAddr,
                                                   &length);
    HwDeviceExtension->responsesArrayBaseDmaAlignedAddr =
    ScsiPortGetPhysicalAddress(HwDeviceExtension,
                               NULL,
                               HwDeviceExtension->responsesArrayBaseAlignedAddr,
                               &length);


    if ((HwDeviceExtension->responsesArrayBaseDmaAlignedAddr.QuadPart
         - pArrayBaseDmaAddr.QuadPart) !=
        (HwDeviceExtension->responsesArrayBaseAlignedAddr -
         pArrayBaseAddr))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d]: Error in Responses Queues Alignment\n",
                 HwDeviceExtension->mvSataAdapter.adapterId);
        ScsiPortLogError(HwDeviceExtension,
                         NULL,
                         0,
                         0,
                         0,
                         SP_INTERNAL_ADAPTER_ERROR,
                         __LINE__);
        return FALSE;
    }
    return TRUE;
}

VOID
InitChannel(
           IN PHW_DEVICE_EXTENSION HwDeviceExtension,
           IN UCHAR channelIndex)
{
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    MV_SATA_CHANNEL     *pSataChannel;
    SCSI_PHYSICAL_ADDRESS   phyAddress;
    pSataChannel = &HwDeviceExtension->mvSataChannels[channelIndex];
    pSataAdapter->sataChannel[channelIndex] = pSataChannel;
    pSataChannel->channelNumber = channelIndex;


    pSataChannel->requestQueue = (struct mvDmaRequestQueueEntry *)
                                 (HwDeviceExtension->requestsArrayBaseAlignedAddr +
                                  (channelIndex * HwDeviceExtension->requestQSize));
    phyAddress.QuadPart =
    HwDeviceExtension->requestsArrayBaseDmaAlignedAddr.QuadPart +
    (channelIndex * HwDeviceExtension->requestQSize);

    /* check the alignment of the request queue*/
    if (phyAddress.LowPart & (HwDeviceExtension->requestQSize - 1))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d]: request queue allocated isn't  aligned,"
                 " dma_addr=%x channel=%d\n", pSataAdapter->adapterId,
                 phyAddress.QuadPart, channelIndex);
        ScsiPortLogError(HwDeviceExtension,
                         NULL,
                         0,
                         0,
                         0,
                         SP_INTERNAL_ADAPTER_ERROR,
                         __LINE__);
        return;
    }

    pSataChannel->requestQueuePciLowAddress = phyAddress.LowPart;
    pSataChannel->requestQueuePciHiAddress = phyAddress.HighPart;
    pSataChannel->responseQueue = (struct mvDmaResponseQueueEntry *)
                                  (HwDeviceExtension->responsesArrayBaseAlignedAddr +
                                   (channelIndex * HwDeviceExtension->responseQSize));
    phyAddress.QuadPart =
    HwDeviceExtension->responsesArrayBaseDmaAlignedAddr.QuadPart +
    (channelIndex * HwDeviceExtension->responseQSize);
    /* check the alignment of the response queue*/
    if (phyAddress.LowPart & (HwDeviceExtension->responseQSize - 1))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d]: response queue allocated isn't  Byte "
                 "aligned, dma_addr=%x\n",
                 pSataAdapter->adapterId, channelIndex, phyAddress);
        ScsiPortLogError(HwDeviceExtension,
                         NULL,
                         0,
                         0,
                         0,
                         SP_INTERNAL_ADAPTER_ERROR,
                         __LINE__);
        return;
    }
    pSataChannel->responseQueuePciLowAddress = phyAddress.LowPart;
    pSataChannel->responseQueuePciHiAddress = phyAddress.HighPart;
    /*mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  " [%d,%d]: response queue allocated: 0x%p\n",
              pSataAdapter->adapterId, channelIndex,
              pSataChannel->responseQueue);*/
}

MV_BOOLEAN IALConfigQueuingMode(MV_SATA_ADAPTER *pSataAdapter,
                                MV_U8 channelIndex,
                                MV_EDMA_MODE mode,
                                MV_SATA_SWITCHING_MODE switchingMode,
                                MV_BOOLEAN  use128Entries)
{

    PHW_DEVICE_EXTENSION HwDeviceExtension =
    (PHW_DEVICE_EXTENSION)pSataAdapter->IALData;

    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  " [%d %d]: IALConfigQueuingMode - mode = %d,"
             "forceNonQueued = %s, forceQueueDepth = %s\n",
             pSataAdapter->adapterId, channelIndex, mode,
             (HwDeviceExtension->channelsInfo[channelIndex].forceNonQueued == MV_TRUE) ? "TRUE":"FALSE",
             (HwDeviceExtension->channelsInfo[channelIndex].forceQueueDepth == MV_TRUE) ? "TRUE":"FALSE");

    if ((mode == MV_EDMA_MODE_NOT_QUEUED) ||
        (HwDeviceExtension->channelsInfo[channelIndex].forceNonQueued ==
         MV_TRUE))
    {
        mode = MV_EDMA_MODE_NOT_QUEUED;
        if (HwDeviceExtension->channelsInfo[channelIndex].forceQueueDepth ==
            MV_FALSE)
        {
            HwDeviceExtension->channelsInfo[channelIndex].queueDepth =
            MV_DEFAULT_QUEUE_DEPTH;
        }
    }

    if (mode != MV_EDMA_MODE_NOT_QUEUED)
    {
        if (HwDeviceExtension->channelsInfo[channelIndex].forceQueueDepth ==
            MV_FALSE)
        {
            HwDeviceExtension->channelsInfo[channelIndex].queueDepth = 31;
        }
    }
    HwDeviceExtension->channelsInfo[channelIndex].channelQueueDepth = MV_SATA_SW_QUEUE_SIZE;
    if (use128Entries == MV_TRUE)
    {
        HwDeviceExtension->channelsInfo[channelIndex].channelQueueDepth = MV_SATA_GEN2E_SW_QUEUE_SIZE;
    }
    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  " [%d %d]: EDMA mode %d, channel queue"
             " depth %d Disk queue depth = %d\n",
             pSataAdapter->adapterId, channelIndex, mode,
             HwDeviceExtension->channelsInfo[channelIndex].channelQueueDepth,
             HwDeviceExtension->channelsInfo[channelIndex].queueDepth);

    if (mvSataConfigEdmaMode(pSataAdapter,
                             channelIndex,
                             mode,
                             31) == MV_FALSE)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d] channel %d Error: "
                 "mvSataConfigEdmaMode failed\n",
                 pSataAdapter->adapterId, channelIndex);
        return MV_FALSE;
    }
    return MV_TRUE;
}

MV_BOOLEAN IALBusChangeNotify(MV_SATA_ADAPTER *pSataAdapter, MV_U8 channelIndex)
{
    mvLogMsg(MV_IAL_LOG_ID, MV_DEBUG_ERROR, "[%d %d] "
             "Notify Windows OS about bus change.\n",
             pSataAdapter->adapterId, channelIndex);
    ScsiPortNotification(BusChangeDetected,
                         (PHW_DEVICE_EXTENSION)pSataAdapter->IALData,
                         (UCHAR)channelIndex);
    return MV_TRUE;
}

MV_BOOLEAN IALBusChangeNotifyEx(MV_SATA_ADAPTER *pSataAdapter,
                                MV_U8 channelIndex,
                                MV_U16 targetsToRemove,
                                MV_U16 targetsToAdd)
{
    return MV_TRUE;
}


void IALReleaseChannel(MV_SATA_ADAPTER *pSataAdapter, MV_U8 channelIndex)
{
}

MV_BOOLEAN IALInitChannel(MV_SATA_ADAPTER *pSataAdapter, MV_U8 channelIndex)
{
    InitChannel((PHW_DEVICE_EXTENSION)pSataAdapter->IALData, channelIndex);
    return MV_TRUE;
}


VOID
HwScsiTimer(
           IN PVOID DeviceExtension
           )
{
    PHW_DEVICE_EXTENSION deviceExtension =
    (PHW_DEVICE_EXTENSION)DeviceExtension;
    MV_SATA_ADAPTER *pSataAdapter = &deviceExtension->mvSataAdapter;

    mvIALTimerCallback(&deviceExtension->ialCommonExt,
                       &deviceExtension->ataScsiAdapterExt);

    ScsiPortNotification(RequestTimerCall,
                         DeviceExtension,
                         HwScsiTimer,
                         MV_IAL_ASYNC_TIMER_PERIOD*1000);
}


BOOLEAN
ResetSataChannel(
                IN PHW_DEVICE_EXTENSION HwDeviceExtension,
                UCHAR   channelIndex)
{
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    if ((!pSataAdapter) || (channelIndex >= pSataAdapter->numberOfChannels) ||
        (!pSataAdapter->sataChannel[channelIndex]))
    {
        return TRUE;
    }
    mvSataDisableChannelDma(pSataAdapter, channelIndex);
    /* Flush pending commands */
    mvSataFlushDmaQueue (pSataAdapter, channelIndex, MV_FLUSH_TYPE_CALLBACK);
    mvSataChannelHardReset(pSataAdapter, channelIndex);
    if (pSataAdapter->sataChannel[channelIndex])
    {
        if (mvSataIsStorageDeviceConnected(pSataAdapter, channelIndex) ==
            MV_FALSE)
        {
            /* Hardware reset channel */
            mvSataChannelHardReset(pSataAdapter, channelIndex);
            mvStopChannel(&HwDeviceExtension->ialCommonExt,
                          channelIndex,
                          &HwDeviceExtension->ataScsiAdapterExt);
            return FALSE;
        }

        mvRestartChannel(&HwDeviceExtension->ialCommonExt,
                         channelIndex,
                         &HwDeviceExtension->ataScsiAdapterExt,
                         MV_TRUE);
    }
    return TRUE;
}

BOOLEAN
sendScsiCommand(
               IN PHW_DEVICE_EXTENSION HwDeviceExtension,
               IN PSCSI_REQUEST_BLOCK Srb)

{
    MV_SATA_ADAPTER     *pSataAdapter = &(HwDeviceExtension->mvSataAdapter);
    PSRB_EXTENSION      SrbExt = Srb->SrbExtension;
    MV_SATA_SCSI_CMD_BLOCK *pCmdBlock = &SrbExt->SALBlock;
    MV_U8               channelIndex = Srb->PathId;
    MV_SCSI_COMMAND_STATUS_TYPE         commandStatus;

    pCmdBlock->bus = Srb->PathId;
    pCmdBlock->target = Srb->TargetId;
    pCmdBlock->lun = Srb->Lun;
    pCmdBlock->completionCallBack = IALCompletion;
    pCmdBlock->dataBufferLength = Srb->DataTransferLength;
    pCmdBlock->IALData = Srb;
    pCmdBlock->pSalAdapterExtension = &HwDeviceExtension->ataScsiAdapterExt;
    pCmdBlock->pIalAdapterExtension = &HwDeviceExtension->ialCommonExt;
    pCmdBlock->pDataBuffer = Srb->DataBuffer;
    pCmdBlock->pSenseBuffer = Srb->SenseInfoBuffer;
    pCmdBlock->ScsiCdb = Srb->Cdb;
    pCmdBlock->ScsiCdbLength = Srb->CdbLength;
    pCmdBlock->senseBufferLength = Srb->SenseInfoBufferLength;
    pCmdBlock->ScsiCommandCompletion = MV_SCSI_COMPLETION_INVALID_STATUS;
#ifdef MV_SATA_STORE_COMMANDS_INFO_ON_IAL_STACK
    pCmdBlock->pCommandInfo = &SrbExt->commandInfo;
#endif    

    if ((Srb->Cdb[0] == SCSIOP_READ) || (Srb->Cdb[0] == SCSIOP_READ6) ||
        (Srb->Cdb[0] == SCSIOP_WRITE) || (Srb->Cdb[0] == SCSIOP_WRITE6))
    {
        if (buildPRDTable(HwDeviceExtension, Srb) == FALSE)
        {
            return TRUE;
        }
    }

    if (HwDeviceExtension->inDump == TRUE)
    {
        commandStatus = mvExecuteScsiCommand(pCmdBlock, MV_TRUE);
    }
    else
    {
        commandStatus = mvExecuteScsiCommand(pCmdBlock, MV_FALSE);

    }

#ifdef MV_LOGGER
    if ((ialDebugLevel & MV_IAL_DEBUG_CDB) ||
        (mvLogGetModuleFilter(MV_IAL_LOG_ID) & MV_DEBUG))
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ Adapter %d: command sent to translation layer: status %d\n",
                 pSataAdapter->adapterId, commandStatus);
    }
#endif
    if (commandStatus == MV_SCSI_COMMAND_STATUS_QUEUED)
    {
        if ((Srb->SrbFlags & SRB_FLAGS_QUEUE_ACTION_ENABLE) &&
            (canQueueCmnds(HwDeviceExtension,pSataAdapter ,channelIndex,
                           Srb->TargetId) == MV_TRUE))
        {
            ScsiPortNotification(NextLuRequest, (PVOID)HwDeviceExtension,
                                 Srb->PathId, Srb->TargetId, Srb->Lun);
        }
        else
        {
            ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
        }
#ifdef IOCTL

        if (mvSataNumOfDmaCommands(pSataAdapter,channelIndex) == HwDeviceExtension->channelsInfo[channelIndex].queueDepth)
        {
#ifdef MV_LOGGER
            if ((ialDebugLevel & MV_IAL_DEBUG_CDB) ||
                (mvLogGetModuleFilter(MV_IAL_LOG_ID) & MV_DEBUG))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ Adapter %d: outstanding command = max queue depth (%d)\n",
                         pSataAdapter->adapterId, HwDeviceExtension->channelsInfo[channelIndex].queueDepth);
            }
#endif
        }
#endif
    }
    return TRUE;
}

MV_BOOLEAN
IOControlIdentifyDeviceCommandCompletionCB(MV_SATA_ADAPTER *pSataAdapter,
                                           MV_U8 channelNum,
                                           MV_COMPLETION_TYPE comp_type,
                                           MV_VOID_PTR commandId,
                                           MV_U16 responseFlags,
                                           MV_U32 timeStamp,
                                           MV_STORAGE_DEVICE_REGISTERS *registerStruct)
{
    PSCSI_REQUEST_BLOCK Srb;
    PVOID HwDeviceExtension;
    PSRB_EXTENSION SrbExt;
    PHW_DEVICE_EXTENSION deviceExtension;
    MV_U8   sataChannel;

    if (commandId == NULL)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  " commandId is NULL, can't hanlde this !!!,adapterId=%d,"
                 " channel=%d \n", pSataAdapter->adapterId, channelNum);
        return MV_FALSE;
    }

    Srb = commandId;
    SrbExt = Srb->SrbExtension;
    HwDeviceExtension = SrbExt->HwDeviceExtension;
    deviceExtension = HwDeviceExtension;
    sataChannel =  Srb->PathId;

    switch (comp_type)
    {
    case MV_COMPLETION_TYPE_NORMAL:
        Srb->SrbStatus = SRB_STATUS_SUCCESS;
        Srb->ScsiStatus = SCSISTAT_GOOD;
        break;
    case MV_COMPLETION_TYPE_ABORT:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "command Aborted. IOControl command\n");
        Srb->SrbStatus =  SRB_STATUS_ABORTED;
        break;
    case MV_COMPLETION_TYPE_ERROR:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "command Failed. IOControl command\n");
        Srb->SrbStatus = SRB_STATUS_ERROR;
        break;
    default:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  " Unknown completion type (%d)\n", comp_type);
        return MV_FALSE;
    }
#ifdef MV_LOGGER
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete,
                         HwDeviceExtension,
                         Srb);
    ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
    return MV_TRUE;
}

MV_BOOLEAN SmartATACommandCompletionCB(MV_SATA_ADAPTER *pSataAdapter,
                                       MV_U8 channelNum,
                                       MV_COMPLETION_TYPE comp_type,
                                       MV_VOID_PTR commandId,
                                       MV_U16 responseFlags,
                                       MV_U32 timeStamp,
                                       MV_STORAGE_DEVICE_REGISTERS *registerStruct)
{
    PSCSI_REQUEST_BLOCK Srb;
    PVOID HwDeviceExtension;
    PSRB_EXTENSION SrbExt;
    PHW_DEVICE_EXTENSION deviceExtension;
    MV_U8   sataChannel;
    PSENDCMDOUTPARAMS cmdOutParams;
    PIDEREGS    ideRegs;
    UCHAR       PathId;
    UCHAR       TargetId;
    UCHAR       Lun;

    if (commandId == NULL)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Unable to handle CommandId = NULL, adapterId=%d,"
                 " channel=%d \n", pSataAdapter->adapterId, channelNum);
        return MV_FALSE;
    }

    Srb = commandId;
    SrbExt = Srb->SrbExtension;
    HwDeviceExtension = SrbExt->HwDeviceExtension;
    deviceExtension = HwDeviceExtension;

    PathId =  Srb->PathId;
    TargetId = Srb->TargetId;
    Lun = Srb->Lun;

    cmdOutParams = (PSENDCMDOUTPARAMS)
                   &((PUCHAR)Srb->DataBuffer)[sizeof(SRB_IO_CONTROL)];
    ideRegs = (PIDEREGS)&cmdOutParams->bBuffer[0];
    switch (comp_type)
    {
    case MV_COMPLETION_TYPE_NORMAL:
        cmdOutParams->DriverStatus.bDriverError = SMART_NO_ERROR;
        ((PSRB_IO_CONTROL)Srb->DataBuffer)->ReturnCode = 0;
        switch (((PSRB_IO_CONTROL)Srb->DataBuffer)->ControlCode)
        {
        case IOCTL_SCSI_MINIPORT_ENABLE_SMART:
        case IOCTL_SCSI_MINIPORT_DISABLE_SMART:
        case IOCTL_SCSI_MINIPORT_ENABLE_DISABLE_AUTOSAVE:
        case IOCTL_SCSI_MINIPORT_EXECUTE_OFFLINE_DIAGS:
        case IOCTL_SCSI_MINIPORT_RETURN_STATUS:
            cmdOutParams->bBuffer[3] =
            (UCHAR)registerStruct->lbaMidRegister & 0xFF;
            cmdOutParams->bBuffer[4] =
            (UCHAR)registerStruct->lbaHighRegister & 0xFF;
            break;
        case IOCTL_SCSI_MINIPORT_IDENTIFY:
        case IOCTL_SCSI_MINIPORT_READ_SMART_ATTRIBS:
        case IOCTL_SCSI_MINIPORT_READ_SMART_THRESHOLDS:
        case IOCTL_SCSI_MINIPORT_SAVE_ATTRIBUTE_VALUES:
#ifdef IOCTL_SCSI_MINIPORT_WRITE_SMART_LOG
        case IOCTL_SCSI_MINIPORT_WRITE_SMART_LOG:
#endif
#ifdef IOCTL_SCSI_MINIPORT_READ_SMART_LOG
        case IOCTL_SCSI_MINIPORT_READ_SMART_LOG:
#endif
            break;
        default:
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "Unknown SMART IOControl command 0x%X\n",
                     comp_type);
            cmdOutParams->DriverStatus.bDriverError =
            SMART_NOT_SUPPORTED;
            cmdOutParams->DriverStatus.bIDEError =
            registerStruct->errorRegister;
            Srb->SrbStatus = SRB_STATUS_ERROR;
            break;
        }
        Srb->SrbStatus = SRB_STATUS_SUCCESS;
        Srb->ScsiStatus = SCSISTAT_GOOD;
        break;
    case MV_COMPLETION_TYPE_ABORT:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "command Aborted. IOControl command\n");
        cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
        cmdOutParams->DriverStatus.bIDEError =
        registerStruct->errorRegister;
        cmdOutParams->cBufferSize = 0;
        Srb->SrbStatus =  SRB_STATUS_ABORTED;
        mvCommandCompletionErrorHandler(&deviceExtension->ialCommonExt,
                                        channelNum);
        break;
    case MV_COMPLETION_TYPE_ERROR:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "command Failed. IOControl command\n");
        cmdOutParams->DriverStatus.bDriverError = SMART_IDE_ERROR;
        cmdOutParams->DriverStatus.bIDEError =
        registerStruct->errorRegister;
        cmdOutParams->cBufferSize = 0;
        Srb->SrbStatus = SRB_STATUS_ERROR;
        mvCommandCompletionErrorHandler(&deviceExtension->ialCommonExt,
                                        channelNum);
        break;
    default:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  " Unknown completion type (%d)\n", comp_type);
        return MV_FALSE;
    }
#ifdef MV_LOGGER
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete,
                         HwDeviceExtension,
                         Srb);
    ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
    return MV_TRUE;
}


/*******************************************************************************
* DriverEntry  -  Installable driver initialization entry point for system.
*
* DESCRIPTION:  Installable driver initialization entry point for system.
*
*
* INPUT:
*      Driver Object
*
* RETURN:
*       Status from ScsiPortInitialize()
*
* COMMENTS:
*       None.
*
*******************************************************************************/
ULONG
DriverEntry(
           IN PVOID DriverObject,
           IN PVOID Argument2
           )
{
    HW_INITIALIZATION_DATA hwInitializationData;
    ULONG   adapterNo = 1;
    ULONG pciStatus, pciStatusTemp;
    ULONG i;
    UCHAR VendorId[4] = {
        '1', '1', 'A', 'B'};
    UCHAR DeviceId5080[4] = {
        '5', '0', '8', '0'};
    UCHAR DeviceId5081[4] = {
        '5', '0', '8', '1'};
    UCHAR DeviceId5040[4] = {
        '5', '0', '4', '0'};
    UCHAR DeviceId5041[4] = {
        '5', '0', '4', '1'};
    UCHAR DeviceId6081[4] = {
        '6', '0', '8', '1'};
    UCHAR DeviceId6041[4] = {
        '6', '0', '4', '1'};
    UCHAR DeviceId6042[4] = {
        '6', '0', '8', '2'};
    UCHAR DeviceId7042[4] = {
        '7', '0', '4', '2'};


#ifdef MV_LOGGER
#if defined (MV_LOG_DEBUG)
    mvLogRegisterModule(MV_CORE_DRIVER_LOG_ID, 0x1FF,
                        szModules[MV_CORE_DRIVER_LOG_ID]);
    mvLogRegisterModule(MV_SAL_LOG_ID, 0x1FF,
                        szModules[MV_SAL_LOG_ID]);
    mvLogRegisterModule(MV_IAL_COMMON_LOG_ID, 0x1FF,
                        szModules[MV_IAL_COMMON_LOG_ID]);
    mvLogRegisterModule(MV_IAL_LOG_ID, 0x1FF,
                        szModules[MV_IAL_LOG_ID]);
#elif defined (MV_LOG_ERROR)
    mvLogRegisterModule(MV_CORE_DRIVER_LOG_ID, MV_DEBUG_FATAL_ERROR|MV_DEBUG_ERROR,
                        szModules[MV_CORE_DRIVER_LOG_ID]);
    mvLogRegisterModule(MV_SAL_LOG_ID, MV_DEBUG_FATAL_ERROR|MV_DEBUG_ERROR,
                        szModules[MV_SAL_LOG_ID]);
    mvLogRegisterModule(MV_IAL_COMMON_LOG_ID, MV_DEBUG_FATAL_ERROR|MV_DEBUG_ERROR,
                        szModules[MV_IAL_COMMON_LOG_ID]);
    mvLogRegisterModule(MV_IAL_LOG_ID, MV_DEBUG_FATAL_ERROR|MV_DEBUG_ERROR,
                        szModules[MV_IAL_LOG_ID]);
#endif
#endif

    mvLogMsg(MV_IAL_LOG_ID, MV_DEBUG,
             "\nEntering SCSI MiniPort DriverEntry: compilation time %s\n",
             __TIME__);

    /* Zero out structure.*/
    for (i=0; i<sizeof(HW_INITIALIZATION_DATA); i++)
    {
        ((PUCHAR)&hwInitializationData)[i] = 0;
    }


    /* Set size of hwInitializationData.*/

    hwInitializationData.HwInitializationDataSize = sizeof(HW_INITIALIZATION_DATA);


    /* Set entry points.*/


    hwInitializationData.HwInitialize = mvHwInitialize;
    hwInitializationData.HwResetBus = mvResetBus;
    hwInitializationData.HwStartIo = mvStartIo;
    hwInitializationData.HwInterrupt = mvInterrupt;
    hwInitializationData.HwFindAdapter = mvFindAdapter;
#ifdef MV_WINDOWS_2000

    hwInitializationData.HwAdapterControl = mvAdapterControl;
#endif

    /* Indicate no buffer mapping but will need physical addresses.*/


    hwInitializationData.NeedPhysicalAddresses = TRUE;

    hwInitializationData.AutoRequestSense = TRUE;
    hwInitializationData.MultipleRequestPerLu = TRUE;
    hwInitializationData.TaggedQueuing = TRUE;
    /* Specify size of extensions.*/

    hwInitializationData.DeviceExtensionSize = sizeof(HW_DEVICE_EXTENSION);
    hwInitializationData.SpecificLuExtensionSize = 0;

    hwInitializationData.VendorId = VendorId;
    hwInitializationData.VendorIdLength = 4;
    hwInitializationData.DeviceId = DeviceId5080;
    hwInitializationData.DeviceIdLength = 4;

    /* Specifiy the bus type.*/
    hwInitializationData.AdapterInterfaceType = PCIBus;
    hwInitializationData.NumberOfAccessRanges = 1;

    /* Ask for SRB extensions for EDMA PRD Tables.*/

    hwInitializationData.SrbExtensionSize = sizeof(SRB_EXTENSION);

    /* The adapter count is used by the find adapter routine to track how   */
    /* which adapter addresses have been tested.                            */
    pciStatus = ScsiPortInitialize(DriverObject, Argument2,
                                   &hwInitializationData, &adapterNo);

    hwInitializationData.DeviceId = DeviceId5081;
    pciStatusTemp = ScsiPortInitialize(DriverObject, Argument2,
                                       &hwInitializationData, &adapterNo);
    if (pciStatus > pciStatusTemp)
    {
        pciStatus = pciStatusTemp;
    }
    hwInitializationData.DeviceId = DeviceId5040;
    pciStatusTemp = ScsiPortInitialize(DriverObject, Argument2,
                                       &hwInitializationData, &adapterNo);
    if (pciStatus > pciStatusTemp)
    {
        pciStatus = pciStatusTemp;
    }
    hwInitializationData.DeviceId = DeviceId5041;
    pciStatusTemp = ScsiPortInitialize(DriverObject, Argument2,
                                       &hwInitializationData, &adapterNo);
    if (pciStatus > pciStatusTemp)
    {
        pciStatus = pciStatusTemp;
    }
    hwInitializationData.DeviceId = DeviceId6081;
    pciStatusTemp = ScsiPortInitialize(DriverObject, Argument2,
                                       &hwInitializationData, &adapterNo);
    if (pciStatus > pciStatusTemp)
    {
        pciStatus = pciStatusTemp;
    }
    hwInitializationData.DeviceId = DeviceId6041;
    pciStatusTemp = ScsiPortInitialize(DriverObject, Argument2,
                                       &hwInitializationData, &adapterNo);
    if (pciStatus > pciStatusTemp)
    {
        pciStatus = pciStatusTemp;
    }
    hwInitializationData.DeviceId = DeviceId6042;
    pciStatusTemp = ScsiPortInitialize(DriverObject, Argument2,
                                       &hwInitializationData, &adapterNo);
    if (pciStatus > pciStatusTemp)
    {
        pciStatus = pciStatusTemp;
    }
    hwInitializationData.DeviceId = DeviceId7042;
    pciStatusTemp = ScsiPortInitialize(DriverObject, Argument2,
                                       &hwInitializationData, &adapterNo);
    if (pciStatus > pciStatusTemp)
    {
        pciStatus = pciStatusTemp;
    }
    return pciStatus;


} /* end DriverEntry()*/

BOOLEAN mvSataStrComp(IN PCHAR String1, IN PCHAR String2)
{
    PCHAR index1 = String1;
    PCHAR index2 = String2;

    while (*index1)
    {
        if (*index1 != *index2)
        {
            return FALSE;
        }
        index1++;
        index2++;
    }
    return TRUE;
}

BOOLEAN InLdrOrDump(IN PCHAR ArgumentString)
{
    PCHAR stringIndex;
    stringIndex = ArgumentString;

    if (stringIndex == NULL)
    {
        return FALSE;
    }

    //
    // Lower-case the string.
    //
    while (*stringIndex)
    {

        if (*stringIndex >= 'A' && *stringIndex <= 'Z')
        {
            *stringIndex = *stringIndex + ('a' - 'A');
        }
        stringIndex++;
    }

    //
    // Reset the pointer.
    //
    stringIndex = ArgumentString;

    //
    // If ntldr=1 or dump=1 return true, as
    // resources are limited in these contexts.
    //

    if (mvSataStrComp("ntldr=1", stringIndex))
    {

        //
        // In the context of the loader.
        //
        return TRUE;

    }
    else
    {
        if (mvSataStrComp("dump=1", stringIndex))
        {

            //
            // In context of a crashdump or hibernate.
            //
            return TRUE;
        }
    }
    //
    // Normal context.
    //
    return FALSE;
}


ULONG
mvFindAdapter(
             IN PVOID HwDeviceExtension,
             IN PVOID Context,
             IN PVOID BusInformation,
             IN PCHAR ArgumentString,
             IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo,
             OUT PBOOLEAN Again
             )
{
    PHW_DEVICE_EXTENSION deviceExtension = HwDeviceExtension;
    MV_SATA_ADAPTER     *pSataAdapter = &(deviceExtension->mvSataAdapter);
    PACCESS_RANGE       accessRange;
    MV_BUS_ADDR_T       pciAddress;
    ULONG               i;
    static ULONG        adapterId = 0;
    MV_U8               channelIndex;
    BOOLEAN             ntldrDump = FALSE;

    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "\nEntering SCSI MiniPort mvFindAdapter: compilation time %s\n", __TIME__);
    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   " Driver file %s \n", __FILE__);
    deviceExtension->inDump = FALSE;

    if (ArgumentString)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   " Adapter %d found. PCI bus %d,"
                 " device %d, Args:'%s'\n", adapterId,
                 ConfigInfo->SystemIoBusNumber, ConfigInfo->SlotNumber,
                 ArgumentString);
        if (InLdrOrDump(ArgumentString))
        {
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                     "Adapter %d. Called during hibernation, or kernel dump. Args:'%s'\n", adapterId,
                     ArgumentString);
            deviceExtension->inDump = TRUE;
            ntldrDump = TRUE;
        }

    }
    else
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   " Adapter %d found. PCI bus %d, device %d\n", adapterId,
                 ConfigInfo->SystemIoBusNumber, ConfigInfo->SlotNumber);
    }

    if ((TRUE == ntldrDump) && (TRUE == mvSataInitializedForDump))
    {
        //DbgBreakPoint();
        mvLogMsg(MV_IAL_LOG_ID, MV_DEBUG_ERROR, "mvFindAdapter() "
                 "Initialize adapter for hibernatinon or kernel dump done only once."
                 "Exiting.\n");
        *Again = FALSE;
        return SP_RETURN_NOT_FOUND;
    }

    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   " Adapter %d found. BusInterruptLevel %d BusInterruptVector %d\n", adapterId,
             ConfigInfo->BusInterruptLevel, ConfigInfo->BusInterruptVector);

#ifdef IOCTL
    deviceExtension->adapterBusNumber = (MV_U8) ConfigInfo->SystemIoBusNumber;
    deviceExtension->adapterDeviceNumber = (MV_U8) ConfigInfo->SlotNumber;
#endif
    accessRange = &((*(ConfigInfo->AccessRanges))[0]);

    if (ScsiPortValidateRange(HwDeviceExtension,
                              ConfigInfo->AdapterInterfaceType,
                              ConfigInfo->SystemIoBusNumber,
                              accessRange->RangeStart,
                              accessRange->RangeLength,
                              FALSE))    /* FALSE: memory space */
    {
        pciAddress = (MV_BUS_ADDR_T) ScsiPortGetDeviceBase(HwDeviceExtension,
                                                           ConfigInfo->AdapterInterfaceType,
                                                           ConfigInfo->SystemIoBusNumber,
                                                           accessRange->RangeStart,
                                                           accessRange->RangeLength,
                                                           FALSE);  /* FALSE: memory space */


        for (i=0; i<sizeof(MV_SATA_ADAPTER); i++)
        {
            ((PUCHAR)pSataAdapter)[i] = 0;
        }
        pSataAdapter->adapterIoBaseAddress = pciAddress;
        pSataAdapter->adapterId = adapterId++;
    }
    else
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "ScsiPortValidateRange failed\n");
        return(SP_RETURN_NOT_FOUND);
    }

    /* Hardware found, let's find out hardware configuration
       and fill out ConfigInfo table for WinNT
    */

    if (setAdapterConfigSpaceParams(deviceExtension, ConfigInfo) == MV_FALSE)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "setAdapterConfigSpaceParams failed\n");
        return(SP_RETURN_ERROR);
    }
    if ((deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x5040) ||
        (deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x5041) ||
        (deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x6041) ||
        (deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x6042) ||
        (deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x7042))
    {
        /* 4 SATA channels */
        ConfigInfo->NumberOfBuses = MV_SATA_CHANNELS_NUM / 2;
    }
    else
    {
        /* 8 SATA channels */
        ConfigInfo->NumberOfBuses = MV_SATA_CHANNELS_NUM;
    }

    ConfigInfo->MaximumTransferLength = MAX_TRANSFER_SIZE;
    if (ConfigInfo->NumberOfPhysicalBreaks >= MAX_SG_DESCRIPTORS - 1)
    {
        ConfigInfo->NumberOfPhysicalBreaks = MAX_SG_DESCRIPTORS - 1;
    }
    ConfigInfo->ScatterGather = TRUE;
    ConfigInfo->Master = TRUE;
    ConfigInfo->NeedPhysicalAddresses = TRUE;
    ConfigInfo->Dma32BitAddresses = TRUE;
    ConfigInfo->InterruptMode = LevelSensitive;
    ConfigInfo->ResetTargetSupported = TRUE;

    ConfigInfo->BufferAccessScsiPortControlled = TRUE;
    ConfigInfo->MapBuffers = TRUE;


    if ((((deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x5040) ||
          (deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x5041) ||
          (deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x5081)) &&
         (deviceExtension->mvSataAdapter.pciConfigRevisionId == 0x2))
        ||
        (((deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x6041) ||
          (deviceExtension->mvSataAdapter.pciConfigDeviceId == 0x6081)) &&
         (deviceExtension->mvSataAdapter.pciConfigRevisionId == 0x1)))
    {
        /*set double DWORD buffer alignment restriction*/
        /* Fix for 88SX50xx FEr PCI#10*/
        /* Fix for 88SX60xx FEr PCI#3*/
        ConfigInfo->AlignmentMask = 0x7;
    }
    else
    {
        ConfigInfo->AlignmentMask = 0x1;
    }


    ConfigInfo->MaximumNumberOfTargets = MV_SATA_PM_MAX_PORTS;
    ConfigInfo->MaximumNumberOfLogicalUnits = 1;

    for (i = 0; i < MV_SATA_CHANNELS_NUM; i++)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  "InitiatorBusId[%d] %d\n", i, ConfigInfo->InitiatorBusId[i]);
        ConfigInfo->InitiatorBusId[i] = MV_SATA_PM_MAX_PORTS;
    }

#ifdef MV_WINDOWS_2000
#ifdef MV_SUPPORT_WMI
    /*
     *  Inform SCSIPORT that we are a WMI data provider and have we GUIDs
     *  to register.
     */

    ConfigInfo->WmiDataProvider = TRUE;
    mvWmiInitialize(deviceExtension);

#else

    /* Inform Scsi port that no support for WMI */
    ConfigInfo->WmiDataProvider = FALSE;
#endif

#endif

    /* Inform Scsi port we support 64 bit DMA */
    if (ConfigInfo->Dma64BitAddresses & SCSI_DMA64_SYSTEM_SUPPORTED)
    {
        deviceExtension->bDma64BitSupported = TRUE;
        ConfigInfo->Dma64BitAddresses |= SCSI_DMA64_MINIPORT_SUPPORTED;
    }
    else
    {
        deviceExtension->bDma64BitSupported = FALSE;
    }
    if (AllocateEdmaQueues(deviceExtension, ConfigInfo) == FALSE)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "AllocateEdmaQueues failed\n");
        return(SP_RETURN_ERROR);
    }

#if 0
    pSataAdapter->intCoalThre[0]= 0;
    pSataAdapter->intCoalThre[1]= 0;
    pSataAdapter->intTimeThre[0] = 0 ;
    pSataAdapter->intTimeThre[1] = 0 ;
#else
    pSataAdapter->intCoalThre[0]= 4;
    pSataAdapter->intCoalThre[1]= 4;
    pSataAdapter->intTimeThre[0] = 150 * 50 ;
    pSataAdapter->intTimeThre[1] = 150 * 50 ;
#endif

    if (initAdapterExtension(deviceExtension, ArgumentString) == FALSE)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "initAdapterExtension failed\n");
        return(SP_RETURN_ERROR);
    }
    /* mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,  "Adapter %d: unit 0 intCoal %x timeThr %x unit 1 intCoal %x timeThe %x\n",
           pSataAdapter->adapterId,
           pSataAdapter->intCoalThre[0], pSataAdapter->intTimeThre[0],
           pSataAdapter->intCoalThre[1], pSataAdapter->intTimeThre[1]);*/
    pSataAdapter->pciSerrMask = MV_PCI_SERR_MASK_REG_ENABLE_ALL;
    pSataAdapter->pciInterruptMask = MV_PCI_INTERRUPT_MASK_REG_ENABLE_ALL;
#undef READ_1_5
#ifdef READ_1_5
    pSataAdapter->pciCommand = (MV_PCI_COMMAND_REG_DEFAULT | MV_BIT31 | MV_BIT19) & ~(MV_BIT5 | MV_BIT7);
#else
    pSataAdapter->pciCommand = MV_PCI_COMMAND_REG_DEFAULT;
#endif
    pSataAdapter->mvSataEventNotify = mvSataEventNotify;


    if (mvSataInitAdapter(pSataAdapter) == MV_FALSE)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "mvSataInitAdapter failed\n");
        return(SP_RETURN_ERROR);
    }
    mvSataScsiInitAdapterExt(&deviceExtension->ataScsiAdapterExt, pSataAdapter);
    pSataAdapter->IALData = deviceExtension;

    if (ntldrDump)
    {
        //
        // When this routine is called by the dump driver to
        // handle writing a crash dump or hibernate file to a drive
        // handled by the IOP, this routine will get called for EACH
        // call that driverEntry makes to StorPortIntialize.  In this
        // case the config info that is passed in will be the config
        // info for the IOP that is in the hibernate or dump path.
        // We only need to initialize the IOP once when in dump mode
        // so set a flag here to insure we don't load again.
        //
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "set mvSataInitializedForDump = TRUE\n");
        mvSataInitializedForDump = TRUE;
        *Again = FALSE;
    }
    else
    {
        *Again = TRUE;
    }
    return(SP_RETURN_FOUND);
}


BOOLEAN
mvHwInitialize(
              IN PVOID HwDeviceExtension
              )
{
    PHW_DEVICE_EXTENSION deviceExtension = HwDeviceExtension;
    MV_SATA_ADAPTER     *pSataAdapter = &(deviceExtension->mvSataAdapter);
    if (MV_FALSE == mvAdapterStartInitialization(pSataAdapter,
                                                 &deviceExtension->ialCommonExt,
                                                 &deviceExtension->ataScsiAdapterExt))
    {
        ScsiPortLogError(HwDeviceExtension,
                         NULL,
                         0,
                         0,
                         0,
                         SP_INTERNAL_ADAPTER_ERROR,
                         __LINE__);
        return FALSE;
    }
    ScsiPortNotification(RequestTimerCall,
                         deviceExtension,
                         HwScsiTimer,
                         MV_IAL_ASYNC_TIMER_PERIOD*1000);
    ScsiPortNotification(ResetDetected, HwDeviceExtension, NULL);
    return TRUE;
}


BOOLEAN
mvStartIo(
         IN PVOID HwDeviceExtension,
         IN PSCSI_REQUEST_BLOCK Srb
         )
{
    PHW_DEVICE_EXTENSION deviceExtension = HwDeviceExtension;
    MV_SATA_ADAPTER     *pSataAdapter = &(deviceExtension->mvSataAdapter);
    UCHAR               channelIndex;

#ifdef MV_LOGGER
    if ((ialDebugLevel & MV_IAL_DEBUG_CDB) ||
        (mvLogGetModuleFilter(MV_IAL_LOG_ID) & MV_DEBUG))
    {
        if (Srb->Function == SRB_FUNCTION_EXECUTE_SCSI)
        {
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++--\\/----\\/---Command Received-----\n");
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ Adapter %d: SCSI cmd received  (PathId %d, TargetId %d, LUN %d, length = %d)\n",
                     pSataAdapter->adapterId, Srb->PathId, Srb->TargetId, Srb->Lun, Srb->Length);
            if (Srb->CdbLength == 6)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ %02x, %02x, %02x, %02x, %02x, %02x\n",
                         Srb->Cdb[0], Srb->Cdb[1], Srb->Cdb[2], Srb->Cdb[3],
                         Srb->Cdb[4], Srb->Cdb[5]);
            }
            else if (Srb->CdbLength == 10)
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ %02x, %02x, %02x, %02x, %02x, %02x, %02x, %02x, %02x, %02x\n",
                         Srb->Cdb[0], Srb->Cdb[1], Srb->Cdb[2], Srb->Cdb[3], Srb->Cdb[4],
                         Srb->Cdb[5], Srb->Cdb[6], Srb->Cdb[7], Srb->Cdb[8], Srb->Cdb[9]);
            }
            else
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ Unsupported length %d in CDB\n",Srb->Length);
            }
            if (Srb->Cdb[0] == SCSIOP_MODE_SELECT)
            {
                MV_U32 i;
                if (Srb->DataBuffer == NULL)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ While dumping MODE SELECT - DataBuffer is NULL\n");
                }
                else
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ Dumping MODE SELECT data (DataTransferLength = %d)\n",
                             Srb->DataTransferLength);
                    for (i = 0 ; i < Srb->DataTransferLength ; i++)
                    {
                        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++ %d : %02x\n",i,((UCHAR *) Srb->DataBuffer)[i]);
                    }
                }

            }

            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "++++--/\\----/\\---Command Received-----\n");
        }
    }
#endif


#ifdef MV_SUPPORT_WMI

    /* Check if command is a WMI request. */

    if (Srb->Function == SRB_FUNCTION_WMI)
    {

        /* Process the WMI request and return. */
        return mvWmiSrb(HwDeviceExtension, (PSCSI_WMI_REQUEST_BLOCK) Srb);
    }
#endif

    switch (Srb->Function)
    {
    case SRB_FUNCTION_EXECUTE_SCSI:
        sendScsiCommand(deviceExtension,Srb);
        return TRUE;
    case SRB_FUNCTION_RESET_BUS:
    case SRB_FUNCTION_RESET_DEVICE:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   "MVSATA: %d %d: Reset Device (Function %d)\n", pSataAdapter->adapterId,
                 Srb->PathId, Srb->Function);
        ResetSataChannel(deviceExtension, Srb->PathId);

#ifdef MV_LOGGER
        reportCDBCompletion(HwDeviceExtension,Srb);
#endif
        ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
        ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
        return TRUE;
#ifdef IOCTL
    case SRB_FUNCTION_IO_CONTROL:
        if (deviceExtension->ialCommonExt.channelState[Srb->PathId]
            != CHANNEL_READY)
        {
            Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
            ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
            ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
            break;
        }
        return mvAdapterIoControl(deviceExtension, Srb);
#else
    case SRB_FUNCTION_IO_CONTROL:
#endif
    case SRB_FUNCTION_ABORT_COMMAND:
    case SRB_FUNCTION_CLAIM_DEVICE:
    case SRB_FUNCTION_RECEIVE_EVENT:
    case SRB_FUNCTION_RELEASE_QUEUE:
    case SRB_FUNCTION_ATTACH_DEVICE:
    case SRB_FUNCTION_RELEASE_DEVICE:
    case SRB_FUNCTION_SHUTDOWN:
    case SRB_FUNCTION_FLUSH:
    case SRB_FUNCTION_RELEASE_RECOVERY:
    case SRB_FUNCTION_TERMINATE_IO:
    case SRB_FUNCTION_FLUSH_QUEUE:
    default:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   " SRB Function %d not supported\n",Srb->Function);

        Srb->SrbStatus = SRB_STATUS_INVALID_REQUEST;
#ifdef MV_LOGGER
        reportCDBCompletion(HwDeviceExtension,Srb);
#endif
        ScsiPortNotification(RequestComplete, HwDeviceExtension, Srb);
        ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
        return TRUE;
    } /* end switch */

    return FALSE;
}

BOOLEAN
mvInterrupt(
           IN PVOID HwDeviceExtension
           )
{
    PHW_DEVICE_EXTENSION deviceExtension = HwDeviceExtension;
    MV_SATA_ADAPTER     *pSataAdapter = &(deviceExtension->mvSataAdapter);
    KIRQL  OldIrql;
    MV_U32 channel;
#ifdef DPC_INT
    mvSataMaskAdapterInterrupt (pSataAdapter);
    ScsiPortNotification(CallEnableInterrupts,HwDeviceExtension ,
                         HwScsiEnableInterruptsCallback);
    return TRUE;

#else
    static MV_U32 valid=0;

    if (mvSataInterruptServiceRoutine(pSataAdapter) == MV_TRUE)
    {
#ifdef IOCTL
        deviceExtension->numOfInterrupts ++;
#endif
        mvSataScsiPostIntService(&deviceExtension->ataScsiAdapterExt);
        return TRUE;
    }
    return FALSE;
#endif
}

BOOLEAN
mvResetBus(
          IN PVOID HwDeviceExtension,
          IN ULONG PathId
          )
{
    MV_U8   channelIndex = (MV_U8)PathId;
    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
             " %d %d: reset Bus called\n",
             ((PHW_DEVICE_EXTENSION)HwDeviceExtension)->mvSataAdapter.adapterId,
             PathId);

    if (PathId > MV_SATA_CHANNELS_NUM)
        return FALSE;

    // When in the crashdump or hibernate paths the dump
    // driver will issue a bus reset after initializing
    // the driver (via the normal driverEntry path).  Since
    // the IOP has just been reset as part of the
    // IOP initialization, don't issue another one.
    //
    // If the driver is already in the middle of a Bus Reset request,
    // don't allow another one to get through or it will corrupt the
    // reset count for the already outstanding resets.  When running
    // in a clustering environment and doing fail overs, some machines
    // seem to get impatient and send multiple resets.
    //
    if (MV_TRUE == ((PHW_DEVICE_EXTENSION)HwDeviceExtension)->inDump)
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,
                 " %d %d: cannot perform reset Bus during hibernation.\n",
                 ((PHW_DEVICE_EXTENSION)HwDeviceExtension)->mvSataAdapter.adapterId,
                 PathId);

        ScsiPortCompleteRequest(HwDeviceExtension,
                                (UCHAR) PathId,
                                SP_UNTAGGED,
                                SP_UNTAGGED,
                                SRB_STATUS_BUS_RESET);

        ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);
        return TRUE;
    }

    if (ResetSataChannel((PHW_DEVICE_EXTENSION)HwDeviceExtension,
                         channelIndex) == FALSE)
    {
        ScsiPortNotification(BusChangeDetected, HwDeviceExtension,
                             PathId);

    }
    ScsiPortCompleteRequest(HwDeviceExtension,
                            (UCHAR) PathId,
                            SP_UNTAGGED,
                            SP_UNTAGGED,
                            SRB_STATUS_BUS_RESET);

    ScsiPortNotification(NextRequest, HwDeviceExtension, NULL);

    return TRUE;
}

#ifdef MV_WINDOWS_2000

SCSI_ADAPTER_CONTROL_STATUS
mvAdapterControl(
                IN PVOID HwDeviceExtension,
                IN SCSI_ADAPTER_CONTROL_TYPE ControlType,
                IN PVOID Parameters
                )
{
    PHW_DEVICE_EXTENSION deviceExtension = HwDeviceExtension;
    MV_SATA_ADAPTER     *pSataAdapter = &(deviceExtension->mvSataAdapter);
    PSCSI_SUPPORTED_CONTROL_TYPE_LIST pControlTypeList;


    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   " [%d]: Adapter Control Called. ControlType: %d\n",
             pSataAdapter->adapterId, ControlType);
    switch (ControlType)
    {
    case ScsiQuerySupportedControlTypes:{
            BOOLEAN supportedTypes[ScsiAdapterControlMax] = {
                TRUE,   /* ScsiQuerySupportedControlTypes   */
                TRUE,   /* ScsiStopAdapter                  */
                FALSE,  /* ScsiRestartAdapter               */
                FALSE,  /* ScsiSetBootConfig                */
                FALSE   /* ScsiSetRunningConfig             */
            };

            ULONG max = ScsiAdapterControlMax;
            ULONG i;

            pControlTypeList = (PSCSI_SUPPORTED_CONTROL_TYPE_LIST) Parameters;

            if (pControlTypeList->MaxControlType < max)
            {
                max = pControlTypeList->MaxControlType;
            }

            for (i = 0; i < max; i++)
            {
                pControlTypeList->SupportedTypeList[i] = supportedTypes[i];
            }


            break;


        }
    case ScsiStopAdapter:
        {
            /*
             * Shut down all interrupts on the adapter.  They'll get re-enabled
             * by the initialization routines.
            */
            ScsiPortNotification(RequestTimerCall,
                                 deviceExtension,
                                 HwScsiTimer,
                                 0);
            mvSataMaskAdapterInterrupt(pSataAdapter);
            break;
        }
    default:{
            return ScsiAdapterControlUnsuccessful;
        }
    }

    return ScsiAdapterControlSuccess;
}


#endif
MV_BOOLEAN IALCompletion(struct mvSataAdapter *pSataAdapter,
                         MV_SATA_SCSI_CMD_BLOCK *pCmdBlock)
{
    PSCSI_REQUEST_BLOCK Srb = (PSCSI_REQUEST_BLOCK)pCmdBlock->IALData;
    PHW_DEVICE_EXTENSION HwDeviceExtension =
    (PHW_DEVICE_EXTENSION)pSataAdapter->IALData;
    UCHAR   pathId = Srb->PathId;

    switch (pCmdBlock->ScsiCommandCompletion)
    {
    case MV_SCSI_COMPLETION_SUCCESS:
        Srb->SrbStatus = SRB_STATUS_SUCCESS;
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;
    case MV_SCSI_COMPLETION_BAD_SCSI_COMMAND:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with BAD_SCSI_COMMAND\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_ERROR | SRB_STATUS_AUTOSENSE_VALID;
        if (Srb->SenseInfoBufferLength > (UCHAR)pCmdBlock->senseDataLength)
        {
            Srb->SenseInfoBufferLength = (UCHAR)pCmdBlock->senseDataLength;
        }
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;
    case MV_SCSI_COMPLETION_ATA_FAILED:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with ATA FAILED\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_ERROR | SRB_STATUS_AUTOSENSE_VALID;
        if (Srb->SenseInfoBufferLength > (UCHAR)pCmdBlock->senseDataLength)
        {
            Srb->SenseInfoBufferLength = (UCHAR)pCmdBlock->senseDataLength;
        }
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;
    case MV_SCSI_COMPLETION_UA_RESET:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed"
                 " with UA BUS RESET\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_ERROR | SRB_STATUS_AUTOSENSE_VALID;
        if (Srb->SenseInfoBufferLength > (UCHAR)pCmdBlock->senseDataLength)
        {
            Srb->SenseInfoBufferLength = (UCHAR)pCmdBlock->senseDataLength;
        }
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;
    case MV_SCSI_COMPLETION_UA_PARAMS_CHANGED:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed"
                 " with UA PARAMETERS CHANGED\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_ERROR | SRB_STATUS_AUTOSENSE_VALID;
        if (Srb->SenseInfoBufferLength > (UCHAR)pCmdBlock->senseDataLength)
        {
            Srb->SenseInfoBufferLength = (UCHAR)pCmdBlock->senseDataLength;
        }
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;
    case MV_SCSI_COMPLETION_ABORTED:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with ABORTED\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_ERROR;
        if (Srb->SenseInfoBufferLength > (UCHAR)pCmdBlock->senseDataLength)
        {
            Srb->SenseInfoBufferLength = (UCHAR)pCmdBlock->senseDataLength;
        }
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;

    case MV_SCSI_COMPLETION_OVERRUN:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with OVERRUN\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_DATA_OVERRUN;
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;
    case MV_SCSI_COMPLETION_UNDERRUN:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with UNDERRUN\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_DATA_OVERRUN;
        Srb->ScsiStatus = pCmdBlock->ScsiStatus;
        Srb->DataTransferLength = pCmdBlock->dataTransfered;
        break;
    case MV_SCSI_COMPLETION_INVALID_BUS:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with INVALID BUS\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_INVALID_PATH_ID;
        break;
    case MV_SCSI_COMPLETION_NO_DEVICE:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with NO DEVICE\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_NO_DEVICE;
        break;
    case MV_SCSI_COMPLETION_PARITY_ERROR:
        /*
         * If completing command with SRB_STATUS_PARITY_ERROR then this will
         * reach application since this status means also that retry failed
         * (by the DDK) ; but in this case we want the SCSI Port to do retry
         */
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "[%d,%d,%d] Scsi command completed with PARITY ERROR - completing"
                 " to NT with SRB_STATUS_ABORTED\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_ABORTED;
        break;
    case MV_SCSI_COMPLETION_DISCONNECT:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Scsi command completed with DISCONNECT\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_UNEXPECTED_BUS_FREE;
        break;
    case MV_SCSI_COMPLETION_QUEUE_FULL:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Scsi command completed with QUEUE FULL\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_BUSY;
        break;
    case MV_SCSI_COMPLETION_BUS_RESET:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Scsi command completed with BUS RESET\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_BUS_RESET;
        ScsiPortNotification(ResetDetected, HwDeviceExtension,NULL);
        break;
    case MV_SCSI_COMPLETION_NOT_READY:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Scsi command status NOT READY - waiting for timeout\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Setting SRB Staus and Scsi Status to 0\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = 0;
        Srb->ScsiStatus = 0;
        return MV_TRUE;
    case MV_SCSI_COMPLETION_INVALID_STATUS:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Scsi command status INVALID STATUS - waiting for timeout\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Setting SRB Staus and Scsi Status to 0\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = 0;
        Srb->ScsiStatus = 0;
        return MV_TRUE;
    case MV_SCSI_COMPLETION_BAD_SCB:
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Scsi command status BAD SCB\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        Srb->SrbStatus = SRB_STATUS_ERROR;
        break;
    default:
        Srb->SrbStatus = 0;
        Srb->ScsiStatus = 0;
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] ERROR: Unknown Scsi completion status %d - waiting for timeout\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId,
                 pCmdBlock->ScsiCommandCompletion);
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  "[%d,%d,%d] Setting SRB Staus and Scsi Status to 0\n",
                 pSataAdapter->adapterId, Srb->PathId, Srb->TargetId);
        return MV_TRUE;

    }
#ifdef MV_LOGGER
    reportCDBCompletion(HwDeviceExtension,Srb);
#endif
    ScsiPortNotification(RequestComplete,
                         HwDeviceExtension,
                         Srb);
    ScsiPortNotification(NextLuRequest,
                         HwDeviceExtension,
                         pathId,
                         Srb->TargetId, Srb->Lun);
    return MV_TRUE;
}
BOOLEAN buildPRDTable(PHW_DEVICE_EXTENSION HwDeviceExtension, PSCSI_REQUEST_BLOCK Srb)
{
    MV_SATA_EDMA_PRD_ENTRY  *PRDTable;
    PUCHAR                  virtualAddress;
    PUCHAR                  PRDAligned;
    ULONG                   length = 0, bufferLength = 0, remainLength, i;
    SCSI_PHYSICAL_ADDRESS   physAddress;
    MV_SATA_SCSI_CMD_BLOCK *pCmdBlock = &((PSRB_EXTENSION)Srb->SrbExtension)->SALBlock;

    if (Srb->DataTransferLength > 0)
    {
        virtualAddress =  Srb->DataBuffer;
        remainLength = Srb->DataTransferLength;
        PRDAligned = (PUCHAR)((PSRB_EXTENSION)Srb->SrbExtension)->PRDTable;
        PRDAligned += 31;
        (ULONG_PTR)PRDAligned &= ~0x1f;
        PRDTable = (MV_SATA_EDMA_PRD_ENTRY *)PRDAligned;
        /* Build PRD table from scatter gather list  */

        /*
        There are 4 spare entries in PRD table for further buffer split, in
        case then either the buffer size exceeds 64 KB or
        32 high address bits of all data in the buffer are not identical
        */
        for (i = 0; (remainLength > 0) && (i < MAX_SG_DESCRIPTORS + 4); i++)
        {

            if (length <= 0)
            {
                /*Get next buffer in SG list*/
                physAddress = ScsiPortGetPhysicalAddress(HwDeviceExtension,
                                                         Srb,
                                                         virtualAddress,
                                                         &length);
                if (length > remainLength)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   "*****IAL***** Warning length(%x)"
                             " > remainLength(%x)\n",
                             length,
                             remainLength);
                    length = remainLength;
                }
                virtualAddress += length;
            }
            bufferLength = length;
            PRDTable[i].flags = 0;
            PRDTable[i].reserved = 0;
            PRDTable[i].lowBaseAddr = physAddress.LowPart;
            PRDTable[i].highBaseAddr = physAddress.HighPart;
#ifdef TRACE_MAX_BUFFER_LEN
            {
                static ULONG maxLength = 0;

                if (length > maxLength)
                {
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "  max length 0x%x\n", length);
                    maxLength = length;
                }

            }
#endif
#ifdef TRACE_MAX_TRANS_LEN
            {
                static ULONG totlaMaxLength = 0;

                if (Srb->DataTransferLength > totlaMaxLength)
                {
                    totlaMaxLength = Srb->DataTransferLength;
                    mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG,   "  total max length 0x%x\n", totlaMaxLength);
                }

            }
#endif
            /*Buffer size must not exceed 64K*/
            if (bufferLength > MV_BIT16)
            {
                bufferLength  =  MV_BIT16;
            }
            /*Split the buffer if the 32 bit high address of its data
            is not identical*/
            if (HwDeviceExtension->bDma64BitSupported == TRUE &&
                ((physAddress.QuadPart | 0xFFFFFFFF) !=
                 ((physAddress.QuadPart + bufferLength - 1) | 0xFFFFFFFF)))
            {
                mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,   "*****IAL***** Warning: buffer crosses"
                         " 4G boundaries\n");
                bufferLength = (ULONG)(0x100000000 -
                                       (physAddress.QuadPart & 0xFFFFFFFF));
            }
            PRDTable[i].byteCount = (USHORT)bufferLength;
            remainLength -= bufferLength;
            length -= bufferLength;
            physAddress.QuadPart += bufferLength;
        }
        if ((i >= MAX_SG_DESCRIPTORS + 4) && (remainLength > 0))
        {
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG_ERROR,  " - Number of SG entries"
                     " (%d) > %d\n", i, MAX_SG_DESCRIPTORS);
            Srb->SrbStatus = SRB_STATUS_ERROR;
#ifdef MV_LOGGER
            reportCDBCompletion(HwDeviceExtension,Srb);
#endif
            ScsiPortNotification(RequestComplete, (PVOID)HwDeviceExtension, Srb);
            ScsiPortNotification(NextRequest, (PVOID)HwDeviceExtension, NULL);
            return FALSE;
        }
#ifdef MV_SATA_SUPPORT_EDMA_SINGLE_DATA_REGION
        if ((HwDeviceExtension->mvSataAdapter.sataAdapterGeneration >= MV_SATA_GEN_IIE) &&
            (i == 1))
        {
            pCmdBlock->singleDataRegion = MV_TRUE;
            pCmdBlock->byteCount = PRDTable[0].byteCount;
            pCmdBlock->PRDTableLowPhyAddress = PRDTable[0].lowBaseAddr;
            pCmdBlock->PRDTableHighPhyAddress = PRDTable[0].highBaseAddr;
            mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG, "send single Data Region buffer (length %x)\n",
                     pCmdBlock->byteCount);
            return TRUE;
        }
        else
        {
            pCmdBlock->singleDataRegion = MV_FALSE;
        }
#endif
        PRDTable[i-1].flags = MV_EDMA_PRD_EOT_FLAG;
    }
    else
    {
        mvLogMsg(MV_IAL_LOG_ID,  MV_DEBUG, " IO recieved with zero transfer length.\n");
        return TRUE;
    }
    //MV_PRINTD ("Number of SG entries in PRD table is %d\n", i);
    virtualAddress = (PUCHAR)PRDTable;
    physAddress = ScsiPortGetPhysicalAddress(HwDeviceExtension,
                                             NULL,
                                             virtualAddress,
                                             &length);
    pCmdBlock->PRDTableLowPhyAddress = physAddress.LowPart;
    pCmdBlock->PRDTableHighPhyAddress = physAddress.HighPart;
    return TRUE;
}

/* Delay function in micro seconds resolution */
void mvMicroSecondsDelay(MV_VOID_PTR pSataAdapter, MV_U32 usecs)
{
    MV_U32  msecs = usecs / 1000;
    MV_U32  i;

    for (i = 0; i < msecs; i++)
    {
        ScsiPortStallExecution(1000);
    }
    ScsiPortStallExecution(usecs - (1000 * msecs));
}

