/*******************************************************************************
Copyright (C) Marvell International Ltd. and its affiliates

This software file (the "File") is owned and distributed by Marvell 
International Ltd. and/or its affiliates ("Marvell") under the following
alternative licensing terms.  Once you have made an election to distribute the
File under one of the following license alternatives, please (i) delete this
introductory statement regarding license alternatives, (ii) delete the two
license alternatives that you have not elected to use and (iii) preserve the
Marvell copyright notice above.

********************************************************************************
Marvell Commercial License Option

If you received this File from Marvell and you have entered into a commercial
license agreement (a "Commercial License") with Marvell, the File is licensed
to you under the terms of the applicable Commercial License.

********************************************************************************
Marvell GPL License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File in accordance with the terms and conditions of the General 
Public License Version 2, June 1991 (the "GPL License"), a copy of which is 
available along with the File in the license.txt file or by writing to the Free 
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 or 
on the worldwide web at http://www.gnu.org/licenses/gpl.txt. 

THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE IMPLIED 
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY 
DISCLAIMED.  The GPL License provides additional details about this warranty 
disclaimer.
********************************************************************************
Marvell BSD License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File under the following licensing terms. 
Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

    *   Redistributions of source code must retain the above copyright notice,
	    this list of conditions and the following disclaimer. 

    *   Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution. 

    *   Neither the name of Marvell nor the names of its contributors may be 
        used to endorse or promote products derived from this software without 
        specific prior written permission. 
    
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/
/*******************************************************************************
* mvWin2kScsiMiniPort.h - Header file for the Scsi MiniPort driver
*
* DESCRIPTION:
*       None.
*
* DEPENDENCIES:
*
*******************************************************************************/
#ifndef __INCmvWin2kScsiMiniPorth
#define __INCmvWin2kScsiMiniPorth

#ifdef MV_WINDOWS_2000
    #undef MV_SUPPORT_WMI
#endif

/* includes */
#include "miniport.h"
#include "scsi.h"
#ifdef MV_SUPPORT_WMI
    #include "scsiwmi.h"
#endif
#include "..\CoreDriver\mvSata.h"
#include "mvScsiAtaLayer.h"
#include "mvIALCommon.h"
#include <stdarg.h>

/* Defines */
#ifdef IOCTL
    #define MV_IOCTL_VERSION_SIGNATURE                  "VRSN_1_0"
    #define MV_IOCTL_VERSION_SIGNATURE_LENGTH           8
    #define MV_IOCTL_BUFFER_LENGTH                      (4096 + 28) /* in bytes */

/* Io control request codes */
    #define MV_IOCTL_GET_STATS                          0xface0000
    #define MV_IOCTL_GET_DRIVE_PARAMS                   0xface0001
    #define MV_IOCTL_READ_REGISTER                      0xface0005
    #define MV_IOCTL_WRITE_REGISTER                     0xface0006
    #define MV_IOCTL_MODIFY_QUEUED_DMA_MODE             0xface0007
    #define MV_IOCTL_READ_CORE_DEBUG_LEVEL              0xface0008
    #define MV_IOCTL_WRITE_CORE_DEBUG_LEVEL             0xface0009
    #define MV_IOCTL_SET_SATA_PHY_STATE                 0xface0010
    #define MV_IOCTL_SET_SATA_UNIT_INT_COAL             0xface0011
    #define MV_IOCTL_GET_ADAPTER_INFO                   0xface0014
    #define MV_IOCTL_READ_IAL_DEBUG_LEVEL               0xface0015
    #define MV_IOCTL_WRITE_IAL_DEBUG_LEVEL              0xface0016
    #define MV_IOCTL_DUMP_PERFORMANCE_COUNTERS          0xface0017
    #define MV_IOCTL_SET_MAX_OUTSTANDING_COMMANDS       0xface0018
#endif

/* Scatter/Gather Segment List Definitions */
/* Adapter limitations */
#define MAX_SG_DESCRIPTORS  40
#define MAX_TRANSFER_SIZE   (128*1024)/*(15*4096)*/

//#define MV_MAX_MODE_SENSE_RESULT_LENGTH 50


/* Typedefs */
#ifdef IOCTL
typedef struct _DRIVER_STATS
{
    MV_U32              totalIOs;
    MV_U32              totalAccumulatedOutstanding;
    MV_U32              totalSectorsTransferred;
    MV_U32              totalReadIOs;
    MV_U32              totalWriteIOs;
    MV_U32              totalVerifyIOs;
    MV_U32              totalReadsSectorsTransferred;
    MV_U32              totalWritesSectorsTransferred;
    MV_U32              totalVerifiesSectorsTransferred;
    MV_U32              totalMediumErrorsOccured;
    MV_U32              totalCRCErrorsOccured;
    MV_U32              totalWriteProtectErrorsOccured;
    MV_U32              totalIllegalRequestsOccured;
    MV_U8               lastErrorTypeOccured;
    MV_U32              lastErrorOccuredTimeStamp;
} DRIVER_STATS, *PDRIVER_STATS;
#endif

typedef struct
{
/*#ifdef IOCTL*/
    MV_BOOLEAN      forceNonQueued;
    MV_BOOLEAN      forceQueueDepth;
/*#endif*/
    MV_U8           queueDepth;
	MV_U8           channelQueueDepth;
}SATA_CHANNEL_INFO, *PSATA_CHANNEL_INFO;


typedef struct _HW_DEVICE_EXTENSION
{
    MV_SATA_ADAPTER         mvSataAdapter;  /* CoreDriver Adapter data structure*/
    PUCHAR                  requestsArrayBaseAlignedAddr;
    SCSI_PHYSICAL_ADDRESS   requestsArrayBaseDmaAlignedAddr;
    PUCHAR                  responsesArrayBaseAlignedAddr;
    SCSI_PHYSICAL_ADDRESS   responsesArrayBaseDmaAlignedAddr;
    MV_SATA_CHANNEL         mvSataChannels[MV_SATA_CHANNELS_NUM];
    SATA_CHANNEL_INFO       channelsInfo[MV_SATA_CHANNELS_NUM];
    MV_U8                   adapterBusNumber;
    MV_U8                   adapterDeviceNumber;
    MV_U32                  BAR0;
    BOOLEAN                 bDma64BitSupported;
    ULONG                   requestQSize;
    ULONG                   responseQSize;
#ifdef IOCTL
    DRIVER_STATS            stats[MV_SATA_CHANNELS_NUM];
    MV_U32                  numOfInterrupts;
#endif
    PVOID                   NotificationHandle;
#ifdef MV_SUPPORT_WMI
    SCSI_WMILIB_CONTEXT     WmiLibContext;
#endif
    MV_SAL_ADAPTER_EXTENSION  ataScsiAdapterExt;
    MV_IAL_COMMON_ADAPTER_EXTENSION ialCommonExt;
    BOOLEAN                 inDump;
} HW_DEVICE_EXTENSION, *PHW_DEVICE_EXTENSION;

typedef struct _SRB_EXTENSION
{
    MV_SATA_EDMA_PRD_ENTRY  PRDTable[MAX_SG_DESCRIPTORS + 2 + 4] ;
    PVOID           HwDeviceExtension;
    MV_UDMA_TYPE    commandType;
    MV_SATA_SCSI_CMD_BLOCK  SALBlock;
#ifdef MV_SATA_STORE_COMMANDS_INFO_ON_IAL_STACK
    MV_QUEUE_COMMAND_INFO   commandInfo;
#endif

} SRB_EXTENSION, *PSRB_EXTENSION;

#ifdef MV_SUPPORT_WMI
BOOLEAN
mvWmiSrb(
        IN     PHW_DEVICE_EXTENSION    HwDeviceExtension,
        IN OUT PSCSI_WMI_REQUEST_BLOCK Srb
        );

void mvWmiInitialize(
                    IN PHW_DEVICE_EXTENSION HwDeviceExtension
                    );
#endif

/* Locals */

#endif /*__INCmvOsWin2kh*/
